



#         ciforth : a generic I86 ISO FORTH by HCC FIG
# $Id: ciriscv.gnr,v 5.2 2023/03/30 19:25:37 albert Exp $
# Copyright (2012): Albert van der Horst by GNU Public License
# Expect for this comment, the file is identical to a working ciarm version 6.120.
#
#HCC FIG Holland : Hobby Computer Club, Forth Interest Group  Holland
# #   ciforth $Revision: 5.2 $
#
#
#
#
# If this is a configured assembly file, it should be accompanied with configured
# documentation (texinfo, ps, html.)
# WITHOUT THE DOCUMENTATION: GIVE UP! GET THE REAL THING!
# You have a configured system, if there are NO curly brackets on the next line.
#                          
#
# Configuration of this particular version:
# 64-bits
# running under Linux  #  with native forth I/O 

# Normally ciforth doesn't observe ISO >IN.
# Contains :
# (there may be no items here.)
#        Security words
#         Loadable words, i.e. all of ISO CORE, more than is needed
#           for a self contained kernel.
# A field in the header to point to source
#
#
# Gnu as version of ciforth created by ``m4'' from generic listing.
# This 64-bit source can be assembled in the following way:
#       as lina.s
#       ld -N a.out -o lina





 /* 
        A generic version of ISO FORTH for ARM processors
                Albert van der Horst

                in cooperation with
                HCC Forth user group
                The Netherlands
                www.forth.hccnet.nl

              based on
              FIG-FORTH
   implemented by:  Charlie Krajewski
                    205 ( BIG ) Blue Rd.
                    Middletown, CT  06457

  The listing has been made possible by the
  prior work of:
               Thomas Newman, Hayward, Ca.

 : other_acknowledgements
         John_Cassidy
         Kim_Harris
         George_Flammer
         Robert_D._Villwock ;

 : for tools
         Richard M. Stallman
         Linus Torvalds

No one who programs with FORTH can afford to be without:
  "Starting Forth  by Leo Brodie" and "Thinking Forth by Leo Brodie".
   Both out of print, but saved on the internet.

This Forth is a descendant in the 300+ (RCS)- generations from fig-Forth.

For nostalgic reasons the following comment has never been removed:
   Although there is much to be said for typing in your own
   listing and getting it running, there is much to be said
   not typing in your own listing.  If you feel that 100+
   pages of plinking is nutty, contact me for availability
   of a disc with source & executable files.  Obtainable at
   a bargain basement price, prepare yourself for bargain
   basement support.

All publications of the FORTH Interest Group are public domain.
They may be further distributed by the inclusion of this
credit notice:
               This publication has been made available by:

               FORTH Interest Group
               P.O. Box 1105
               San Carlos, Ca.  94070
[I feel obliged to keep this last one in (AH). Note that although it is
based on fig-Forth no stone is left unturned.]
 */          #
# ########################################################################################
#                       PREPARATION (no code)
# ########################################################################################
FIGREL  =     6       # FIG RELEASE #
FIGREV  =     0       # FIG REVISION #
USRVER  =     0      # USER VERSION NUMBER, a digit now
#
#      VERY ELEMENTARY .
CW      =     8    # Size of a cell in Forth, not in the bootcode.
ERRORSCREEN =     48    # Screen where the error messages start.
#
#      MEMORY LAYOUT.
# Normally this is specified at the m4 configuration level.
# For a configured system these values can be changed at this single place. 
NBUF    =     16   # No. of buffers, or screens 
KBBUF   =     1024      # Data bytes per disk buffer
US      =     0x40*CW  # User variable space, eats into next data stack, if any.
EM      =     (0x200000)     # Where the memory ends w.r.t. ORIG. 1) 
EMP     =     (EM-1)/0x1000+1 # Number of pages.
TAS     =     0x100000    #Size of unitialised area per task.
STACKSIZE   = TAS/4  #For both stacks.  but for now
TIBSIZE     = TAS/4  #TIB size. 
#


#
#      ASCII CHARACTER EQUIVALENTS
#
ABL     =     ' '     # SPACE
ACR     =     0x0D     # CR
AMS     =     '-'          # MINUS SIGN 
ASO     =     '['          # SQUARE BRACKET OPEN 
ASC     =     ']'          # SQUARE BRACKET CLOSE 
ADOT    =     '.'          # PERIOD
ALF     =     0x0A     # LINE FEED, USED INTERNALLY AS LINE ENDER 
AFF     =     0x0C     # FORM FEED
BELL    =     0x07     # ^G
BSIN    =     0x08     # INPUT DELETE CHARACTER
BSOUT   =     0x08     # OUTPUT BACKSPACE ( ^H )
#
#      HEADER RELATED EQUATES
#      dea is the dictionary entry address 
B_DUMMY   =     0x01     # dea is dummy, from namespace link
B_INVIS   =     0x02     # dea is invisible, "smudged".
B_IMMED   =     0x04     # dea is a immediate.
B_DENOT   =     0x08     # dea is a denotation.
C_HOFFSET =     0       # Offsets of code field in cells, w.r.t. dea
D_HOFFSET =     1       # Same for data field
F_HOFFSET =     2       # Same for flag field
L_HOFFSET =     3       # Same for link field
N_HOFFSET =     4       # Same for name field
S_HOFFSET =     5       # Same for source field
PH_OFFSET =     5+1   # Past header field: Start of data area. 
BD_OFFSET =     5+1+1 # Start of BODY for CREATEd word.
HEADSIZE  =     (CW*(PH_OFFSET))  #In cells, only to clean up source.
#
 #  


BPS     =     512             #Bytes/sector, common to all of MSDOS
SPB     =     KBBUF/BPS       #Sectors per buffer 
#







# Bring the names in line with what Forth want immediately. 

# ------------------------------------------------------------      
#   Start of constants stolen from C.                                 
# ------------------------------------------------------------      

SEEK_SET	=	0x0
SEEK_CUR	=	0x1
TCGETS	=	0x5401
TCSETS	=	0x5402
ECHO	=	0x8
EAGAIN	=	0xb
EINTR	=	0x4
EPIPE	=	0x20
VMIN	=	0x6
VTIME	=	0x5
ICANON	=	0x2
O_RDWR	=	0x2
O_RDONLY	=	0x0
O_WRONLY	=	0x1
O_CREAT	=	0x40
O_TRUNC	=	0x200
O_NONBLOCK	=	0x800
WEXITED	=	0x4
P_PID	=	0x1
AT_FDCWD	=	0xffffff9c
AT_REMOVEDIR	=	0x200
MAP_SHARED	=	0x1
MAP_PRIVATE	=	0x2
SIZE_TERMIO	=	0x3c

# Numbers of system calls. See "Linux kernel Internals" Appendix A. 
# By M.Beck, H. Boehme e.a. Addison Wesley.                         
# or in practice                                                    
# https://github.com/torvalds/linux/tree/master/arch/x86/entry/syscalls
# The system calls themselves are extensively documented in chapter 
# 2 of the man pages, e.g. "man 2 exit"
exit	=	0x5d
openat	=	0x38
close	=	0x39
unlinkat	=	0x23
mmap	=	0xde
chdir	=	0x31
read	=	0x3f
pselect6	=	0x48
write	=	0x40
ioctl	=	0x1d
lseek	=	0x3e
execve	=	0xdd
waitid	=	0x5f
clone	=	0xdc
kill	=	0x81
times	=	0x99
# ------------------------------------------------------------      
#   End of constants stolen from C.                                 
# ------------------------------------------------------------      

select = pselect6     # Can you believe this ?

RAWIO           =     (ECHO | ICANON)

# 


_main:



# allocatable, executable 
# Must be normal .text or otherwise linking with c-lib fails. 
# .text is normally not writable but linker -N option takes care of that. 
        
        .text

        #
        .global  _start       # Entry point.
        .global  start        # Entry point.
_ORIG:
_start:      # Entry point.
# 









NOBOOT:         # Skip till here if not booting.


ENDBOOT:

# ########################################################################################
#                       FORTH startup CODE + restarts + debug 
# ########################################################################################
#


# 4.15.3.2
start:
#
COLD_ENTRY:

        LA   a4, USINI
        MV     a0, sp              #Remember ARGS.
        SD     a0, CW*(31)(a4)
#


# Initialisation of the Forth engine.
        LA     a4, USINI
        LD     a0, CW*2(a4)  #Parameter stack
        MV     sp, a0           #Parameter stack
        LD     fp, CW*3(a4)  #Return stack
        LA     a7, COLD + CW*(PH_OFFSET)    #Interpreter pointer
      LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  



#
#
# ########################################################################################
#                       FORTH ITSELF
# ########################################################################################
#
 /* 
   FORTH REGISTERS
   The names under FORTH are used in the generic source.

   FORTH   ARM      FORTH PRESERVATION RULES
   -----   ----     ----- ------------ -----
           a0-a3   Used within and cluttered by Forth words.
                    across FORTH words.
                    The first of those is WOR (probably)

   WOR   a0      Working register.  When entering a word
                    via its code field the DEA is passed in WOR.

            a0      Scratch register.  Used to pass data from
                    FORTH words, see label APUSH or macro _APUSH

            a1      General register.  Used when two data items are passed
                    from FORTH words, see label DPUSH or macro _DPUSH

   HIP   a7      High level Interpreter Pointer.  Must be preserved

   RPO   fp      Return stack pointer.  Must be preserved across
                    FORTH words.

   SPO   sp      Parameter stack pointer.  Must be preserved
                    across FORTH words.

           PC       Machine Process Counter, sometimes used
                    for relative addressing

This is the ARM asignment, not valid for 64 bit.

   a13     sp      ARM stack pointer.
                   In thumb this can be used with pop and push.

   a14     RL      Machine Link Register for calls.
                    Can be used within Forth words, with care.

  a15     PC        Machine Process Counter (Program counter)
                    Cannot be used within Forth words.



            The remaining registers are free to use.
            a0-a3 are the normal Forth work registers.
            Within assembler words a0 can be used too,
            as it has served its purpose.

            Here are some  useful code sequences.
            For next:
                    LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
            For push single and next:
                    ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
            For push double and next:
                    ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  



----------------------------------------------------------
 */          #
 /* 
---------------------------------------------

   COMMENT CONVENTIONS
   ------- -----------

   =       IS EQUAL TO
   <-      ASSIGNMENT

  NAME        =  Address of name
  (NAME)      =  Contents of name
  DEA         =  DICTIONARY ENTRY ADDRESS
                 fields have fixed offsets from DEA
  CFA         =  CODE FIELD ADDRESS : a pointer to executable code
  DFA         =  DATA FIELD ADDRESS : a pointer to
                        data/high level code/ DOES> pointer
  FFA         =  FLAG FIELD ADDRESS: contains flags
  LFA         =  LINK FIELD ADDRESS: a pointer to the next DEA
  NFA         =  NAME FIELD ADDRESS: a pointer to a name buffer
  SFA =  SOURCE FIELD ADDRESS: a pointer into source code
  PHA         =  POST HEADER ADDRESS: After all fields

  S1          =  Parameter stack - 1st cell
  S2          =  Parameter stack - 2nd cell
  RP1          =  Return stack    - 1st cell
  a2          =  Return stack    - 2nd cell

  LSB         =  Least significant bit
  MSB         =  Most  significant bit
  LB          =  Low byte
  HB          =  High byte
  LC          =  Low  cell
  HC          =  High cell

------------------------------------------------------------
 */  # 
        #

#
# 
# There may be no jumps to NEXT at all 
# The label NEXT1 is rarely relevant (for _OLDDEBUG_) 
DPUSH:  ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp)       # Fall through.
APUSH:  ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
NEXT:
#
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))                  # R0 <- (HIP) 
NEXT1:
        LD     a1, 0(a0)
        JALR      a1     # TO `CFA'
#
#       Dictionary starts here.


#  ************
#  *   ONLY   *
#  ************
#
        .align 3,0
N_ONLY:
        .dword      4
        .ascii      "ONLY"
        .align 3,0
ONLY:
        .dword    DODOE
        .dword    ONLY+HEADSIZE
        .dword    0x0
        .dword    0
        .dword    N_ONLY
        .dword    0

        .dword      DOVOC
        .dword      FORTH     # NEXT NAMESPACE 
ONLYBODY:

        .align 3,0

        .dword    0x0
        .dword    0
        .dword    B_DUMMY
        .dword    STARTOFONLY
        .dword    (ZERO+(CW*(D_HOFFSET)))
        .dword    0

#

#  *******************
#  *   ENVIRONMENT   *
#  *******************
#
        .align 3,0
N_ENV:
        .dword      11
        .ascii      "ENVIRONMENT"
        .align 3,0
ENV:
        .dword    DODOE
        .dword    ENV+HEADSIZE
        .dword    0x0
        .dword    ONLY
        .dword    N_ENV
        .dword    0

        .dword      DOVOC
        .dword      ONLY       # NEXT NAMESPACE 

        .align 3,0

        .dword    0x0
        .dword    0
        .dword    B_DUMMY
        .dword    STARTOFENVIRONMENT
        .dword    (ZERO+(CW*(D_HOFFSET)))
        .dword    0

#

#  ************
#  *   COLD   *
#  ************
#
        .align 3,0
N_COLD:
        .dword      4
        .ascii      "COLD"
        .align 3,0
COLD:
        .dword    DOCOL
        .dword    COLD+HEADSIZE
        .dword    0x0
        .dword    ENV
        .dword    N_COLD
        .dword    0

# Fill user area for single task.
        .dword      LIT, USINI      # Fails if SP/RP are not usable!
        .dword      LIT, USINI+(CW*(1)), FETCH
        .dword      LIT, US
        .dword      LMOVE
        .dword      INIT
        .dword      ONE            # Sign on wanted.


        .dword      ZERO, LIT, TCGETS, TERMIO
        .dword      LIT, ioctl, XOS, ZLESS
        .dword      ZBRAN
        .dword      COLD1-.-CW
        # Not a terminal. Cannot set the terminal, get rid of "OK".
        .dword      DROP, ZERO      # Sign on suppressed.
        .dword      LIT,TASK  # Suppress OK for pipes.
        .dword      LIT,OK
        .dword      LIT, (CW*(3)), LMOVE
COLD1:
 # 

        .dword      SKIP
         .dword      8
SB0: .ascii      "/dev/mem"
       .align 3,0
        .dword      LIT, SB0
        .dword      LIT, 8
        .dword      LIT, 0x00181002
        .dword      OFILE, LOR, DEVMEM, STOR
#
       .dword      OPTIONS
        .dword      ZBRAN
        .dword      COLD5-.-CW

        .dword      SIGNON    # Suppressed for scripting! Or any options.
COLD5:
        .dword      ABORT
        .dword      BYE     # In case of turnkey programs.
        .dword      SEMIS   # Unnecessary, but helpful for decompilation.
#

#  ************
#  *   USER   *
#  ************
#
        .align 3,0
N_USER:
        .dword      4
        .ascii      "USER"
        .align 3,0
USER:
        .dword    DOCOL
        .dword    USER+HEADSIZE
        .dword    0x0
        .dword    COLD
        .dword    N_USER
        .dword    0

        .dword      LCONST
        .dword      PSCOD
DOUSE:  LD     a3, (CW*(D_HOFFSET))(a0)    #PFA  
        LA     a0, USINI
        LD     a0,0(a0)
        ADD     a0, a0,a3      #ADDR OF VARIABLE
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  

#

#  ***************
#  *   +ORIGIN   *
#  ***************
#
        .align 3,0
N_PORIG:
        .dword      7
        .ascii      "+ORIGIN"
        .align 3,0
PORIG:
        .dword    DOCOL
        .dword    PORIG+HEADSIZE
        .dword    0x0
        .dword    USER
        .dword    N_PORIG
        .dword    0

        .dword      LIT
        .dword      USINI
        .dword      PLUS
        .dword      SEMIS
#
#      Initialisation block for user variables through DOC-LINK
#       <<<<< must be in same order as user variables >>>>>
# To be fig compatible, there should be two vectors here  
# for the cold and warm start.
#
# ------------------------ Forth entry code, hooks for restart -------------
# FIXME. You want to jump here from -2 CELLS +USINI / -4 CELLS +USINI 
        LA     a4, COLD
        MV     a0, a4
        LD     a1, 0(a0)
        JALR      a1          #Hope stacks are still okay.
        .align 3,0
# Make it possible to revector WARM. You can even make it low level. 
        LA      a4, WARM
        MV      a0,a4
        LD     a1, 0(a0)          #Hope stacks are still okay.
        JALR      a1
        .align 3,0
USINI:  .dword      STRUSA
        .dword      STRUSA  # INIT (U0) user area of the main task 1
        .dword      INITS0  # INIT (S0)         2
        .dword      INITR0  - (CW*(1))
                        # INIT (R0)         3
        .dword      STRTIB  # INIT (TIB)        4
        .dword      BSIN    # RUBOUT: get rid of latest char 5
        .dword      0       # AVAILABLE         6
        .dword      1       # INIT (WARNING)     7
        .dword      INITDP  #      INIT (FENCE)  8
DPA:    .dword      INITDP  #      INIT (DP)     9
        .dword      ENV #       INIT (VOC-LINK) 10
        .dword      39+16        # INIT (MAX-USER) 
#
        .dword      0, 0            # WHERE             12 13 
        .dword      0, 0            #REMAINDER   14 15 

        .space    US-(. - USINI)        # All user can be initialised.
#
#      <<<<< end of data used by cold start >>>>>
#


#  **************
#  *   VMA-IO   *
#  **************
#
        .align 3,0
N_VMAIO:
        .dword      6
        .ascii      "VMA-IO"
        .align 3,0
VMAIO:
        .dword    DOVAR
        .dword    VMAIO+HEADSIZE
        .dword    0x0
        .dword    PORIG
        .dword    N_VMAIO
        .dword    0

        .dword 0xB2000000
#

#  *******************
#  *   PERIPHERALS   *
#  *******************
#
        .align 3,0
N_PERIPHERALS:
        .dword      11
        .ascii      "PERIPHERALS"
        .align 3,0
PERIPHERALS:
        .dword    DOVAR
        .dword    PERIPHERALS+HEADSIZE
        .dword    0x0
        .dword    VMAIO
        .dword    N_PERIPHERALS
        .dword    0

        .dword 0x03000000
#

#  ***************
#  *   DEV-MEM   *
#  ***************
#
        .align 3,0
N_DEVMEM:
        .dword      7
        .ascii      "DEV-MEM"
        .align 3,0
DEVMEM:
        .dword    DOVAR
        .dword    DEVMEM+HEADSIZE
        .dword    0x0
        .dword    PERIPHERALS
        .dword    N_DEVMEM
        .dword    0

        .dword 0
#

#  ***************
#  *   MMAP-IO   *
#  ***************
#
        .align 3,0
N_MMAPIO:
        .dword      7
        .ascii      "MMAP-IO"
        .align 3,0
MMAPIO:
        .dword    DOCOL
        .dword    MMAPIO+HEADSIZE
        .dword    0x0
        .dword    DEVMEM
        .dword    N_MMAPIO
        .dword    0

# The libc mmap definition, not necessarily documenting the system call! 
# mmap( void *addr, size_t length, int prot, int flags, int fd,
# off_t offset );     see man mmap

      .dword     DEVMEM, FETCH, LDUP, ZERO, GREAT
      .dword      ZBRAN
        .dword      MMAP2-.-CW
      .dword     TOR
      .dword     VMAIO, FETCH
      .dword     LIT, 0x010000000 # length of mapped area
      .dword     LIT, 3                  # protection-bits
      .dword     LIT, MAP_SHARED         # flags
      .dword     FROMR
      .dword     PERIPHERALS, FETCH

      .dword     ZERO                # dummy 
      .dword     LIT, mmap, XOS7
MMAP2:
      .dword     VMAIO, STOR
      .dword      SEMIS   #Unnecessary, but helpful for decompilation.
#
#

#

#  *************
#  *   ABORT   *
#  *************
#
        .align 3,0
N_ABORT:
        .dword      5
        .ascii      "ABORT"
        .align 3,0
ABORT:
        .dword    DOCOL
        .dword    ABORT+HEADSIZE
        .dword    0x0
        .dword    MMAPIO
        .dword    N_ABORT
        .dword    0

        .dword      SZERO, FETCH, SPSTO
        .dword      ZERO, HANDLER, STOR
        .dword      QUIT
        .dword      SEMIS   #Unnecessary, but helpful for decompilation.
#
#

#  ************
#  *   WARM   *
#  ************
#
        .align 3,0
N_WARM:
        .dword      4
        .ascii      "WARM"
        .align 3,0
WARM:
        .dword    DOCOL
        .dword    WARM+HEADSIZE
        .dword    0x0
        .dword    ABORT
        .dword    N_WARM
        .dword    0

        .dword      INIT
        .dword      SIGNON
        .dword      ABORT
        .dword      SEMIS   #Unnecessary, but helpful for decompilation.
#

#  ************
#  *   NOOP   *
#  ************
#
        .align 3,0
N_NOOP:
        .dword      4
        .ascii      "NOOP"
        .align 3,0
NOOP:
        .dword    NOOP+HEADSIZE
        .dword    NOOP+HEADSIZE
        .dword    0x0
        .dword    WARM
        .dword    N_NOOP
        .dword    0

       LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#
# 
#

#  ***********
#  *   LIT   *
#  ***********
#
        .align 3,0
N_LIT:
        .dword      3
        .ascii      "LIT"
        .align 3,0
LIT:
        .dword    LIT+HEADSIZE
        .dword    LIT+HEADSIZE
        .dword    0x0
        .dword    NOOP
        .dword    N_LIT
        .dword    0

        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))                  # R0 <- (HIP) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0            # TO TOP OF STACK
#

#todo EXECUTE
#  ***************
#  *   EXECUTE   *
#  ***************
#
        .align 3,0
N_EXEC:
        .dword      7
        .ascii      "EXECUTE"
        .align 3,0
EXEC:
        .dword    EXEC+HEADSIZE
        .dword    EXEC+HEADSIZE
        .dword    0x0
        .dword    LIT
        .dword    N_EXEC
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      # GET XT
        LD    t1, 0(a0)
        JALR     t1  #(IP) <- (PFA)

#

#  ***************
#  *   RECURSE   *
#  ***************
#
        .align 3,0
N_RECURSE:
        .dword      7
        .ascii      "RECURSE"
        .align 3,0
RECURSE:
        .dword    DOCOL
        .dword    RECURSE+HEADSIZE
        .dword    B_IMMED
        .dword    EXEC
        .dword    N_RECURSE
        .dword    0

        .dword      LATEST, COMMA
        .dword      SEMIS
#

#  **************
#  *   BRANCH   *
#  **************
#
        .align 3,0
N_BRAN:
        .dword      6
        .ascii      "BRANCH"
        .align 3,0
BRAN:
        .dword    (SKIP+HEADSIZE)
        .dword    BRAN+HEADSIZE
        .dword    0x0
        .dword    RECURSE
        .dword    N_BRAN
        .dword    0

#

#  ************
#  *   SKIP   *
#  ************
#
        .align 3,0
N_SKIP:
        .dword      4
        .ascii      "SKIP"
        .align 3,0
SKIP:
        .dword    SKIP+HEADSIZE
        .dword    SKIP+HEADSIZE
        .dword    0x0
        .dword    BRAN
        .dword    N_SKIP
        .dword    0

BRAN1:
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))                  # R0 <- (HIP) 

        ADDI     a0, a0, -1
        ORI     a0, a0, (CW-1)
        ADDI     a0, a0, 1
#
        ADD     a7,a7,a0
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***************
#  *   0BRANCH   *
#  ***************
#
        .align 3,0
N_ZBRAN:
        .dword      7
        .ascii      "0BRANCH"
        .align 3,0
ZBRAN:
        .dword    ZBRAN+HEADSIZE
        .dword    ZBRAN+HEADSIZE
        .dword    0x0
        .dword    SKIP
        .dword    N_ZBRAN
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      # GET STACK VALUE
        BNEZ      a0, ZBRAN1   # YES, BRANCH
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))                  # R0 <- (HIP) 
        ADD     a7,a7,a0
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
ZBRAN1:
        ADDI      a7,a7,(CW*(1))
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***************
#  *   (+LOOP)   *
#  ***************
#
        .align 3,0
N_XPLOOP:
        .dword      7
        .ascii      "(+LOOP)"
        .align 3,0
XPLOOP:
        .dword    XPLOOP+HEADSIZE
        .dword    XPLOOP+HEADSIZE
        .dword    0x0
        .dword    ZBRAN
        .dword    N_XPLOOP
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      # GET LOOP VALUE
        LD     a2,0(fp)  # INDEX = INDEX + INCR
        ADD     a2, a2,a3
        SD     a2,0(fp)  # GET NEW INDEX
        LD     a4, (CW*(1))(fp)
	SUB	a2, a2, a4
        XOR     a0, a2,a3  # DEFAULT
        SLT     a0, a0, zero     # LT, KEEP LOOPING
        ADDI     a0, a0, -1      #TRUE
XPO1: ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   (DO)   *
#  ************
#
        .align 3,0
N_XDO:
        .dword      4
        .ascii      "(DO)"
        .align 3,0
XDO:
        .dword    XDO+HEADSIZE
        .dword    XDO+HEADSIZE
        .dword    0x0
        .dword    XPLOOP
        .dword    N_XDO
        .dword    0

        LD      a2, 0(a7)
        ADDI    a7, a7, (CW*(1))                  # R0 <- (HIP) 
        ADD     a2, a2,a7  #Make absolute
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))           # INITIAL INDEX VALUE
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))           # LIMIT VALUE
        
         MV     t0, fp
         MV     fp, sp
         MV     sp, t0
   # GET RETURN STACK, USES REGISTER t0 ! 
        ADDI    sp, sp, -(CW*(1))
        SD      a2, 0(sp)       # Target location.
        ADDI    sp, sp, -(CW*(1))
        SD      a3, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        
         MV     t0, fp
         MV     fp, sp
         MV     sp, t0
   # GET PARAMETER STACK, USES REGISTER t0 ! 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   (?DO)   *
#  *************
#
        .align 3,0
N_XQDO:
        .dword      5
        .ascii      "(?DO)"
        .align 3,0
XQDO:
        .dword    XQDO+HEADSIZE
        .dword    XQDO+HEADSIZE
        .dword    0x0
        .dword    XDO
        .dword    N_XQDO
        .dword    0

        LD      a2, 0(a7)
        ADDI    a7, a7, (CW*(1))                  # R0 <- (HIP) 
        ADD     a2, a2,a7  #Make absolute
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      # INITIAL INDEX VALUE
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      # LIMIT VALUE
        BGE            a1,a3, QXDO1
        
         MV     t0, fp
         MV     fp, sp
         MV     sp, t0
   # GET RETURN STACK, USES REGISTER t0 ! 
        ADDI    sp, sp, -(CW*(1))
        SD      a2, 0(sp)       # Target location.
        ADDI    sp, sp, -(CW*(1))
        SD      a3, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        
         MV     t0, fp
         MV     fp, sp
         MV     sp, t0
   # GET PARAMETER STACK, USES REGISTER t0 ! 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
QXDO1:  MV     a7,a2
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *********
#  *   I   *
#  *********
#
        .align 3,0
N_IDO:
        .dword      1
        .ascii      "I"
        .align 3,0
IDO:
        .dword    IDO+HEADSIZE
        .dword    IDO+HEADSIZE
        .dword    0x0
        .dword    XQDO
        .dword    N_IDO
        .dword    0

        LD     a0,0(fp) # GET INDEX VALUE
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0            # TO PARAMETER STACK
#

#  *********
#  *   J   *
#  *********
#
        .align 3,0
N_JDO:
        .dword      1
        .ascii      "J"
        .align 3,0
JDO:
        .dword    JDO+HEADSIZE
        .dword    JDO+HEADSIZE
        .dword    0x0
        .dword    IDO
        .dword    N_JDO
        .dword    0

        LD     a0,(CW*(3))(fp)  # GET INDEX VALUE
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0            # TO PARAMETER STACK
#

#  **************
#  *   UNLOOP   *
#  **************
#
        .align 3,0
N_UNLOOP:
        .dword      6
        .ascii      "UNLOOP"
        .align 3,0
UNLOOP:
        .dword    UNLOOP+HEADSIZE
        .dword    UNLOOP+HEADSIZE
        .dword    0x0
        .dword    JDO
        .dword    N_UNLOOP
        .dword    0

        ADDI fp,fp,(CW*(3))
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   DIGIT   *
#  *************
#
        .align 3,0
N_DIGIT:
        .dword      5
        .ascii      "DIGIT"
        .align 3,0
DIGIT:
        .dword    DOCOL
        .dword    DIGIT+HEADSIZE
        .dword    0x0
        .dword    UNLOOP
        .dword    N_DIGIT
        .dword    0

        .dword      TOR
        .dword      LDUP, LIT, 'A', LESS
        .dword      LIT,  'A'-'0'-10, LAND, PLUS
        .dword      LIT,  'A'-10, LSUB
        .dword      LDUP, FROMR, ULESS
        .dword SEMIS
#
#  **************
#  *   DIGIT0   *
#  **************
#
        .align 3,0
N_DIGIT0:
        .dword      6
        .ascii      "DIGIT0"
        .align 3,0
DIGIT0:
        .dword    DIGIT0+HEADSIZE
        .dword    DIGIT0+HEADSIZE
        .dword    0x0
        .dword    DIGIT
        .dword    N_DIGIT0
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #NUMBER BASE
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ASCII DIGIT
        AND     a0, a0,0xFF
        AND     a1, a1,0xFF
        LI      t0,  '0'
        BLT     a0,t0, DIGI2   #NUMBER ERROR
        LI      t0, 10
        BLT     t0, a0, DIGI1   #NUMBER = 0 THRU 9
        ADDI    a0, a0,-7
        LI      t0,  10
        BLT     t0, a0, DIGI2   #NUMBER 'A' THRU 'Z'?
DIGI1:
        BGE     a0, a1, DIGI2   # COMPARE NUMBER TO BASE
        MV     a1,a0   #NEW BINARY NUMBER

        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0            #ADD TO STACK
#   NUMBER ERROR
DIGI2:  LI    a0, 0 #FALSE FLAG
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **************
#  *   ~MATCH   *
#  **************
#
        .align 3,0
N_NMATCH:
        .dword      6
        .ascii      "~MATCH"
        .align 3,0
NMATCH:
        .dword    DOCOL
        .dword    NMATCH+HEADSIZE
        .dword    0x0
        .dword    DIGIT0
        .dword    N_NMATCH
        .dword    0

#
        .dword      TOR                           # Remember dea
        .dword      OVER
        .dword      RR, TNFA, FETCH, SFET
        .dword      CORA
        .dword      FROMR, SWAP                   # Leave flag on top of dea.
        .dword      SEMIS                            #
#

#  **************
#  *   ?BLANK   *
#  **************
#
        .align 3,0
N_QBL:
        .dword      6
        .ascii      "?BLANK"
        .align 3,0
QBL:
        .dword    DOCOL
        .dword    QBL+HEADSIZE
        .dword    0x0
        .dword    NMATCH
        .dword    N_QBL
        .dword    0

        .dword      LBL, ONEP, LESS
        .dword      SEMIS
#

#  ************
#  *   PP@@   *
#  ************
#
        .align 3,0
N_PPFET:
        .dword      4
        .ascii      "PP@@"
        .align 3,0
PPFET:
        .dword    DOCOL
        .dword    PPFET+HEADSIZE
        .dword    0x0
        .dword    QBL
        .dword    N_PPFET
        .dword    0

        .dword      SRC, CELLP, TFET
        .dword      OVER, EQUAL
        .dword      ZBRAN
        .dword      PPFET1-.-CW
        .dword      ZERO
        .dword      BRAN
        .dword      PPFET2-.-CW
PPFET1:
        .dword      LDUP
        
        .dword      CFET
        .dword      ONE, PP, PSTOR
PPFET2:
        .dword      SEMIS
#

#  ************
#  *   NAME   *
#  ************
#
        .align 3,0
N_LNAME:
        .dword      4
        .ascii      "NAME"
        .align 3,0
LNAME:
        .dword    DOCOL
        .dword    LNAME+HEADSIZE
        .dword    0x0
        .dword    PPFET
        .dword    N_LNAME
        .dword    0

        .dword      X
PWORD0: .dword      DROP
        .dword      PPFET, QBL
        .dword      OVER, SRC, CELLP, FETCH, LSUB # At end?
        .dword      LAND, ZEQU
        .dword      ZBRAN
        .dword      PWORD0-.-CW

        .dword      X
PWORD1: .dword      DROP
        .dword      PPFET, QBL
        .dword      ZBRAN
        .dword      PWORD1-.-CW

        .dword      OVER, LSUB
        
        .dword      SEMIS
#

#  *************
#  *   PARSE   *
#  *************
#
        .align 3,0
N_PARSE:
        .dword      5
        .ascii      "PARSE"
        .align 3,0
PARSE:
        .dword    DOCOL
        .dword    PARSE+HEADSIZE
        .dword    0x0
        .dword    LNAME
        .dword    N_PARSE
        .dword    0

        .dword      SRC, CELLP, TFET
        .dword      OVER, LSUB
        
        .dword      ROT, SSLAS, TSWAP
        .dword      ZEQU
        .dword      ZBRAN
        .dword      PARSE8-.-CW
        .dword      DROP, SRC, CELLP, FETCH
PARSE8: 
        .dword      PP, STOR
        .dword SEMIS
#

#  ***********
#  *   SRC   *
#  ***********
#
        .align 3,0
N_SRC:
        .dword      3
        .ascii      "SRC"
        .align 3,0
SRC:
        .dword    DOUSE
        .dword    (CW*(27))
        .dword    0x0
        .dword    PARSE
        .dword    N_SRC
        .dword    0
      # And 28 and 29.
#

#  **************
#  *   SOURCE   *
#  **************
#
        .align 3,0
N_SOURCE:
        .dword      6
        .ascii      "SOURCE"
        .align 3,0
SOURCE:
        .dword    DOCOL
        .dword    SOURCE+HEADSIZE
        .dword    0x0
        .dword    SRC
        .dword    N_SOURCE
        .dword    0

        .dword      SRC, TFET
        .dword      SWAP, OVER, LSUB
        .dword      SEMIS
#

#

#  **********
#  *   CR   *
#  **********
#
        .align 3,0
N_CR:
        .dword      2
        .ascii      "CR"
        .align 3,0
CR:
        .dword    DOCOL
        .dword    CR+HEADSIZE
        .dword    0x0
        .dword    SOURCE
        .dword    N_CR
        .dword    0

        .dword      LIT,ALF
        .dword      EMIT
        .dword      ZERO, LOUT, STOR
        .dword      SEMIS
#

#  *************
#  *   CMOVE   *
#  *************
#
        .align 3,0
N_LCMOVE:
        .dword      5
        .ascii      "CMOVE"
        .align 3,0
LCMOVE:
        .dword    LCMOVE+HEADSIZE
        .dword    LCMOVE+HEADSIZE
        .dword    0x0
        .dword    CR
        .dword    N_LCMOVE
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))      #count
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #dest
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #source
        BEQZ a2, MOVRP2
MOVREP: LB     a0, 0(a3)
        ADDI    a3, a3, 1 
        SB     a0, 0(a1)
        ADDI    a1, a1, 1 
        ADDI    a2, a2, -1
        BNEZ    a2, MOVREP
MOVRP2: 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   MOVE   *
#  ************
#
        .align 3,0
N_LMOVE:
        .dword      4
        .ascii      "MOVE"
        .align 3,0
LMOVE:
        .dword    LMOVE+HEADSIZE
        .dword    LMOVE+HEADSIZE
        .dword    0x0
        .dword    LCMOVE
        .dword    N_LMOVE
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))      #count,  R2
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #dest,   R1
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #source, R3
        BEQZ     a2, MOVE3
        BGEU     a3, a1, MOVEF
        ADD     a1, a1,a2
        ADD     a3, a3,a2
        ADDI     a1, a1, -1
        ADDI     a3, a3, -1
MOVER:
        LB     a0, 0(a3)
        ADDI    a3, a3, -1 
        SB     a0, 0(a1)
        ADDI    a1, a1, -1 
        ADDI     a2, a2, -1
        BNEZ    a2, MOVER
        JAL     MOVE3
MOVEF:
        LB     a0, 0(a3)
        ADDI    a3, a3, 1 
        SB     a0, 0(a1)
        ADDI    a1, a1, 1 
        ADDI     a2, a2, -1
        BNEZ    a2, MOVEF
MOVE3:
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   UM*   *
#  ***********
#
        .align 3,0
N_USTAR:
        .dword      3
        .ascii      "UM*"
        .align 3,0
USTAR:
        .dword    USTAR+HEADSIZE
        .dword    USTAR+HEADSIZE
        .dword    0x0
        .dword    LMOVE
        .dword    N_USTAR
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        MUL      a1,a3,a2
        MULHU    a0,a3,a2
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#




 /* 
_BITS_32(  division of unsigned-64/unsigned-32)
_BITS_64(  division of unsigned-128/unsigned-64)
  input:
  div  = a2  = divisor
  modh = a1 = MSW of dividend
  modl = a0  = LSW of dividend
  output:
  quot = a0  = quotient
  rem = a1 = remainder

  Internally a3 is used as a mask, with 1 bit set.
  mask = a3

 */  
#  **************
#  *   UM/MOD   *
#  **************
#
        .align 3,0
N_USLAS:
        .dword      6
        .ascii      "UM/MOD"
        .align 3,0
USLAS:
        .dword    USLAS+HEADSIZE
        .dword    USLAS+HEADSIZE
        .dword    0x0
        .dword    USTAR
        .dword    N_USLAS
        .dword    0

        ld      a4, (CW*(0))(sp)        #LSW OF DIVISOR
        ld      a0, (CW*(1))(sp)        #MSW OF DIVIDEND
        ld      a3, (CW*(2))(sp)        #LSW OF DIVIDEND
        add     sp, sp, (CW*(3))

        ori     t1, zero, 1		
        slli    t1, t1, (CW*8-1)   # init mask with highest bit set
        ori     a1, zero,0		# init quot

        bgeu     a0, a4, DZERO      #Overflow.

UMSLAS3:
        slt   t0, a3, zero
        slt   a6,  a0, zero
        sll     a0, a0, 1	#   double precision shift (modh, modl)
        sll     a3, a3, 1	#   ADD with carry and set flags again !
        add     a0, a0, t0
#        sleu    t0, a4, a0               
#        or      t0, t0, a6       
#        beqz    t0,UMSLAS2       
        bleu    a4, a0, UMSLAS4
        beqz    a6,UMSLAS2
UMSLAS4:
        add     a1, a1, t1
        sub     a0, a0, a4
UMSLAS2:

        srli    t1, t1,1
        bnez    t1, UMSLAS3
UMSLAS9:

    add    sp, sp, -2*CW
    sd     a0, CW(sp)
    sd      a1, 0(sp)
    LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
# Don't crash return zero's. 
DZERO:
    add    sp, sp, -2*CW
    sd     a1, CW(sp)
    sd      a1, 0(sp)
    LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  

#

#  ***********
#  *   AND   *
#  ***********
#
        .align 3,0
N_LAND:
        .dword      3
        .ascii      "AND"
        .align 3,0
LAND:
        .dword    LAND+HEADSIZE
        .dword    LAND+HEADSIZE
        .dword    0x0
        .dword    USLAS
        .dword    N_LAND
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        AND     a0, a0,a3
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   OR   *
#  **********
#
        .align 3,0
N_LOR:
        .dword      2
        .ascii      "OR"
        .align 3,0
LOR:
        .dword    LOR+HEADSIZE
        .dword    LOR+HEADSIZE
        .dword    0x0
        .dword    LAND
        .dword    N_LOR
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      # (S1) <- (S1) OR (S2)
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        OR      a0, a0,a3
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   XOR   *
#  ***********
#
        .align 3,0
N_LXOR:
        .dword      3
        .ascii      "XOR"
        .align 3,0
LXOR:
        .dword    LXOR+HEADSIZE
        .dword    LXOR+HEADSIZE
        .dword    0x0
        .dword    LOR
        .dword    N_LXOR
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      # (S1) <- (S1) XOR (S2)
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        XOR     a0, a0,a3
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **************
#  *   INVERT   *
#  **************
#
        .align 3,0
N_INVERT:
        .dword      6
        .ascii      "INVERT"
        .align 3,0
INVERT:
        .dword    INVERT+HEADSIZE
        .dword    INVERT+HEADSIZE
        .dword    0x0
        .dword    LXOR
        .dword    N_INVERT
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      # (S1) <- (S1) XOR (S2)
        NOT  a0, a0
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   DSP@   *
#  ************
#
        .align 3,0
N_SPFET:
        .dword      4
        .ascii      "DSP@"
        .align 3,0
SPFET:
        .dword    SPFET+HEADSIZE
        .dword    SPFET+HEADSIZE
        .dword    0x0
        .dword    INVERT
        .dword    N_SPFET
        .dword    0

        MV     a0,sp   # (S1) <- (SP)
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   DSP!   *
#  ************
#
        .align 3,0
N_SPSTO:
        .dword      4
        .ascii      "DSP!"
        .align 3,0
SPSTO:
        .dword    SPSTO+HEADSIZE
        .dword    SPSTO+HEADSIZE
        .dword    0x0
        .dword    SPFET
        .dword    N_SPSTO
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        MV     sp,a0        #RESET PARAM STACK POINTER
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   DEPTH   *
#  *************
#
        .align 3,0
N_DEPTH:
        .dword      5
        .ascii      "DEPTH"
        .align 3,0
DEPTH:
        .dword    DOCOL
        .dword    DEPTH+HEADSIZE
        .dword    0x0
        .dword    SPSTO
        .dword    N_DEPTH
        .dword    0

        .dword      SZERO, FETCH
        .dword      SPFET
        .dword      LSUB
        .dword      LIT, CW, SLASH
        .dword      ONEM
        .dword      SEMIS
#

#  ***********
#  *   CLS   *
#  ***********
#
        .align 3,0
N_CLS:
        .dword      3
        .ascii      "CLS"
        .align 3,0
CLS:
        .dword    DOCOL
        .dword    CLS+HEADSIZE
        .dword    0x0
        .dword    DEPTH
        .dword    N_CLS
        .dword    0

        .dword      SZERO, FETCH
        .dword      SPSTO
        .dword      SEMIS
#



#  ************
#  *   RSP@   *
#  ************
#
        .align 3,0
N_RPFET:
        .dword      4
        .ascii      "RSP@"
        .align 3,0
RPFET:
        .dword    RPFET+HEADSIZE
        .dword    RPFET+HEADSIZE
        .dword    0x0
        .dword    CLS
        .dword    N_RPFET
        .dword    0
      #(S1) <- (RP)
        ADDI    sp, sp, -(CW*(1))
        SD      fp, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   RSP!   *
#  ************
#
        .align 3,0
N_RPSTO:
        .dword      4
        .ascii      "RSP!"
        .align 3,0
RPSTO:
        .dword    RPSTO+HEADSIZE
        .dword    RPSTO+HEADSIZE
        .dword    0x0
        .dword    RPFET
        .dword    N_RPSTO
        .dword    0

        LD      fp, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   EXIT   *
#  ************
#
        .align 3,0
N_EXIT:
        .dword      4
        .ascii      "EXIT"
        .align 3,0
EXIT:
        .dword    EXIT+HEADSIZE
        .dword    EXIT+HEADSIZE
        .dword    0x0
        .dword    RPSTO
        .dword    N_EXIT
        .dword    0

        LD      a7, 0(fp)
        ADDI    fp, fp, (CW*(1))  #(IP) <- (RP1)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   CO   *
#  **********
#
        .align 3,0
N_CO:
        .dword      2
        .ascii      "CO"
        .align 3,0
CO:
        .dword    CO+HEADSIZE
        .dword    CO+HEADSIZE
        .dword    0x0
        .dword    EXIT
        .dword    N_CO
        .dword    0

        LD     a1,0(fp)
        SD     a7, 0(fp)
        MV     a7, a1
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   (;)   *
#  ***********
#
        .align 3,0
N_SEMIS:
        .dword      3
        .ascii      "(;)"
        .align 3,0
SEMIS:
        .dword    (EXIT+HEADSIZE)
        .dword    SEMIS+HEADSIZE
        .dword    0x0
        .dword    CO
        .dword    N_SEMIS
        .dword    0

#

#  *************
#  *   LEAVE   *
#  *************
#
        .align 3,0
N_LLEAV:
        .dword      5
        .ascii      "LEAVE"
        .align 3,0
LLEAV:
        .dword    DOCOL
        .dword    LLEAV+HEADSIZE
        .dword    0x0
        .dword    SEMIS
        .dword    N_LLEAV
        .dword    0
  #LIMIT <- INDEX
        .dword      UNLOOP
        .dword      SEMIS
#

#  **********
#  *   >R   *
#  **********
#
        .align 3,0
N_TOR:
        .dword      2
        .ascii      ">R"
        .align 3,0
TOR:
        .dword    TOR+HEADSIZE
        .dword    TOR+HEADSIZE
        .dword    0x0
        .dword    LLEAV
        .dword    N_TOR
        .dword    0
        # (RP1) <- (S1)
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #GET STACK PARAMETER
        ADDI    fp, fp, -(CW*(1))
        SD      a3, 0(fp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   R>   *
#  **********
#
        .align 3,0
N_FROMR:
        .dword      2
        .ascii      "R>"
        .align 3,0
FROMR:
        .dword    FROMR+HEADSIZE
        .dword    FROMR+HEADSIZE
        .dword    0x0
        .dword    TOR
        .dword    N_FROMR
        .dword    0
      #(S1) <- (RP1)
        LD      a0, 0(fp)
        ADDI    fp, fp, (CW*(1)) # GET RETURN STACK VALUE
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   RDROP   *
#  *************
#
        .align 3,0
N_RDROP:
        .dword      5
        .ascii      "RDROP"
        .align 3,0
RDROP:
        .dword    RDROP+HEADSIZE
        .dword    RDROP+HEADSIZE
        .dword    0x0
        .dword    FROMR
        .dword    N_RDROP
        .dword    0
      #(S1) <- (RP1)
        LD      a0, 0(fp)
        ADDI    fp, fp, (CW*(1)) # DISCARD RETURN STACK VALUE
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   R@   *
#  **********
#
        .align 3,0
N_RR:
        .dword      2
        .ascii      "R@"
        .align 3,0
RR:
        .dword    (IDO+HEADSIZE)
        .dword    RR+HEADSIZE
        .dword    0x0
        .dword    RDROP
        .dword    N_RR
        .dword    0

#

#  **********
#  *   0=   *
#  **********
#
        .align 3,0
N_ZEQU:
        .dword      2
        .ascii      "0="
        .align 3,0
ZEQU:
        .dword    ZEQU+HEADSIZE
        .dword    ZEQU+HEADSIZE
        .dword    0x0
        .dword    RR
        .dword    N_ZEQU
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        SEQZ a0, a0
        NEG  a0, a0
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   0<   *
#  **********
#
        .align 3,0
N_ZLESS:
        .dword      2
        .ascii      "0<"
        .align 3,0
ZLESS:
        .dword    ZLESS+HEADSIZE
        .dword    ZLESS+HEADSIZE
        .dword    0x0
        .dword    ZEQU
        .dword    N_ZLESS
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LI      t0, 0
        SLT     a0, a0, t0
        NEG     a0, a0
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *********
#  *   +   *
#  *********
#
        .align 3,0
N_PLUS:
        .dword      1
        .ascii      "+"
        .align 3,0
PLUS:
        .dword    PLUS+HEADSIZE
        .dword    PLUS+HEADSIZE
        .dword    0x0
        .dword    ZLESS
        .dword    N_PLUS
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #(S1) <- (S1) + (S2)
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADD     a0, a0,a3
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   D+   *
#  **********
#
        .align 3,0
N_DPLUS:
        .dword      2
        .ascii      "D+"
        .align 3,0
DPLUS:
        .dword    DPLUS+HEADSIZE
        .dword    DPLUS+HEADSIZE
        .dword    0x0
        .dword    PLUS
        .dword    N_DPLUS
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      # YHC
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      # YLC
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      # XHC
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))      # XLC
        ADD    a1, a1,a2
        SLTU   a4, a1,a2  # R4 is carry
        ADD    a0, a0,a3
        ADD    a0, a0,a4
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **************
#  *   NEGATE   *
#  **************
#
        .align 3,0
N_NEGATE:
        .dword      6
        .ascii      "NEGATE"
        .align 3,0
NEGATE:
        .dword    NEGATE+HEADSIZE
        .dword    NEGATE+HEADSIZE
        .dword    0x0
        .dword    DPLUS
        .dword    N_NEGATE
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        NEG a0, a0
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***************
#  *   DNEGATE   *
#  ***************
#
        .align 3,0
N_DNEGA:
        .dword      7
        .ascii      "DNEGATE"
        .align 3,0
DNEGA:
        .dword    DNEGA+HEADSIZE
        .dword    DNEGA+HEADSIZE
        .dword    0x0
        .dword    NEGATE
        .dword    N_DNEGA
        .dword    0

         LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))   # MSW
         LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))   # LSW

         NEG a1, a1
         SLTU a3, zero, a1   # R3 is borrow
         NEG a0, a0
         SUB a0,a0,a3

        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#
        #
#

#  ************
#  *   OVER   *
#  ************
#
        .align 3,0
N_OVER:
        .dword      4
        .ascii      "OVER"
        .align 3,0
OVER:
        .dword    OVER+HEADSIZE
        .dword    OVER+HEADSIZE
        .dword    0x0
        .dword    DNEGA
        .dword    N_OVER
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   DROP   *
#  ************
#
        .align 3,0
N_DROP:
        .dword      4
        .ascii      "DROP"
        .align 3,0
DROP:
        .dword    DROP+HEADSIZE
        .dword    DROP+HEADSIZE
        .dword    0x0
        .dword    OVER
        .dword    N_DROP
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   NIP   *
#  ***********
#
        .align 3,0
N_NIP:
        .dword      3
        .ascii      "NIP"
        .align 3,0
NIP:
        .dword    NIP+HEADSIZE
        .dword    NIP+HEADSIZE
        .dword    0x0
        .dword    DROP
        .dword    N_NIP
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   2DROP   *
#  *************
#
        .align 3,0
N_TDROP:
        .dword      5
        .ascii      "2DROP"
        .align 3,0
TDROP:
        .dword    TDROP+HEADSIZE
        .dword    TDROP+HEADSIZE
        .dword    0x0
        .dword    NIP
        .dword    N_TDROP
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   SWAP   *
#  ************
#
        .align 3,0
N_SWAP:
        .dword      4
        .ascii      "SWAP"
        .align 3,0
SWAP:
        .dword    SWAP+HEADSIZE
        .dword    SWAP+HEADSIZE
        .dword    0x0
        .dword    TDROP
        .dword    N_SWAP
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   DUP   *
#  ***********
#
        .align 3,0
N_LDUP:
        .dword      3
        .ascii      "DUP"
        .align 3,0
LDUP:
        .dword    LDUP+HEADSIZE
        .dword    LDUP+HEADSIZE
        .dword    0x0
        .dword    SWAP
        .dword    N_LDUP
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   2DUP   *
#  ************
#
        .align 3,0
N_TDUP:
        .dword      4
        .ascii      "2DUP"
        .align 3,0
TDUP:
        .dword    TDUP+HEADSIZE
        .dword    TDUP+HEADSIZE
        .dword    0x0
        .dword    LDUP
        .dword    N_TDUP
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   2SWAP   *
#  *************
#
        .align 3,0
N_TSWAP:
        .dword      5
        .ascii      "2SWAP"
        .align 3,0
TSWAP:
        .dword    TSWAP+HEADSIZE
        .dword    TSWAP+HEADSIZE
        .dword    0x0
        .dword    TDUP
        .dword    N_TSWAP
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADDI    sp, sp, -(CW*(1))
        SD      a2, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a3, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   2OVER   *
#  *************
#
        .align 3,0
N_TOVER:
        .dword      5
        .ascii      "2OVER"
        .align 3,0
TOVER:
        .dword    TOVER+HEADSIZE
        .dword    TOVER+HEADSIZE
        .dword    0x0
        .dword    TSWAP
        .dword    N_TOVER
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a2, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a3, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   +!   *
#  **********
#
        .align 3,0
N_PSTOR:
        .dword      2
        .ascii      "+!"
        .align 3,0
PSTOR:
        .dword    PSTOR+HEADSIZE
        .dword    PSTOR+HEADSIZE
        .dword    0x0
        .dword    TOVER
        .dword    N_PSTOR
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ADDRESS
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #INCREMENT
        LD     a2, 0(a3)
        ADD     a2, a2, a0
        SD     a2, 0(a3)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **************
#  *   TOGGLE   *
#  **************
#
        .align 3,0
N_TOGGL:
        .dword      6
        .ascii      "TOGGLE"
        .align 3,0
TOGGL:
        .dword    TOGGL+HEADSIZE
        .dword    TOGGL+HEADSIZE
        .dword    0x0
        .dword    PSTOR
        .dword    N_TOGGL
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #BIT PATTERN
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ADDR
        LD     a1, 0(a3)
        XOR     a1, a1, a0
        SD     a1, 0(a3)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *********
#  *   @   *
#  *********
#
        .align 3,0
N_FETCH:
        .dword      1
        .ascii      "@"
        .align 3,0
FETCH:
        .dword    FETCH+HEADSIZE
        .dword    FETCH+HEADSIZE
        .dword    0x0
        .dword    TOGGL
        .dword    N_FETCH
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD     a0,0(a3)
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   P@   *
#  **********
#
        .align 3,0
N_PFETCH:
        .dword      2
        .ascii      "P@"
        .align 3,0
PFETCH:
        .dword    PFETCH+HEADSIZE
        .dword    PFETCH+HEADSIZE
        .dword    0x0
        .dword    FETCH
        .dword    N_PFETCH
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD     a0,0(a3)
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#


#  **********
#  *   L@   *
#  **********
#
        .align 3,0
N_LFETCH:
        .dword      2
        .ascii      "L@"
        .align 3,0
LFETCH:
        .dword    LFETCH+HEADSIZE
        .dword    LFETCH+HEADSIZE
        .dword    0x0
        .dword    PFETCH
        .dword    N_LFETCH
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LW      a0,0(a3)
        LI      t0, 0xFFFFFFFF
        AND     a0, a0, t0
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#
#
#


#  ***********
#  *   BIC   *
#  ***********
#
        .align 3,0
N_BITCLR:
        .dword      3
        .ascii      "BIC"
        .align 3,0
BITCLR:
        .dword    BITCLR+HEADSIZE
        .dword    BITCLR+HEADSIZE
        .dword    0x0
        .dword    LFETCH
        .dword    N_BITCLR
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LW     a2,0(a3)
        NOT    a0,a0
        AND    a2,a2,a0
        SW     a2,0(a3)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   BIS   *
#  ***********
#
        .align 3,0
N_BITSET:
        .dword      3
        .ascii      "BIS"
        .align 3,0
BITSET:
        .dword    BITSET+HEADSIZE
        .dword    BITSET+HEADSIZE
        .dword    0x0
        .dword    BITCLR
        .dword    N_BITSET
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LW     a2,0(a3)
        OR      a2, a2,a0
        SW     a2,0(a3)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#
#

#  **********
#  *   C@   *
#  **********
#
        .align 3,0
N_CFET:
        .dword      2
        .ascii      "C@"
        .align 3,0
CFET:
        .dword    CFET+HEADSIZE
        .dword    CFET+HEADSIZE
        .dword    0x0
        .dword    BITSET
        .dword    N_CFET
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LB    a0,0(a3)
        AND   a0, a0, 0xFF
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   2@   *
#  **********
#
        .align 3,0
N_TFET:
        .dword      2
        .ascii      "2@"
        .align 3,0
TFET:
        .dword    TFET+HEADSIZE
        .dword    TFET+HEADSIZE
        .dword    0x0
        .dword    CFET
        .dword    N_TFET
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ADDR
        LD     a0,0(a3)                      #MSW
        LD     a1,(CW*(1))(a3)               #LSW
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *********
#  *   !   *
#  *********
#
        .align 3,0
N_STOR:
        .dword      1
        .ascii      "!"
        .align 3,0
STOR:
        .dword    STOR+HEADSIZE
        .dword    STOR+HEADSIZE
        .dword    0x0
        .dword    TFET
        .dword    N_STOR
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ADDR
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #DATA
        SD     a0, 0(a3)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   P!   *
#  **********
#
        .align 3,0
N_PRTSTR:
        .dword      2
        .ascii      "P!"
        .align 3,0
PRTSTR:
        .dword    PRTSTR+HEADSIZE
        .dword    PRTSTR+HEADSIZE
        .dword    0x0
        .dword    STOR
        .dword    N_PRTSTR
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ADDR
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #DATA
        SD     a0, 0(a3)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#


#  **********
#  *   L!   *
#  **********
#
        .align 3,0
N_LSTOR:
        .dword      2
        .ascii      "L!"
        .align 3,0
LSTOR:
        .dword    LSTOR+HEADSIZE
        .dword    LSTOR+HEADSIZE
        .dword    0x0
        .dword    PRTSTR
        .dword    N_LSTOR
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ADDR
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #DATA
        SW     a0, 0(a3)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#
#

#  **********
#  *   C!   *
#  **********
#
        .align 3,0
N_CSTOR:
        .dword      2
        .ascii      "C!"
        .align 3,0
CSTOR:
        .dword    CSTOR+HEADSIZE
        .dword    CSTOR+HEADSIZE
        .dword    0x0
        .dword    LSTOR
        .dword    N_CSTOR
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ADDR
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #DATA
        SB            a0, 0(a3)  #Store byte
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   2!   *
#  **********
#
        .align 3,0
N_TSTOR:
        .dword      2
        .ascii      "2!"
        .align 3,0
TSTOR:
        .dword    TSTOR+HEADSIZE
        .dword    TSTOR+HEADSIZE
        .dword    0x0
        .dword    CSTOR
        .dword    N_TSTOR
        .dword    0

        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #ADDR
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #MSW
        SD     a0, 0(a3)
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #LSW
        SD     a0, CW(a3)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **************
#  *   WITHIN   *
#  **************
#
        .align 3,0
N_WITHIN:
        .dword      6
        .ascii      "WITHIN"
        .align 3,0
WITHIN:
        .dword    DOCOL
        .dword    WITHIN+HEADSIZE
        .dword    0x0
        .dword    TSTOR
        .dword    N_WITHIN
        .dword    0

        .dword      OVER, LSUB, TOR
        .dword      LSUB, FROMR
        .dword      ULESS
        .dword      SEMIS
#
#

#  *********
#  *   :   *
#  *********
#
        .align 3,0
N_COLON:
        .dword      1
        .ascii      ":"
        .align 3,0
COLON:
        .dword    DOCOL
        .dword    COLON+HEADSIZE
        .dword    0x0
        .dword    WITHIN
        .dword    N_COLON
        .dword    0

        .dword      SCSP
        .dword      LNAME
        .dword      PCREAT
        .dword      LATEST, HIDDEN
        .dword      RBRAC
        .dword      PSCOD
DOCOL:  ADDI    fp, fp, -(CW*(1))
        SD      a7, 0(fp) 
        LD     a7, (CW*(D_HOFFSET))(a0)   #(IP) <- (PFA)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  


#

#  *********
#  *   ;   *
#  *********
#
        .align 3,0
N_SEMI:
        .dword      1
        .ascii      ";"
        .align 3,0
SEMI:
        .dword    DOCOL
        .dword    SEMI+HEADSIZE
        .dword    B_IMMED
        .dword    COLON
        .dword    N_SEMI
        .dword    0

        .dword      QCSP
        .dword      LIT, SEMIS, COMMA
        .dword      LATEST, HIDDEN
        .dword      LBRAC
        .dword      SEMIS
#

#  ****************
#  *   CONSTANT   *
#  ****************
#
        .align 3,0
N_LCONST:
        .dword      8
        .ascii      "CONSTANT"
        .align 3,0
LCONST:
        .dword    DOCOL
        .dword    LCONST+HEADSIZE
        .dword    0x0
        .dword    SEMI
        .dword    N_LCONST
        .dword    0

        .dword      LNAME
        .dword      PCREAT
        .dword      LATEST, TDFA, STOR
        .dword      PSCOD
DOCON:  LD     a0, (CW*(D_HOFFSET))(a0)      #GET DATA FROM PFA
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ****************
#  *   VARIABLE   *
#  ****************
#
        .align 3,0
N_VAR:
        .dword      8
        .ascii      "VARIABLE"
        .align 3,0
VAR:
        .dword    DOCOL
        .dword    VAR+HEADSIZE
        .dword    0x0
        .dword    LCONST
        .dword    N_VAR
        .dword    0

        .dword      LDATA
        .dword      ZERO, COMMA
        .dword      SEMIS
#

#  ************
#  *   DATA   *
#  ************
#
        .align 3,0
N_LDATA:
        .dword      4
        .ascii      "DATA"
        .align 3,0
LDATA:
        .dword    DOCOL
        .dword    LDATA+HEADSIZE
        .dword    0x0
        .dword    VAR
        .dword    N_LDATA
        .dword    0

        .dword      LNAME
        .dword      PCREAT
        .dword      PSCOD
DOVAR:  LD     a0, (CW*(D_HOFFSET))(a0)        #(R0) <- PFA
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *********
#  *   _   *
#  *********
#
        .align 3,0
N_X:
        .dword      1
        .ascii      "_"
        .align 3,0
X:
        .dword    X+HEADSIZE
        .dword    X+HEADSIZE
        .dword    0x0
        .dword    LDATA
        .dword    N_X
        .dword    0

        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0   #Whatever happens to be in a0, i.e. the dea of ``_''.
#

#  *********
#  *   0   *
#  *********
#
        .align 3,0
N_ZERO:
        .dword      1
        .ascii      "0"
        .align 3,0
ZERO:
        .dword    DOCON
        .dword    0
        .dword    0x0
        .dword    X
        .dword    N_ZERO
        .dword    0

# The data field of 0 is used for an empty string.
#

#  *********
#  *   1   *
#  *********
#
        .align 3,0
N_ONE:
        .dword      1
        .ascii      "1"
        .align 3,0
ONE:
        .dword    DOCON
        .dword    1
        .dword    0x0
        .dword    ZERO
        .dword    N_ONE
        .dword    0

#

#  *********
#  *   2   *
#  *********
#
        .align 3,0
N_TWO:
        .dword      1
        .ascii      "2"
        .align 3,0
TWO:
        .dword    DOCON
        .dword    2
        .dword    0x0
        .dword    ONE
        .dword    N_TWO
        .dword    0

#

#  **********
#  *   BL   *
#  **********
#
        .align 3,0
N_LBL:
        .dword      2
        .ascii      "BL"
        .align 3,0
LBL:
        .dword    DOCON
        .dword    ABL
        .dword    0x0
        .dword    TWO
        .dword    N_LBL
        .dword    0

#

#
#  **********
#  *   $@   *
#  **********
#
        .align 3,0
N_SFET:
        .dword      2
        .ascii      "\$@"
        .align 3,0
SFET:
        .dword    SFET+HEADSIZE
        .dword    SFET+HEADSIZE
        .dword    0x0
        .dword    LBL
        .dword    N_SFET
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(a1)
        ADDI    a1, a1, (CW*(1))
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   $!   *
#  **********
#
        .align 3,0
N_SSTOR:
        .dword      2
        .ascii      "\$!"
        .align 3,0
SSTOR:
        .dword    DOCOL
        .dword    SSTOR+HEADSIZE
        .dword    0x0
        .dword    SFET
        .dword    N_SSTOR
        .dword    0

        .dword TDUP, STOR, CELLP, SWAP, LMOVE
        .dword SEMIS
#

# 

#  ***********
#  *   $+!   *
#  ***********
#
        .align 3,0
N_SADD:
        .dword      3
        .ascii      "\$+!"
        .align 3,0
SADD:
        .dword    DOCOL
        .dword    SADD+HEADSIZE
        .dword    0x0
        .dword    SSTOR
        .dword    N_SADD
        .dword    0

        .dword   LDUP, FETCH, TOR # Remember old count.
        .dword   TDUP, PSTOR
        .dword   CELLP, FROMR, PLUS, SWAP, LMOVE
        .dword SEMIS
#

#  ***********
#  *   $C+   *
#  ***********
#
        .align 3,0
N_CHAPP:
        .dword      3
        .ascii      "\$C+"
        .align 3,0
CHAPP:
        .dword    DOCOL
        .dword    CHAPP+HEADSIZE
        .dword    0x0
        .dword    SADD
        .dword    N_CHAPP
        .dword    0

        .dword   LDUP, TOR
        .dword   LDUP, FETCH, PLUS, CELLP, CSTOR
        .dword   ONE, FROMR, PSTOR
        .dword SEMIS
#

#  **********
#  *   $,   *
#  **********
#
        .align 3,0
N_SCOMMA:
        .dword      2
        .ascii      "\$,"
        .align 3,0
SCOMMA:
        .dword    DOCOL
        .dword    SCOMMA+HEADSIZE
        .dword    0x0
        .dword    CHAPP
        .dword    N_SCOMMA
        .dword    0

        .dword HERE, TOR
        .dword LDUP, CELLP, ALLOT
        .dword RR, SSTOR, FROMR
 .dword LALIGN
        .dword SEMIS
#

#  ***********
#  *   C/L   *
#  ***********
#
        .align 3,0
N_CSLL:
        .dword      3
        .ascii      "C/L"
        .align 3,0
CSLL:
        .dword    DOCON
        .dword    64
        .dword    0x0
        .dword    SCOMMA
        .dword    N_CSLL
        .dword    0

#


#  **************
#  *   _FIRST   *
#  **************
#
        .align 3,0
N__FIRST:
        .dword      6
        .ascii      "_FIRST"
        .align 3,0
_FIRST:
        .dword    DOCON
        .dword    BUF1
        .dword    0x0
        .dword    CSLL
        .dword    N__FIRST
        .dword    0

 #  
#
# 
#

#  **************
#  *   _LIMIT   *
#  **************
#
        .align 3,0
N__LIMIT:
        .dword      6
        .ascii      "_LIMIT"
        .align 3,0
_LIMIT:
        .dword    DOCON
        .dword    BUF1+(KBBUF+2*CW)*NBUF
        .dword    0x0
        .dword    _FIRST
        .dword    N__LIMIT
        .dword    0

# THE END  OF THE MEMORY 

#  **********
#  *   EM   *
#  **********
#
        .align 3,0
N_LEM:
        .dword      2
        .ascii      "EM"
        .align 3,0
LEM:
        .dword    DOCON
        .dword    ACTUAL_EM
        .dword    0x0
        .dword    _LIMIT
        .dword    N_LEM
        .dword    0

#

#  **********
#  *   BM   *
#  **********
#
        .align 3,0
N_LBM:
        .dword      2
        .ascii      "BM"
        .align 3,0
LBM:
        .dword    DOCON
        .dword    _ORIG
        .dword    0x0
        .dword    LEM
        .dword    N_LBM
        .dword    0

#

#  *************
#  *   B/BUF   *
#  *************
#
        .align 3,0
N_BBUF:
        .dword      5
        .ascii      "B/BUF"
        .align 3,0
BBUF:
        .dword    DOCON
        .dword    KBBUF
        .dword    0x0
        .dword    LBM
        .dword    N_BBUF
        .dword    0

#
# All user variables are initialised 
# with the values from USINI.
# The implementation relies on the initialisation of 
# those with numbers (1..11), so change in concord with USINI.

#  **********
#  *   U0   *
#  **********
#
        .align 3,0
N_UZERO:
        .dword      2
        .ascii      "U0"
        .align 3,0
UZERO:
        .dword    DOUSE
        .dword    (CW*(1))
        .dword    0x0
        .dword    BBUF
        .dword    N_UZERO
        .dword    0

#

#  **********
#  *   S0   *
#  **********
#
        .align 3,0
N_SZERO:
        .dword      2
        .ascii      "S0"
        .align 3,0
SZERO:
        .dword    DOUSE
        .dword    (CW*(2))
        .dword    0x0
        .dword    UZERO
        .dword    N_SZERO
        .dword    0

#

#  **********
#  *   R0   *
#  **********
#
        .align 3,0
N_RZERO:
        .dword      2
        .ascii      "R0"
        .align 3,0
RZERO:
        .dword    DOUSE
        .dword    (CW*(3))
        .dword    0x0
        .dword    SZERO
        .dword    N_RZERO
        .dword    0

#

#  ***********
#  *   TIB   *
#  ***********
#
        .align 3,0
N_TIB:
        .dword      3
        .ascii      "TIB"
        .align 3,0
TIB:
        .dword    DOUSE
        .dword    (CW*(4))
        .dword    0x0
        .dword    RZERO
        .dword    N_TIB
        .dword    0

#

#  **************
#  *   RUBOUT   *
#  **************
#
        .align 3,0
N_RUBOUT:
        .dword      6
        .ascii      "RUBOUT"
        .align 3,0
RUBOUT:
        .dword    DOUSE
        .dword    (CW*(5))
        .dword    0x0
        .dword    TIB
        .dword    N_RUBOUT
        .dword    0

#

#  ***************
#  *   WARNING   *
#  ***************
#
        .align 3,0
N_LWARN:
        .dword      7
        .ascii      "WARNING"
        .align 3,0
LWARN:
        .dword    DOUSE
        .dword    (CW*(7))
        .dword    0x0
        .dword    RUBOUT
        .dword    N_LWARN
        .dword    0

#

#  *************
#  *   FENCE   *
#  *************
#
        .align 3,0
N_FENCE:
        .dword      5
        .ascii      "FENCE"
        .align 3,0
FENCE:
        .dword    DOUSE
        .dword    (CW*(8))
        .dword    0x0
        .dword    LWARN
        .dword    N_FENCE
        .dword    0

#

#  **********
#  *   DP   *
#  **********
#
        .align 3,0
N_LDP:
        .dword      2
        .ascii      "DP"
        .align 3,0
LDP:
        .dword    DOUSE
        .dword    (CW*(9))
        .dword    0x0
        .dword    FENCE
        .dword    N_LDP
        .dword    0

#

#  ****************
#  *   VOC-LINK   *
#  ****************
#
        .align 3,0
N_VOCL:
        .dword      8
        .ascii      "VOC-LINK"
        .align 3,0
VOCL:
        .dword    DOUSE
        .dword    (CW*(10))
        .dword    0x0
        .dword    LDP
        .dword    N_VOCL
        .dword    0

#

#  **************
#  *   OFFSET   *
#  **************
#
        .align 3,0
N_LOFFSET:
        .dword      6
        .ascii      "OFFSET"
        .align 3,0
LOFFSET:
        .dword    DOUSE
        .dword    (CW*(37))
        .dword    0x0
        .dword    VOCL
        .dword    N_LOFFSET
        .dword    0

#

#  ****************
#  *   MAX-USER   *
#  ****************
#
        .align 3,0
N_MAXUSER:
        .dword      8
        .ascii      "MAX-USER"
        .align 3,0
MAXUSER:
        .dword    DOUSE
        .dword    (CW*(11))
        .dword    0x0
        .dword    LOFFSET
        .dword    N_MAXUSER
        .dword    0

#
# End of user variables with fixed place.
#
#

#  *************
#  *   WHERE   *
#  *************
#
        .align 3,0
N_LWHERE:
        .dword      5
        .ascii      "WHERE"
        .align 3,0
LWHERE:
        .dword    DOUSE
        .dword    (CW*(12))
        .dword    0x0
        .dword    MAXUSER
        .dword    N_LWHERE
        .dword    0
    #  Occupies two CELLS! 
#

#  ***********
#  *   SCR   *
#  ***********
#
        .align 3,0
N_SCR:
        .dword      3
        .ascii      "SCR"
        .align 3,0
SCR:
        .dword    DOUSE
        .dword    (CW*(33))
        .dword    0x0
        .dword    LWHERE
        .dword    N_SCR
        .dword    0

#

#  *************
#  *   STATE   *
#  *************
#
        .align 3,0
N_STATE:
        .dword      5
        .ascii      "STATE"
        .align 3,0
STATE:
        .dword    DOUSE
        .dword    (CW*(18))
        .dword    0x0
        .dword    SCR
        .dword    N_STATE
        .dword    0

#

#  ************
#  *   BASE   *
#  ************
#
        .align 3,0
N_BASE:
        .dword      4
        .ascii      "BASE"
        .align 3,0
BASE:
        .dword    DOUSE
        .dword    (CW*(19))
        .dword    0x0
        .dword    STATE
        .dword    N_BASE
        .dword    0

#

#  ***********
#  *   DPL   *
#  ***********
#
        .align 3,0
N_DPL:
        .dword      3
        .ascii      "DPL"
        .align 3,0
DPL:
        .dword    DOUSE
        .dword    (CW*(20))
        .dword    0x0
        .dword    BASE
        .dword    N_DPL
        .dword    0

#

#  ***********
#  *   ECL   *
#  ***********
#
        .align 3,0
N_ECL:
        .dword      3
        .ascii      "ECL"
        .align 3,0
ECL:
        .dword    DOUSE
        .dword    (CW*(38))
        .dword    0x0
        .dword    DPL
        .dword    N_ECL
        .dword    0

#

#  ***********
#  *   FLD   *
#  ***********
#
        .align 3,0
N_LFLD:
        .dword      3
        .ascii      "FLD"
        .align 3,0
LFLD:
        .dword    DOUSE
        .dword    (CW*(21))
        .dword    0x0
        .dword    ECL
        .dword    N_LFLD
        .dword    0

#


#  ***********
#  *   CSP   *
#  ***********
#
        .align 3,0
N_LCSP:
        .dword      3
        .ascii      "CSP"
        .align 3,0
LCSP:
        .dword    DOUSE
        .dword    (CW*(22))
        .dword    0x0
        .dword    LFLD
        .dword    N_LCSP
        .dword    0

#
#

#  **********
#  *   R#   *
#  **********
#
        .align 3,0
N_RNUM:
        .dword      2
        .ascii      "R#"
        .align 3,0
RNUM:
        .dword    DOUSE
        .dword    (CW*(23))
        .dword    0x0
        .dword    LCSP
        .dword    N_RNUM
        .dword    0

#

#  ***********
#  *   HLD   *
#  ***********
#
        .align 3,0
N_HLD:
        .dword      3
        .ascii      "HLD"
        .align 3,0
HLD:
        .dword    DOUSE
        .dword    (CW*(24))
        .dword    0x0
        .dword    RNUM
        .dword    N_HLD
        .dword    0

#

#  ***********
#  *   OUT   *
#  ***********
#
        .align 3,0
N_LOUT:
        .dword      3
        .ascii      "OUT"
        .align 3,0
LOUT:
        .dword    DOUSE
        .dword    (CW*(25))
        .dword    0x0
        .dword    HLD
        .dword    N_LOUT
        .dword    0

#

#  *************
#  *   (BLK)   *
#  *************
#
        .align 3,0
N_PBLK:
        .dword      5
        .ascii      "(BLK)"
        .align 3,0
PBLK:
        .dword    DOUSE
        .dword    (CW*(26))
        .dword    0x0
        .dword    LOUT
        .dword    N_PBLK
        .dword    0

#

#  **********
#  *   PP   *
#  **********
#
        .align 3,0
N_PP:
        .dword      2
        .ascii      "PP"
        .align 3,0
PP:
        .dword    DOUSE
        .dword    (CW*(29))
        .dword    0x0
        .dword    PBLK
        .dword    N_PP
        .dword    0

#

#  *************
#  *   (>IN)   *
#  *************
#
        .align 3,0
N_PIIN:
        .dword      5
        .ascii      "(>IN)"
        .align 3,0
PIIN:
        .dword    DOUSE
        .dword    (CW*(30))
        .dword    0x0
        .dword    PP
        .dword    N_PIIN
        .dword    0

#

#  ************
#  *   ARGS   *
#  ************
#
        .align 3,0
N_ARGS:
        .dword      4
        .ascii      "ARGS"
        .align 3,0
ARGS:
        .dword    DOUSE
        .dword    (CW*(31))
        .dword    0x0
        .dword    PIIN
        .dword    N_ARGS
        .dword    0

#

#  ***************
#  *   HANDLER   *
#  ***************
#
        .align 3,0
N_HANDLER:
        .dword      7
        .ascii      "HANDLER"
        .align 3,0
HANDLER:
        .dword    DOUSE
        .dword    (CW*(32))
        .dword    0x0
        .dword    ARGS
        .dword    N_HANDLER
        .dword    0

#

#  ***************
#  *   CURRENT   *
#  ***************
#
        .align 3,0
N_CURR:
        .dword      7
        .ascii      "CURRENT"
        .align 3,0
CURR:
        .dword    DOUSE
        .dword    (CW*(34))
        .dword    0x0
        .dword    HANDLER
        .dword    N_CURR
        .dword    0

#

#  *****************
#  *   REMAINDER   *
#  *****************
#
        .align 3,0
N_REMAIND:
        .dword      9
        .ascii      "REMAINDER"
        .align 3,0
REMAIND:
        .dword    DOUSE
        .dword    (CW*(14))
        .dword    0x0
        .dword    CURR
        .dword    N_REMAIND
        .dword    0

#      IMPORTANT
# REQUIRES ONE MORE CELL!
#

#  ***************
#  *   CONTEXT   *
#  ***************
#
        .align 3,0
N_CONTEXT:
        .dword      7
        .ascii      "CONTEXT"
        .align 3,0
CONTEXT:
        .dword    DOUSE
        .dword    (CW*(39))
        .dword    0x0
        .dword    REMAIND
        .dword    N_CONTEXT
        .dword    0
 # Up to  39+16
#      IMPORTANT
#     16 USER SPACE CELLS MUST BE KEPT FREE
#     IN ADDITION TO THE ONE FOR CONTEXT
#
#========== END USER VARIABLES =============
#

#  **********
#  *   1+   *
#  **********
#
        .align 3,0
N_ONEP:
        .dword      2
        .ascii      "1+"
        .align 3,0
ONEP:
        .dword    DOCOL
        .dword    ONEP+HEADSIZE
        .dword    0x0
        .dword    CONTEXT
        .dword    N_ONEP
        .dword    0

        .dword      ONE
        .dword      PLUS
        .dword      SEMIS
#

#  *************
#  *   CELL+   *
#  *************
#
        .align 3,0
N_CELLP:
        .dword      5
        .ascii      "CELL+"
        .align 3,0
CELLP:
        .dword    DOCOL
        .dword    CELLP+HEADSIZE
        .dword    0x0
        .dword    ONEP
        .dword    N_CELLP
        .dword    0

        .dword      LIT, CW
        .dword      PLUS
        .dword      SEMIS
#

#
#  *************
#  *   CELLS   *
#  *************
#
        .align 3,0
N_LCELLS:
        .dword      5
        .ascii      "CELLS"
        .align 3,0
LCELLS:
        .dword    DOCOL
        .dword    LCELLS+HEADSIZE
        .dword    0x0
        .dword    CELLP
        .dword    N_LCELLS
        .dword    0

        .dword        LIT, 3
        .dword      LSHIFT
        .dword      SEMIS
#

#  *************
#  *   CHAR+   *
#  *************
#
        .align 3,0
N_CHARP:
        .dword      5
        .ascii      "CHAR+"
        .align 3,0
CHARP:
        .dword    DOCOL
        .dword    (ONEP+HEADSIZE)
        .dword    0x0
        .dword    LCELLS
        .dword    N_CHARP
        .dword    0

#

#  *************
#  *   CHARS   *
#  *************
#
        .align 3,0
N_CHARS:
        .dword      5
        .ascii      "CHARS"
        .align 3,0
CHARS:
        .dword    DOCOL 
        .dword    CHARS+HEADSIZE
        .dword    0x0
        .dword    CHARP
        .dword    N_CHARS
        .dword    0

        .dword      SEMIS
#

#  *************
#  *   ALIGN   *
#  *************
#
        .align 3,0
N_LALIGN:
        .dword      5
        .ascii      "ALIGN"
        .align 3,0
LALIGN:
        .dword    DOCOL 
        .dword    LALIGN+HEADSIZE
        .dword    0x0
        .dword    CHARS
        .dword    N_LALIGN
        .dword    0


        .dword      LDP, FETCH
        .dword      ALIGNED
        .dword      LDP, STOR
#
        .dword      SEMIS
#

#  ***************
#  *   ALIGNED   *
#  ***************
#
        .align 3,0
N_ALIGNED :
        .dword      7
        .ascii      "ALIGNED"
        .align 3,0
ALIGNED :
        .dword    ALIGNED +HEADSIZE
        .dword    ALIGNED +HEADSIZE
        .dword    0x0
        .dword    LALIGN
        .dword    N_ALIGNED 
        .dword    0


        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        ADDI     a0, a0, -1
        ORI     a0, a0, (CW-1)
        ADDI     a0, a0, 1
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
#
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   HERE   *
#  ************
#
        .align 3,0
N_HERE:
        .dword      4
        .ascii      "HERE"
        .align 3,0
HERE:
        .dword    DOCOL
        .dword    HERE+HEADSIZE
        .dword    0x0
        .dword    ALIGNED 
        .dword    N_HERE
        .dword    0

        .dword      LDP
        .dword      FETCH
        .dword      SEMIS
#

#  *************
#  *   ALLOT   *
#  *************
#
        .align 3,0
N_ALLOT:
        .dword      5
        .ascii      "ALLOT"
        .align 3,0
ALLOT:
        .dword    DOCOL
        .dword    ALLOT+HEADSIZE
        .dword    0x0
        .dword    HERE
        .dword    N_ALLOT
        .dword    0

        .dword      LDP
        .dword      PSTOR
        .dword      SEMIS
#

#  *********
#  *   ,   *
#  *********
#
        .align 3,0
N_COMMA:
        .dword      1
        .ascii      ","
        .align 3,0
COMMA:
        .dword    DOCOL
        .dword    COMMA+HEADSIZE
        .dword    0x0
        .dword    ALLOT
        .dword    N_COMMA
        .dword    0

        .dword      HERE
        .dword      LIT
        .dword      CW
        .dword      ALLOT
        .dword      STOR
        .dword      SEMIS
#

#  **********
#  *   C,   *
#  **********
#
        .align 3,0
N_CCOMM:
        .dword      2
        .ascii      "C,"
        .align 3,0
CCOMM:
        .dword    DOCOL
        .dword    CCOMM+HEADSIZE
        .dword    0x0
        .dword    COMMA
        .dword    N_CCOMM
        .dword    0

        .dword      HERE
        .dword      ONE
        .dword      ALLOT
        .dword      CSTOR
        .dword      SEMIS
#

#  **********
#  *   2,   *
#  **********
#
        .align 3,0
N_TCOMM:
        .dword      2
        .ascii      "2,"
        .align 3,0
TCOMM:
        .dword    DOCOL
        .dword    TCOMM+HEADSIZE
        .dword    0x0
        .dword    CCOMM
        .dword    N_TCOMM
        .dword    0

        .dword      HERE
        .dword      LIT
        .dword      CW*2
        .dword      ALLOT
        .dword      TSTOR
        .dword      SEMIS
#

#  *********
#  *   -   *
#  *********
#
        .align 3,0
N_LSUB:
        .dword      1
        .ascii      "-"
        .align 3,0
LSUB:
        .dword    LSUB+HEADSIZE
        .dword    LSUB+HEADSIZE
        .dword    0x0
        .dword    TCOMM
        .dword    N_LSUB
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #S1
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        SUB     a0, a0,a1
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0     #S1 = S2 - S1
#

#  *********
#  *   =   *
#  *********
#
        .align 3,0
N_EQUAL:
        .dword      1
        .ascii      "="
        .align 3,0
EQUAL:
        .dword    DOCOL
        .dword    EQUAL+HEADSIZE
        .dword    0x0
        .dword    LSUB
        .dword    N_EQUAL
        .dword    0

        .dword      LSUB
        .dword      ZEQU
        .dword      SEMIS
#

#  *********
#  *   <   *
#  *********
#
        .align 3,0
N_LESS:
        .dword      1
        .ascii      "<"
        .align 3,0
LESS:
        .dword    LESS+HEADSIZE
        .dword    LESS+HEADSIZE
        .dword    0x0
        .dword    EQUAL
        .dword    N_LESS
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #S1
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #S2
        SLT a0, a1, a0   # set R0 to 1 if R1 < R0, otherwise set R0 to 0
        NEG     a0, a0
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   U<   *
#  **********
#
        .align 3,0
N_ULESS:
        .dword      2
        .ascii      "U<"
        .align 3,0
ULESS:
        .dword    ULESS+HEADSIZE
        .dword    ULESS+HEADSIZE
        .dword    0x0
        .dword    LESS
        .dword    N_ULESS
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))
        SLTU    a0, a1, a0  # set R0 to 1 if R1 <U R0, otherwise set R0 to 0
        NEG     a0, a0
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *********
#  *   >   *
#  *********
#
        .align 3,0
N_GREAT:
        .dword      1
        .ascii      ">"
        .align 3,0
GREAT:
        .dword    DOCOL
        .dword    GREAT+HEADSIZE
        .dword    0x0
        .dword    ULESS
        .dword    N_GREAT
        .dword    0

        .dword      SWAP
        .dword      LESS
        .dword      SEMIS
#

#  **********
#  *   <>   *
#  **********
#
        .align 3,0
N_UNEQ:
        .dword      2
        .ascii      "<>"
        .align 3,0
UNEQ:
        .dword    DOCOL
        .dword    UNEQ+HEADSIZE
        .dword    0x0
        .dword    GREAT
        .dword    N_UNEQ
        .dword    0

        .dword      LSUB
        .dword      ZEQU
        .dword      ZEQU
        .dword      SEMIS
#

#  ***********
#  *   ROT   *
#  ***********
#
        .align 3,0
N_ROT:
        .dword      3
        .ascii      "ROT"
        .align 3,0
ROT:
        .dword    ROT+HEADSIZE
        .dword    ROT+HEADSIZE
        .dword    0x0
        .dword    UNEQ
        .dword    N_ROT
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #S1
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      #S2
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #S3
        ADDI    sp, sp, -(CW*(1))
        SD      a3, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   SPACE   *
#  *************
#
        .align 3,0
N_SPACE:
        .dword      5
        .ascii      "SPACE"
        .align 3,0
SPACE:
        .dword    DOCOL
        .dword    SPACE+HEADSIZE
        .dword    0x0
        .dword    ROT
        .dword    N_SPACE
        .dword    0

        .dword      LBL
        .dword      EMIT
        .dword      SEMIS
#

#  ************
#  *   ?DUP   *
#  ************
#
        .align 3,0
N_QDUP:
        .dword      4
        .ascii      "?DUP"
        .align 3,0
QDUP:
        .dword    DOCOL
        .dword    QDUP+HEADSIZE
        .dword    0x0
        .dword    SPACE
        .dword    N_QDUP
        .dword    0

        .dword      LDUP
        .dword      ZBRAN
        .dword      QDUP1-.-CW # IF
        .dword      LDUP    #THEN
QDUP1:  .dword      SEMIS
#

#  **************
#  *   LATEST   *
#  **************
#
        .align 3,0
N_LATEST:
        .dword      6
        .ascii      "LATEST"
        .align 3,0
LATEST:
        .dword    DOCOL
        .dword    LATEST+HEADSIZE
        .dword    0x0
        .dword    QDUP
        .dword    N_LATEST
        .dword    0

        .dword      CURR
        .dword      FETCH
        .dword      TLFA
        .dword      FETCH
        .dword      SEMIS
#

#  ************
#  *   >CFA   *
#  ************
#
        .align 3,0
N_TCFA:
        .dword      4
        .ascii      ">CFA"
        .align 3,0
TCFA:
        .dword    DOCOL
        .dword    TCFA+HEADSIZE
        .dword    0x0
        .dword    LATEST
        .dword    N_TCFA
        .dword    0

        .dword      LIT, (CW*(C_HOFFSET))
        .dword      PLUS
        .dword      SEMIS
#

#  ************
#  *   >DFA   *
#  ************
#
        .align 3,0
N_TDFA:
        .dword      4
        .ascii      ">DFA"
        .align 3,0
TDFA:
        .dword    DOCOL
        .dword    TDFA+HEADSIZE
        .dword    0x0
        .dword    TCFA
        .dword    N_TDFA
        .dword    0

        .dword      LIT, (CW*(D_HOFFSET))
        .dword      PLUS
        .dword      SEMIS
#

#  ************
#  *   >FFA   *
#  ************
#
        .align 3,0
N_TFFA:
        .dword      4
        .ascii      ">FFA"
        .align 3,0
TFFA:
        .dword    DOCOL
        .dword    TFFA+HEADSIZE
        .dword    0x0
        .dword    TDFA
        .dword    N_TFFA
        .dword    0

        .dword      LIT, (CW*(F_HOFFSET))
        .dword      PLUS
        .dword      SEMIS
#

#  ************
#  *   >LFA   *
#  ************
#
        .align 3,0
N_TLFA:
        .dword      4
        .ascii      ">LFA"
        .align 3,0
TLFA:
        .dword    DOCOL
        .dword    TLFA+HEADSIZE
        .dword    0x0
        .dword    TFFA
        .dword    N_TLFA
        .dword    0

        .dword      LIT, (CW*(L_HOFFSET))
        .dword      PLUS
        .dword      SEMIS
#

#  ************
#  *   >NFA   *
#  ************
#
        .align 3,0
N_TNFA:
        .dword      4
        .ascii      ">NFA"
        .align 3,0
TNFA:
        .dword    DOCOL
        .dword    TNFA+HEADSIZE
        .dword    0x0
        .dword    TLFA
        .dword    N_TNFA
        .dword    0

        .dword      LIT,(CW*(N_HOFFSET))
        .dword      PLUS
        .dword      SEMIS
#


#
#  ************
#  *   >SFA   *
#  ************
#
        .align 3,0
N_TSFA:
        .dword      4
        .ascii      ">SFA"
        .align 3,0
TSFA:
        .dword    DOCOL
        .dword    TSFA+HEADSIZE
        .dword    0x0
        .dword    TNFA
        .dword    N_TSFA
        .dword    0

        .dword      LIT,(CW*(S_HOFFSET))
        .dword      PLUS
        .dword      SEMIS
#
#
#

#  ************
#  *   >PHA   *
#  ************
#
        .align 3,0
N_TPHA:
        .dword      4
        .ascii      ">PHA"
        .align 3,0
TPHA:
        .dword    DOCOL
        .dword    TPHA+HEADSIZE
        .dword    0x0
        .dword    TSFA
        .dword    N_TPHA
        .dword    0

        .dword      LIT,(CW*(PH_OFFSET))
        .dword      PLUS
        .dword      SEMIS
#

#  *************
#  *   >BODY   *
#  *************
#
        .align 3,0
N_TOBODY:
        .dword      5
        .ascii      ">BODY"
        .align 3,0
TOBODY:
        .dword    DOCOL
        .dword    TOBODY+HEADSIZE
        .dword    0x0
        .dword    TPHA
        .dword    N_TOBODY
        .dword    0

        .dword      TDFA, FETCH
        .dword      CELLP           # Skip DOES> pointer.
        .dword      SEMIS
#

#  *************
#  *   BODY>   *
#  *************
#
        .align 3,0
N_BODYF:
        .dword      5
        .ascii      "BODY>"
        .align 3,0
BODYF:
        .dword    DOCOL
        .dword    BODYF+HEADSIZE
        .dword    0x0
        .dword    TOBODY
        .dword    N_BODYF
        .dword    0

        .dword      LIT,(CW*(BD_OFFSET))
        .dword      LSUB
        .dword      SEMIS
#

#  ************
#  *   >WID   *
#  ************
#
        .align 3,0
N_TWID:
        .dword      4
        .ascii      ">WID"
        .align 3,0
TWID:
        .dword    DOCOL
        .dword    TWID+HEADSIZE
        .dword    0x0
        .dword    BODYF
        .dword    N_TWID
        .dword    0

        .dword      TOBODY
        .dword      CELLP # Skip vfa link.
        .dword      SEMIS
#

#  ************
#  *   >VFA   *
#  ************
#
        .align 3,0
N_TVFA:
        .dword      4
        .ascii      ">VFA"
        .align 3,0
TVFA:
        .dword    DOCOL
        .dword    TVFA+HEADSIZE
        .dword    0x0
        .dword    TWID
        .dword    N_TVFA
        .dword    0

        .dword      TOBODY
        .dword      SEMIS
#


#  ************
#  *   !CSP   *
#  ************
#
        .align 3,0
N_SCSP:
        .dword      4
        .ascii      "!CSP"
        .align 3,0
SCSP:
        .dword    DOCOL
        .dword    SCSP+HEADSIZE
        .dword    0x0
        .dword    TVFA
        .dword    N_SCSP
        .dword    0

        .dword      SPFET
        .dword      LCSP
        .dword      STOR
        .dword      SEMIS
#
#

#  **************
#  *   ?ERROR   *
#  **************
#
        .align 3,0
N_QERR:
        .dword      6
        .ascii      "?ERROR"
        .align 3,0
QERR:
        .dword    DOCOL
        .dword    QERR+HEADSIZE
        .dword    0x0
        .dword    SCSP
        .dword    N_QERR
        .dword    0

        .dword      SWAP
        .dword      ZBRAN
        .dword      QERR1-.-CW #IF
        .dword      PP, FETCH
        .dword      SRC, FETCH
        .dword      LWHERE, TSTOR
        .dword      THROW
        .dword      BRAN
        .dword      QERR2-.-CW  #ELSE
QERR1:  .dword      DROP    #THEN
QERR2:  .dword      SEMIS
#

#  **************
#  *   ?ERRUR   *
#  **************
#
        .align 3,0
N_QERRUR:
        .dword      6
        .ascii      "?ERRUR"
        .align 3,0
QERRUR:
        .dword    DOCOL
        .dword    QERRUR+HEADSIZE
        .dword    0x0
        .dword    QERR
        .dword    N_QERRUR
        .dword    0

        .dword      ZERO, MIN, LDUP, QERR
        .dword      SEMIS
#


#  **************
#  *   ?DELIM   *
#  **************
#
        .align 3,0
N_QDELIM:
        .dword      6
        .ascii      "?DELIM"
        .align 3,0
QDELIM:
        .dword    DOCOL
        .dword    QDELIM+HEADSIZE
        .dword    0x0
        .dword    QERRUR
        .dword    N_QDELIM
        .dword    0

        .dword      PPFET
        .dword      QBL
        .dword      ZEQU
        .dword      LIT, 10, QERR
        .dword      DROP
        .dword      SEMIS
#

#  ************
#  *   ?CSP   *
#  ************
#
        .align 3,0
N_QCSP:
        .dword      4
        .ascii      "?CSP"
        .align 3,0
QCSP:
        .dword    DOCOL
        .dword    QCSP+HEADSIZE
        .dword    0x0
        .dword    QDELIM
        .dword    N_QCSP
        .dword    0

        .dword      SPFET
        .dword      LCSP
        .dword      FETCH
        .dword      LSUB
        .dword      LIT, 20, QERR
        .dword      SEMIS
#

#  *************
#  *   ?COMP   *
#  *************
#
        .align 3,0
N_QCOMP:
        .dword      5
        .ascii      "?COMP"
        .align 3,0
QCOMP:
        .dword    DOCOL
        .dword    QCOMP+HEADSIZE
        .dword    0x0
        .dword    QCSP
        .dword    N_QCOMP
        .dword    0

        .dword      STATE
        .dword      FETCH
        .dword      ZEQU
        .dword      LIT, 17, QERR
        .dword      SEMIS
#

# 

#  **************
#  *   ?PAIRS   *
#  **************
#
        .align 3,0
N_QPAIR:
        .dword      6
        .ascii      "?PAIRS"
        .align 3,0
QPAIR:
        .dword    DOCOL
        .dword    QPAIR+HEADSIZE
        .dword    0x0
        .dword    QCOMP
        .dword    N_QPAIR
        .dword    0

        .dword      LSUB
        .dword      LIT, 19, QERR
        .dword      SEMIS
#

# 
#
#
#

#  *********
#  *   [   *
#  *********
#
        .align 3,0
N_LBRAC:
        .dword      1
        .ascii      "["
        .align 3,0
LBRAC:
        .dword    DOCOL
        .dword    LBRAC+HEADSIZE
        .dword    B_IMMED
        .dword    QPAIR
        .dword    N_LBRAC
        .dword    0

        .dword      ZERO
        .dword      STATE
        .dword      STOR
        .dword      SEMIS
#

#  *********
#  *   ]   *
#  *********
#
        .align 3,0
N_RBRAC:
        .dword      1
        .ascii      "]"
        .align 3,0
RBRAC:
        .dword    DOCOL
        .dword    RBRAC+HEADSIZE
        .dword    0x0
        .dword    LBRAC
        .dword    N_RBRAC
        .dword    0

        .dword      ONE
        .dword      STATE
        .dword      STOR
        .dword      SEMIS
#

#  **************
#  *   HIDDEN   *
#  **************
#
        .align 3,0
N_HIDDEN:
        .dword      6
        .ascii      "HIDDEN"
        .align 3,0
HIDDEN:
        .dword    DOCOL
        .dword    HIDDEN+HEADSIZE
        .dword    0x0
        .dword    RBRAC
        .dword    N_HIDDEN
        .dword    0

        .dword      TFFA
        .dword      LIT,B_INVIS
        .dword      TOGGL
        .dword      SEMIS
#

#  ***********
#  *   HEX   *
#  ***********
#
        .align 3,0
N_HEX:
        .dword      3
        .ascii      "HEX"
        .align 3,0
HEX:
        .dword    DOCOL
        .dword    HEX+HEADSIZE
        .dword    0x0
        .dword    HIDDEN
        .dword    N_HEX
        .dword    0

        .dword      LIT,16
        .dword      BASE
        .dword      STOR
        .dword      SEMIS
#

#  ***************
#  *   DECIMAL   *
#  ***************
#
        .align 3,0
N_DECA:
        .dword      7
        .ascii      "DECIMAL"
        .align 3,0
DECA:
        .dword    DOCOL
        .dword    DECA+HEADSIZE
        .dword    0x0
        .dword    HEX
        .dword    N_DECA
        .dword    0

        .dword      LIT,10
        .dword      BASE
        .dword      STOR
        .dword      SEMIS
#

#  ***************
#  *   (;CODE)   *
#  ***************
#
        .align 3,0
N_PSCOD:
        .dword      7
        .ascii      "(;CODE)"
        .align 3,0
PSCOD:
        .dword    DOCOL
        .dword    PSCOD+HEADSIZE
        .dword    0x0
        .dword    DECA
        .dword    N_PSCOD
        .dword    0

        .dword      FROMR
        .dword      LATEST
        .dword      TCFA
        .dword      STOR
        .dword      SEMIS
#

#

#  **************
#  *   CREATE   *
#  **************
#
        .align 3,0
N_LCREATE:
        .dword      6
        .ascii      "CREATE"
        .align 3,0
LCREATE:
        .dword    DOCOL
        .dword    LCREATE+HEADSIZE
        .dword    0x0
        .dword    PSCOD
        .dword    N_LCREATE
        .dword    0

        .dword      LNAME
        .dword      PCREAT
        .dword      LIT, HLNOOP, COMMA
        .dword      PSCOD
DODOE:  ADDI    fp, fp, -(CW*(1))
        SD      a7, 0(fp) 
        LD     a7, (CW*(D_HOFFSET))(a0)        #NEW IP 
        MV     a0, a7
        ADDI     a0, a0, (CW*(1))
        LD     a7,0(a7)
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
HLNOOP: .dword      SEMIS
#

#  *************
#  *   DOES>   *
#  *************
#
        .align 3,0
N_DOES:
        .dword      5
        .ascii      "DOES>"
        .align 3,0
DOES:
        .dword    DOCOL
        .dword    DOES+HEADSIZE
        .dword    0x0
        .dword    LCREATE
        .dword    N_DOES
        .dword    0

        .dword      FROMR
        .dword      LATEST
        .dword      TDFA
        .dword      FETCH
        .dword      STOR
        .dword      SEMIS
#

#  *************
#  *   COUNT   *
#  *************
#
        .align 3,0
N_COUNT:
        .dword      5
        .ascii      "COUNT"
        .align 3,0
COUNT:
        .dword    DOCOL
        .dword    COUNT+HEADSIZE
        .dword    0x0
        .dword    DOES
        .dword    N_COUNT
        .dword    0

        .dword      LDUP
        .dword      ONEP
        .dword      SWAP
        .dword      CFET
        .dword      SEMIS
#

#  *****************
#  *   -TRAILING   *
#  *****************
#
        .align 3,0
N_DTRAI:
        .dword      9
        .ascii      "-TRAILING"
        .align 3,0
DTRAI:
        .dword    DOCOL
        .dword    DTRAI+HEADSIZE
        .dword    0x0
        .dword    COUNT
        .dword    N_DTRAI
        .dword    0

        .dword      LDUP
        .dword      ZERO
        .dword     XQDO
        .dword      DTRA4-.-CW
DTRA1:  .dword      OVER
        .dword      OVER
        .dword      PLUS
        .dword      ONE
        .dword      LSUB
        .dword      CFET
        .dword      QBL
        .dword      ZEQU
        .dword      ZBRAN
        .dword      DTRA2-.-CW #IF
        .dword      LLEAV
DTRA2:  .dword      ONE
        .dword      LSUB    # THEN
        .dword     ONE, XPLOOP, ZBRAN
        .dword      DTRA1-.-CW
        .dword      UNLOOP    # LOOP
DTRA4:
        .dword      SEMIS
#

#  **********
#  *   S"   *
#  **********
#
        .align 3,0
N_SQUOT:
        .dword      2
        .ascii      "S\""
        .align 3,0
SQUOT:
        .dword    DOCOL
        .dword    SQUOT+HEADSIZE
        .dword    B_IMMED
        .dword    DTRAI
        .dword    N_SQUOT
        .dword    0

        .dword      DENQ
        .dword      SEMIS
#

#  **********
#  *   ."   *
#  **********
#
        .align 3,0
N_DOTQ:
        .dword      2
        .ascii      ".\""
        .align 3,0
DOTQ:
        .dword    DOCOL
        .dword    DOTQ+HEADSIZE
        .dword    B_IMMED
        .dword    SQUOT
        .dword    N_DOTQ
        .dword    0

        .dword      DENQ
        .dword      STATE
        .dword      FETCH
        .dword      ZBRAN
        .dword      DOTQ1-.-CW # IF
        .dword      LIT, LTYPE, COMMA
        .dword      BRAN
        .dword      DOTQ2-.-CW
DOTQ1:
        .dword      LTYPE
DOTQ2:
        .dword      SEMIS   # THEN
#

#  **********
#  *   .(   *
#  **********
#
        .align 3,0
N_DOTP:
        .dword      2
        .ascii      ".("
        .align 3,0
DOTP:
        .dword    DOCOL
        .dword    DOTP+HEADSIZE
        .dword    B_IMMED
        .dword    DOTQ
        .dword    N_DOTP
        .dword    0

        .dword      LIT, ')'
        .dword      PARSE
        .dword      LTYPE
        .dword      SEMIS
#

#  ***************
#  *   SET-SRC   *
#  ***************
#
        .align 3,0
N_SETSRC:
        .dword      7
        .ascii      "SET-SRC"
        .align 3,0
SETSRC:
        .dword    DOCOL
        .dword    SETSRC+HEADSIZE
        .dword    0x0
        .dword    DOTP
        .dword    N_SETSRC
        .dword    0

        .dword      OVER, PLUS
        .dword      OVER, SRC, TSTOR
        .dword      PP, STOR #  IN

#       DC      DOTS
        .dword      SEMIS
#

#  ****************
#  *   EVALUATE   *
#  ****************
#
        .align 3,0
N_EVALUATE:
        .dword      8
        .ascii      "EVALUATE"
        .align 3,0
EVALUATE:
        .dword    DOCOL
        .dword    EVALUATE+HEADSIZE
        .dword    0x0
        .dword    SETSRC
        .dword    N_EVALUATE
        .dword    0

        .dword      SAVE
        .dword      SETSRC
        .dword      LIT, INTER, CATCH
        .dword      RESTO
        .dword      THROW
        .dword      SEMIS
#

#  ************
#  *   FILL   *
#  ************
#
        .align 3,0
N_FILL:
        .dword      4
        .ascii      "FILL"
        .align 3,0
FILL:
        .dword    FILL+HEADSIZE
        .dword    FILL+HEADSIZE
        .dword    0x0
        .dword    EVALUATE
        .dword    N_FILL
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      # FILL CHAR
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))      # FILL COUNT
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      # BEGIN ADDR
        BEQ a2, zero, FILL2    # COUNT=0? 
FILL1:  SB     a0, 0(a1)
        ADDI    a1, a1, 1 
        ADDI    a2, a2,-1
        BNEZ    a2,FILL1
FILL2:  LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   CORA   *
#  ************
#
        .align 3,0
N_CORA:
        .dword      4
        .ascii      "CORA"
        .align 3,0
CORA:
        .dword    CORA+HEADSIZE
        .dword    CORA+HEADSIZE
        .dword    0x0
        .dword    FILL
        .dword    N_CORA
        .dword    0

        LI    a0,0   # Result, assume all equal
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      # count
        LD      a4, 0(sp)
        ADDI    sp, sp, (CW*(1))      # addr2
        LD      a5, 0(sp)
        ADDI    sp, sp, (CW*(1))      # addr1
CORA1:
        ADDI     a3, a3,-1
        BLT      a3, zero, CORA3        # Stop looping if R3 < 0  and return 0 
        LB     a1, 0(a5)
        ADDI    a5, a5, 1           # load byte at R5 in R1, auto-increment R5 
        LB     a2, 0(a4)
        ADDI    a4, a4, 1           # load byte at R4 in R2, auto-increment R4 
        BEQ      a1,a2, CORA1        # chars equal, keep looping 
        LI       a0, -1
        BLT     a1,a2,CORA3        # R2 is lexicographic later, return 1 
        LI       a0, 1
CORA3:
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   $^   *
#  **********
#
        .align 3,0
N_SINDEX:
        .dword      2
        .ascii      "\$^"
        .align 3,0
SINDEX:
        .dword    SINDEX+HEADSIZE
        .dword    SINDEX+HEADSIZE
        .dword    0x0
        .dword    CORA
        .dword    N_SINDEX
        .dword    0

#     search address of given byte starting from R0 
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))      # char
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))      # count
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      # addr
        LI      a1, 0 #Clear R1 
SINDEX1:
        BEQ     a2, zero, SINDEX3
        LB     a1, 0(a0)
        ADDI    a0, a0, 1          # load next byte from R0 increment R0
        BEQ     a3,a1, SINDEX2          # Compare BYTE 
        ADDI    a2, a2,-1
        BNEZ    a2, SINDEX1
SINDEX3:
        LI    a0,0       # Not found: 0 
        ADDI     a0, a0, 1
SINDEX2:
        ADDI     a0, a0, -1
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0                  # Return address, or 0 
#

#  **********
#  *   $/   *
#  **********
#
        .align 3,0
N_SSLAS:
        .dword      2
        .ascii      "\$/"
        .align 3,0
SSLAS:
        .dword    SSLAS+HEADSIZE
        .dword    SSLAS+HEADSIZE
        .dword    0x0
        .dword    SINDEX
        .dword    N_SSLAS
        .dword    0

        
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))          # char
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))          # count
        MV     a5,a2
        LD      a4, 0(sp)
        ADDI    sp, sp, (CW*(1))          # addr
        MV     a1,a4   # Copy
        BEQZ     a4, SSLAS3   # Nil pointer 
        BEQZ     a2, SSLAS1A
SSLAS1:
        LB     a3, 0(a4)
        ADDI    a4, a4, 1          # load byte at R4 in R3, then auto-increment R4
        BEQ     a0, a3,  SSLAS2
        ADDI    a2, a2, -1
        BNEZ     a2, SSLAS1

# Not present. 
SSLAS1A:
        ADDI    sp, sp, -(CW*(1))
        SD      a2, 0(sp)     # Nil pointer.
        JAL     t0, SSLAS4
SSLAS2:
        SUB     a5,a5,a2
        ADDI     a2, a2, -1         #Delimiter is not part of first string.
SSLAS3:
        ADDI    sp, sp, -(CW*(1))
        SD      a4, 0(sp) 
SSLAS4:
        ADDI    sp, sp, -(CW*(1))
        SD      a2, 0(sp)    #Remaining length
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp)    #Start of first string.
        ADDI    sp, sp, -(CW*(1))
        SD      a5, 0(sp)    #Skipped length.
        
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#
#  **********
#  *   $\   *
#  **********
#
        .align 3,0
N_BSLAS:
        .dword      2
        .ascii      "\$\\"
        .align 3,0
BSLAS:
        .dword    BSLAS+HEADSIZE
        .dword    BSLAS+HEADSIZE
        .dword    0x0
        .dword    SSLAS
        .dword    N_BSLAS
        .dword    0
   #_dummy
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  

#  *************
#  *   ERASE   *
#  *************
#
        .align 3,0
N_LERASE:
        .dword      5
        .ascii      "ERASE"
        .align 3,0
LERASE:
        .dword    DOCOL
        .dword    LERASE+HEADSIZE
        .dword    0x0
        .dword    BSLAS
        .dword    N_LERASE
        .dword    0

        .dword      ZERO
        .dword      FILL
        .dword      SEMIS
#

#  *************
#  *   BLANK   *
#  *************
#
        .align 3,0
N_BLANK:
        .dword      5
        .ascii      "BLANK"
        .align 3,0
BLANK:
        .dword    DOCOL
        .dword    BLANK+HEADSIZE
        .dword    0x0
        .dword    LERASE
        .dword    N_BLANK
        .dword    0

        .dword      LBL
        .dword      FILL
        .dword      SEMIS
#

#  ************
#  *   HOLD   *
#  ************
#
        .align 3,0
N_HOLD:
        .dword      4
        .ascii      "HOLD"
        .align 3,0
HOLD:
        .dword    DOCOL
        .dword    HOLD+HEADSIZE
        .dword    0x0
        .dword    BLANK
        .dword    N_HOLD
        .dword    0

        .dword      LIT,-1
        .dword      HLD
        .dword      PSTOR
        .dword      HLD
        .dword      FETCH
        .dword      CSTOR
        .dword      SEMIS
#

#  ***********
#  *   PAD   *
#  ***********
#
        .align 3,0
N_PAD:
        .dword      3
        .ascii      "PAD"
        .align 3,0
PAD:
        .dword    DOCOL
        .dword    PAD+HEADSIZE
        .dword    0x0
        .dword    HOLD
        .dword    N_PAD
        .dword    0

        .dword      HERE
# Allow for a one line name, a double binary number and some hold char's
        .dword      LIT,84+128+64
        .dword      PLUS
        .dword      SEMIS
#

# 
#

#  ************
#  *   CHAR   *
#  ************
#
        .align 3,0
N_LCHAR:
        .dword      4
        .ascii      "CHAR"
        .align 3,0
LCHAR:
        .dword    DOCOL
        .dword    LCHAR+HEADSIZE
        .dword    0x0
        .dword    PAD
        .dword    N_LCHAR
        .dword    0

        .dword      LNAME, DROP, CFET
        .dword      SEMIS
#

#  **************
#  *   [CHAR]   *
#  **************
#
        .align 3,0
N_BCHAR:
        .dword      6
        .ascii      "[CHAR]"
        .align 3,0
BCHAR:
        .dword    DOCOL
        .dword    BCHAR+HEADSIZE
        .dword    B_IMMED
        .dword    LCHAR
        .dword    N_BCHAR
        .dword    0

        .dword      LCHAR, LITER
        .dword      SEMIS
#

#  ****************
#  *   (NUMBER)   *
#  ****************
#
        .align 3,0
N_PNUMB:
        .dword      8
        .ascii      "(NUMBER)"
        .align 3,0
PNUMB:
        .dword    DOCOL
        .dword    PNUMB+HEADSIZE
        .dword    0x0
        .dword    BCHAR
        .dword    N_PNUMB
        .dword    0

        .dword      ZERO, ZERO
        .dword      ZERO, DPL, STOR
NPNUM1:  .dword      PPFET   # BEGIN
        .dword      LDUP, LIT, ADOT, EQUAL
        .dword      ZBRAN
        .dword      NPNUM2-.-CW # case 1 
        .dword      DROP, DPL, STOR
        .dword      BRAN
        .dword      NPNUM1-.-CW # continue 
NPNUM2:
        .dword      LDUP, LIT, ',', EQUAL
        .dword      ZBRAN
        .dword      NPNUM3-.-CW # case 2 
        .dword      TDROP
        .dword      BRAN
        .dword      NPNUM1-.-CW # continue 
NPNUM3:
        .dword      LDUP, QBL, ZEQU
        .dword      ZBRAN
        .dword      NPNUM5-.-CW  # while 
        .dword      NIP
        .dword      BASE, FETCH, DIGIT
        .dword      ZEQU
        .dword      LIT, 10, QERR

        .dword      SWAP
        .dword      BASE
        .dword      FETCH
        .dword      USTAR
        .dword      DROP
        .dword      ROT
        .dword      BASE
        .dword      FETCH
        .dword      USTAR
        .dword      DPLUS
        .dword      BRAN
        .dword      NPNUM1-.-CW # repeat 
NPNUM5:
        .dword      DROP, PP, STOR
        .dword      SEMIS
#

#  **************
#  *   NUMBER   *
#  **************
#
        .align 3,0
N_NUMB:
        .dword      6
        .ascii      "NUMBER"
        .align 3,0
NUMB:
        .dword    DOCOL
        .dword    NUMB+HEADSIZE
        .dword    0x0
        .dword    PNUMB
        .dword    N_NUMB
        .dword    0

LNUMB:
        .dword      LIT, -1, PP, PSTOR
        .dword      PNUMB, SDLITE
        .dword      SEMIS
#

#  ***************
#  *   >NUMBER   *
#  ***************
#
        .align 3,0
N_TONUMB:
        .dword      7
        .ascii      ">NUMBER"
        .align 3,0
TONUMB:
        .dword    DOCOL
        .dword    TONUMB+HEADSIZE
        .dword    0x0
        .dword    NUMB
        .dword    N_TONUMB
        .dword    0

        .dword      TDUP, PLUS, TOR     # End available on return stack.
        .dword      ZERO
        .dword     XQDO
        .dword      TONUM9-.-CW
TONUM1:
        .dword      LDUP, CFET, BASE, FETCH, DIGIT
        .dword      ZEQU
        .dword      ZBRAN
        .dword      TONUM4-.-CW # IF
        .dword      DROP
        .dword      LLEAV
TONUM4:
        .dword      SWAP, TOR # Address out of the way.
        .dword      SWAP
        .dword      BASE
        .dword      FETCH
        .dword      USTAR
        .dword      DROP
        .dword      ROT
        .dword      BASE
        .dword      FETCH
        .dword      USTAR
        .dword      DPLUS
        .dword      FROMR, ONEP     # Address back.
        .dword     ONE, XPLOOP, ZBRAN
        .dword      TONUM1-.-CW
        .dword      UNLOOP
TONUM9:
        .dword      FROMR
        .dword      OVER, LSUB
        .dword      SEMIS
#

#  *************
#  *   FOUND   *
#  *************
#
        .align 3,0
N_FOUND:
        .dword      5
        .ascii      "FOUND"
        .align 3,0
FOUND:
        .dword    DOCOL
        .dword    FOUND+HEADSIZE
        .dword    0x0
        .dword    TONUMB
        .dword    N_FOUND
        .dword    0

        .dword      CONTEXT, TOR
FOUND1: .dword      RR, FETCH
#        DC      DOTS
        .dword      PFIND, LDUP, ZEQU
        .dword      ZBRAN
        .dword      FOUND3-.-CW
        .dword      DROP
        .dword      RR, FETCH, LIT, ONLYBODY, LSUB #Was this ONLY?
        .dword      ZBRAN
        .dword      FOUND2-.-CW
        .dword      FROMR, CELLP, TOR
        .dword      BRAN
        .dword      FOUND1-.-CW
FOUND2: .dword      ZERO
FOUND3: .dword      RDROP
        .dword      NIP,NIP
        .dword      SEMIS
#

#  ***************
#  *   PRESENT   *
#  ***************
#
        .align 3,0
N_PRESENT:
        .dword      7
        .ascii      "PRESENT"
        .align 3,0
PRESENT:
        .dword    DOCOL
        .dword    PRESENT+HEADSIZE
        .dword    0x0
        .dword    FOUND
        .dword    N_PRESENT
        .dword    0

        .dword      LDUP, TOR
        .dword      FOUND
        .dword      LDUP
        .dword      ZBRAN
        .dword      PRES1-.-CW
        .dword      LDUP
        .dword      TNFA, FETCH, FETCH #  Get precise length.
        .dword      RR, EQUAL
        .dword      LAND
PRES1:
        .dword      RDROP
        .dword      SEMIS
#


# 
#

#

#  **************
#  *   (FIND)   *
#  **************
#
        .align 3,0
N_PFIND:
        .dword      6
        .ascii      "(FIND)"
        .align 3,0
PFIND:
        .dword    DOCOL
        .dword    PFIND+HEADSIZE
        .dword    0x0
        .dword    PRESENT
        .dword    N_PFIND
        .dword    0

# The idea is to have a fast slim loop.
# The optimisation tries to skip, using the extra links.
# At all labels the stack is (sc, dea).
# PFIND0 : nothing known yet about dea
# PFCONTINUE : dea is disapproved, go on 
# PFIND1 : dea is a match, inasfar name and prefix
# PFEXIT : dea is the match, or null.

PFIND0:                       #Start of fast loop.
         .dword      NMATCH
         .dword      ZEQU
        .dword      ZBRAN
        .dword      PFCONTINUE-.-CW
# dea's name and sc have equal length
        .dword      TDUP, TNFA, FETCH, FETCH
        .dword      EQUAL
# dea is a prefix 
        .dword      OVER, TFFA, FETCH
        .dword      LIT, B_DENOT, LAND
        .dword      LOR             # One of those ,ist apply.
        .dword      ZBRAN
        .dword      PFCONTINUE-.-CW
# Discard still, if invisible ("smudged") or dummy.
        .dword      LDUP, TFFA, FETCH
        .dword      LIT, B_INVIS | B_DUMMY, LAND
        .dword      ZEQU
        .dword      ZBRAN
        .dword      PFCONTINUE-.-CW
        .dword      BRAN
        .dword      PFEXIT-.-CW         # Match! 
PFCONTINUE:
        .dword      TLFA, FETCH
        .dword      LDUP, ZEQU
        .dword      ZBRAN
        .dword      PFIND0-.-CW
PFEXIT:
        .dword      SEMIS
#

#  *************
#  *   ERROR   *
#  *************
#
        .align 3,0
N_LERROR:
        .dword      5
        .ascii      "ERROR"
        .align 3,0
LERROR:
        .dword    DOCOL
        .dword    LERROR+HEADSIZE
        .dword    0x0
        .dword    PFIND
        .dword    N_LERROR
        .dword    0

# Make sure the whole word is shown in case of denotations. 
        .dword      LWHERE, TFET
        .dword      OVER
        .dword      ZBRAN
        .dword      ERR2-.-CW      # Virgin: print empty string.
        .dword      OVER, LIT, 20, LSUB
        .dword      MAX
        .dword      SWAP
        .dword      ONEM         #Backup for if we are at next word.
ERR1:   .dword      LDUP, CFET, QBL, ZEQU  #Show whole word!
        .dword      ZBRAN
        .dword      ERR2 -.-CW
        .dword      ONEP
        .dword      BRAN
        .dword      ERR1 -.-CW
ERR2:
        .dword      OVER, LSUB
        .dword      ETYPE
        .dword      SKIP
         .dword      19
SB1: .ascii      " ? ciforth ERROR # "
       .align 3,0
        .dword      LIT, SB1
        .dword      LIT, 19
        .dword      ETYPE
        .dword      BASE, FETCH
        .dword      DECA
        .dword      OVER
        .dword      STOD, ZERO, PDDOTR      #This is about (.) 
        .dword      ETYPE
        .dword      BASE, STOR
        .dword      LIT, -256, MAX
        .dword      LIT, 63, MIN
        .dword      LIT, MESS, CATCH, DROP  #Prevent error loops 
        .dword      SEMIS
#

#  *************
#  *   CATCH   *
#  *************
#
        .align 3,0
N_CATCH:
        .dword      5
        .ascii      "CATCH"
        .align 3,0
CATCH:
        .dword    DOCOL
        .dword    CATCH+HEADSIZE
        .dword    0x0
        .dword    LERROR
        .dword    N_CATCH
        .dword    0

        .dword      SPFET, CELLP, TOR
        .dword      HANDLER, FETCH, TOR
        .dword      RPFET, HANDLER, STOR
        .dword      EXEC
        .dword      FROMR, HANDLER, STOR
        .dword      RDROP, ZERO
        .dword      SEMIS
#

#  *************
#  *   THROW   *
#  *************
#
        .align 3,0
N_THROW:
        .dword      5
        .ascii      "THROW"
        .align 3,0
THROW:
        .dword    DOCOL
        .dword    THROW+HEADSIZE
        .dword    0x0
        .dword    CATCH
        .dword    N_THROW
        .dword    0

        .dword      LDUP
        .dword      ZBRAN
        .dword      THROW1-.-CW
        .dword      HANDLER, FETCH, ZEQU
        .dword      ZBRAN
        .dword      THROW2-.-CW
        .dword      LERROR
        .dword      CLS
        .dword      QUIT
THROW2:
        .dword      HANDLER, FETCH, RPSTO
        .dword      FROMR, HANDLER, STOR
        .dword      FROMR, SWAP, TOR
        .dword      SPSTO
        .dword      FROMR
        .dword      X
THROW1:
        .dword      DROP
        .dword      SEMIS
#

# 

# 
#

#  ***********
#  *   ID.   *
#  ***********
#
        .align 3,0
N_IDDOT:
        .dword      3
        .ascii      "ID."
        .align 3,0
IDDOT:
        .dword    DOCOL
        .dword    IDDOT+HEADSIZE
        .dword    0x0
        .dword    THROW
        .dword    N_IDDOT
        .dword    0

        .dword      LDUP, TFFA, FETCH
        .dword      LIT, B_DUMMY, LAND
        .dword      ZBRAN
        .dword      IDDOT1-.-CW
        .dword      DROP
        .dword      BRAN
        .dword      IDDOT2-.-CW
IDDOT1:
        .dword      TNFA
        .dword      FETCH
        .dword      SFET
        .dword      LTYPE
        .dword      SPACE
        .dword      SPACE
        .dword      SPACE
IDDOT2:
        .dword      SEMIS
#

#  **************
#  *   HEADER   *
#  **************
#
        .align 3,0
N_LHEAD:
        .dword      6
        .ascii      "HEADER"
        .align 3,0
LHEAD:
        .dword    DOCOL
        .dword    LHEAD+HEADSIZE
        .dword    0x0
        .dword    IDDOT
        .dword    N_LHEAD
        .dword    0

        .dword      LDUP
        .dword      ZEQU
        .dword      LIT, 5, QERR
        .dword      TDUP
        .dword      PRESENT
        .dword      LDUP
        .dword      ZBRAN
        .dword      CREA1-.-CW #IF
        .dword      TNFA, FETCH, SFET
        .dword      ETYPE
        .dword      LIT,4
        .dword      MESS
        .dword      X       #THEN
CREA1:  .dword      DROP
 .dword      LALIGN
        .dword      SCOMMA
 .dword      LALIGN
        .dword      HERE,TOR

        .dword      RR, TPHA, COMMA         #Code field.
        .dword      RR, TPHA, COMMA         #Data field.
        .dword      ZERO, COMMA #Flag field.
        .dword      ZERO, COMMA #Link field.
        .dword      COMMA   #Name field.


        .dword      BLK, FETCH, LDUP, ZEQU
        .dword      ZBRAN
        .dword      CREA2-.-CW
        .dword      DROP, PP, FETCH
CREA2:  .dword      COMMA  #Source field.

                .dword      FROMR
        .dword      SEMIS
#

#  ************
#  *   LINK   *
#  ************
#
        .align 3,0
N_LLINK:
        .dword      4
        .ascii      "LINK"
        .align 3,0
LLINK:
        .dword    DOCOL
        .dword    LLINK+HEADSIZE
        .dword    0x0
        .dword    LHEAD
        .dword    N_LLINK
        .dword    0

        .dword      TLFA
        .dword      TDUP
        .dword      FETCH
        .dword      SWAP, TLFA, STOR
        .dword      STOR
        .dword      SEMIS
#

#  ****************
#  *   (CREATE)   *
#  ****************
#
        .align 3,0
N_PCREAT:
        .dword      8
        .ascii      "(CREATE)"
        .align 3,0
PCREAT:
        .dword    DOCOL
        .dword    PCREAT+HEADSIZE
        .dword    0x0
        .dword    LLINK
        .dword    N_PCREAT
        .dword    0

        .dword      LHEAD
        .dword      CURR, FETCH
        .dword      LLINK
        .dword      SEMIS
#

#  *****************
#  *   [COMPILE]   *
#  *****************
#
        .align 3,0
N_BCOMP:
        .dword      9
        .ascii      "[COMPILE]"
        .align 3,0
BCOMP:
        .dword    DOCOL
        .dword    BCOMP+HEADSIZE
        .dword    B_IMMED
        .dword    PCREAT
        .dword    N_BCOMP
        .dword    0

        .dword      LNAME
        .dword      PRESENT
        .dword      LDUP
        .dword      ZEQU
        .dword      LIT, 16, QERR
        .dword      TCFA
        .dword      COMMA
        .dword      SEMIS
#

#  ****************
#  *   POSTPONE   *
#  ****************
#
        .align 3,0
N_POSTP:
        .dword      8
        .ascii      "POSTPONE"
        .align 3,0
POSTP:
        .dword    DOCOL
        .dword    POSTP+HEADSIZE
        .dword    B_IMMED
        .dword    BCOMP
        .dword    N_POSTP
        .dword    0

        .dword      LNAME
        .dword      PRESENT
        .dword      LDUP
        .dword      ZEQU
        .dword      LIT, 15, QERR
        .dword      LDUP, TFFA, FETCH
        .dword      LIT, B_IMMED, LAND, ZEQU
        .dword      ZBRAN
        .dword      POSTP1-.-CW
         .dword      LIT, LIT, COMMA
         .dword      COMMA
         .dword      LIT, COMMA
POSTP1:
         .dword      COMMA
        .dword      SEMIS
#

#  ***************
#  *   LITERAL   *
#  ***************
#
        .align 3,0
N_LITER:
        .dword      7
        .ascii      "LITERAL"
        .align 3,0
LITER:
        .dword    DOCOL
        .dword    LITER+HEADSIZE
        .dword    B_IMMED
        .dword    POSTP
        .dword    N_LITER
        .dword    0

        .dword      STATE
        .dword      FETCH
        .dword      ZBRAN
        .dword      LITE1-.-CW #IF
        .dword      LIT, LIT, COMMA
        .dword      COMMA   #THEN
LITE1:  .dword      SEMIS
#

#  ****************
#  *   DLITERAL   *
#  ****************
#
        .align 3,0
N_DLITE:
        .dword      8
        .ascii      "DLITERAL"
        .align 3,0
DLITE:
        .dword    DOCOL
        .dword    DLITE+HEADSIZE
        .dword    B_IMMED
        .dword    LITER
        .dword    N_DLITE
        .dword    0

        .dword      STATE
        .dword      FETCH
        .dword      ZBRAN
        .dword      DLIT1-.-CW # IF
        .dword      SWAP
        .dword      LITER
        .dword      LITER   # THEN
DLIT1:  .dword      SEMIS
#
#

#  *****************
#  *   SDLITERAL   *
#  *****************
#
        .align 3,0
N_SDLITE:
        .dword      9
        .ascii      "SDLITERAL"
        .align 3,0
SDLITE:
        .dword    DOCOL
        .dword    SDLITE+HEADSIZE
        .dword    B_IMMED
        .dword    DLITE
        .dword    N_SDLITE
        .dword    0

        .dword      DPL
        .dword      FETCH
        .dword      ZBRAN
        .dword      SDLIT1-.-CW # IF
        .dword      DLITE
        .dword      BRAN
        .dword      SDLIT2-.-CW # IF
SDLIT1:
        .dword      DROP, LITER
SDLIT2:
        .dword      SEMIS
#


#  **************
#  *   ?STACK   *
#  **************
#
        .align 3,0
N_QSTAC:
        .dword      6
        .ascii      "?STACK"
        .align 3,0
QSTAC:
        .dword    DOCOL
        .dword    QSTAC+HEADSIZE
        .dword    0x0
        .dword    SDLITE
        .dword    N_QSTAC
        .dword    0

        .dword      SPFET
        .dword      SZERO
        .dword      FETCH
        .dword      SWAP
        .dword      ULESS
        .dword      ONE, QERR
        .dword      SPFET
        .dword      HERE
        .dword      LIT,0x80
        .dword      PLUS
        .dword      ULESS
        .dword      LIT, 7, QERR
        .dword      SEMIS
        #
#
#

#  *****************
#  *   INTERPRET   *
#  *****************
#
        .align 3,0
N_INTER:
        .dword      9
        .ascii      "INTERPRET"
        .align 3,0
INTER:
        .dword    DOCOL
        .dword    INTER+HEADSIZE
        .dword    0x0
        .dword    QSTAC
        .dword    N_INTER
        .dword    0

INTE1:
        .dword      LNAME
        .dword      LDUP      # Zero length.
        .dword      ZBRAN
        .dword      INTE8-.-CW #WHILE
#       DC      DOTS
#       DC      TDUP, LTYPE
        .dword      OVER, TOR       # Save old parse pointer.
        .dword      FOUND
        .dword      LDUP, ZEQU
        .dword      LIT, 12, QERR
        .dword      LDUP, TFFA, FETCH
        .dword      LDUP, LIT, B_DENOT, LAND #Retain copy of flags.
        .dword      ZBRAN
        .dword      INTE3B-.-CW #IF
        .dword      OVER, TNFA, FETCH, FETCH
        .dword      RR, PLUS, PP, STOR  #Skip over prefix.
INTE3B:                  #THEN 
        .dword      RDROP           # Drop old parse pointer.
        .dword      LIT, B_IMMED, LAND
        .dword      STATE, FETCH, ZEQU, LOR
        .dword      ZBRAN
        .dword      INTE3-.-CW #IF
        .dword      EXEC
        .dword      BRAN
        .dword      INTE4-.-CW #IF
INTE3:
        .dword      COMMA
                        #THEN
INTE4:
        .dword      QSTAC
        .dword      BRAN
        .dword      INTE1-.-CW  #AGAIN
INTE8:  .dword      DROP, DROP
        .dword      SEMIS
#

#  *****************
#  *   IMMEDIATE   *
#  *****************
#
        .align 3,0
N_IMMED:
        .dword      9
        .ascii      "IMMEDIATE"
        .align 3,0
IMMED:
        .dword    DOCOL
        .dword    IMMED+HEADSIZE
        .dword    0x0
        .dword    INTER
        .dword    N_IMMED
        .dword    0

        .dword      LATEST
        .dword      TFFA
        .dword      LIT, B_IMMED
        .dword      TOGGL
        .dword      SEMIS
#

#  **************
#  *   PREFIX   *
#  **************
#
        .align 3,0
N_PREFX:
        .dword      6
        .ascii      "PREFIX"
        .align 3,0
PREFX:
        .dword    DOCOL
        .dword    PREFX+HEADSIZE
        .dword    0x0
        .dword    IMMED
        .dword    N_PREFX
        .dword    0

        .dword      LATEST
        .dword      TFFA
        .dword      LIT, B_DENOT
        .dword      TOGGL
        .dword      SEMIS
#

#  *****************
#  *   NAMESPACE   *
#  *****************
#
        .align 3,0
N_VOCAB:
        .dword      9
        .ascii      "NAMESPACE"
        .align 3,0
VOCAB:
        .dword    DOCOL
        .dword    VOCAB+HEADSIZE
        .dword    0x0
        .dword    PREFX
        .dword    N_VOCAB
        .dword    0

        .dword      LCREATE
        .dword      LATEST   # Link this DEA into VOC-LINK chain.
        .dword      VOCL
        .dword      FETCH
        .dword      COMMA
        .dword      VOCL
        .dword      STOR
        .dword      WORDLIST, DROP
        .dword      DOES
DOVOC:
        .dword      ALSO
        .dword      CELLP   # Make it a WID. 
        .dword      CONTEXT
        .dword      STOR
        .dword      SEMIS
        #
#
#   The link to task is a cold start value only.
#   It is updated each time a definition is
#   appended to the 'FORTH' namespace.
#

#  ****************
#  *   WORDLIST   *
#  ****************
#
        .align 3,0
N_WORDLIST:
        .dword      8
        .ascii      "WORDLIST"
        .align 3,0
WORDLIST:
        .dword    DOCOL
        .dword    WORDLIST+HEADSIZE
        .dword    0x0
        .dword    VOCAB
        .dword    N_WORDLIST
        .dword    0

        .dword      HERE
        .dword      ZERO, COMMA   # Dummy code field
        .dword      ZERO, COMMA   # Dummy data field
        .dword      LIT, B_DUMMY, COMMA # Dummy flag field
        .dword      ZERO, COMMA #Link field: empty chain
        .dword      LIT, (ZERO+(CW*(D_HOFFSET))), COMMA #Empty string for name.
        .dword      ZERO, COMMA
                .dword      SEMIS
        #
#

#

#  *******************
#  *   DEFINITIONS   *
#  *******************
#
        .align 3,0
N_DEFIN:
        .dword      11
        .ascii      "DEFINITIONS"
        .align 3,0
DEFIN:
        .dword    DOCOL
        .dword    DEFIN+HEADSIZE
        .dword    0x0
        .dword    WORDLIST
        .dword    N_DEFIN
        .dword    0

        .dword      CONTEXT
        .dword      FETCH
        .dword      CURR
        .dword      STOR
        .dword      SEMIS
#

#  ************
#  *   ALSO   *
#  ************
#
        .align 3,0
N_ALSO:
        .dword      4
        .ascii      "ALSO"
        .align 3,0
ALSO:
        .dword    DOCOL
        .dword    ALSO+HEADSIZE
        .dword    0x0
        .dword    DEFIN
        .dword    N_ALSO
        .dword    0

        .dword      CONTEXT, LDUP, CELLP
        .dword      LIT, (CW*(16-1))
        .dword      LMOVE
        .dword      LIT, ONLYBODY  #End sentinel for array of word lists.
        .dword      CONTEXT, LIT, (CW*(16)), PLUS
        .dword      STOR #Trim sets of wordset.
        .dword      SEMIS
#

#  ****************
#  *   PREVIOUS   *
#  ****************
#
        .align 3,0
N_PREVI:
        .dword      8
        .ascii      "PREVIOUS"
        .align 3,0
PREVI:
        .dword    DOCOL
        .dword    PREVI+HEADSIZE
        .dword    0x0
        .dword    ALSO
        .dword    N_PREVI
        .dword    0

        .dword      CONTEXT, LDUP, CELLP, SWAP
        .dword      LIT, (CW*(16))
        .dword      LMOVE
        .dword      SEMIS
#

#  *********
#  *   (   *
#  *********
#
        .align 3,0
N_PAREN:
        .dword      1
        .ascii      "("
        .align 3,0
PAREN:
        .dword    DOCOL
        .dword    PAREN+HEADSIZE
        .dword    B_IMMED
        .dword    PREVI
        .dword    N_PAREN
        .dword    0

        .dword      LIT,')'
        .dword      PARSE
        .dword      TDROP
        .dword      SEMIS
#
#

#  *********
#  *   \   *
#  *********
#
        .align 3,0
N_BACKS:
        .dword      1
        .ascii      "\\"
        .align 3,0
BACKS:
        .dword    DOCOL
        .dword    BACKS+HEADSIZE
        .dword    B_IMMED
        .dword    PAREN
        .dword    N_BACKS
        .dword    0

# Backup one character, just in case we are at the end of a line.
        .dword      LIT, -1, PP, PSTOR
        .dword      LIT,ALF
        .dword      PARSE
        .dword      TDROP
        .dword      SEMIS
#

#  ************
#  *   QUIT   *
#  ************
#
        .align 3,0
N_QUIT:
        .dword      4
        .ascii      "QUIT"
        .align 3,0
QUIT:
        .dword    DOCOL
        .dword    QUIT+HEADSIZE
        .dword    0x0
        .dword    BACKS
        .dword    N_QUIT
        .dword    0

        .dword      LBRAC
QUIT1:                  #BEGIN
        .dword      RZERO
        .dword      FETCH
        .dword      RPSTO

# debugging 
#        DC      REFTIB  
# end debugging 

        .dword      LIT
        .dword      PACCEP
        .dword      CATCH
        .dword      LDUP, LIT, -EPIPE, EQUAL
        .dword      ZBRAN
        .dword      ENDIF7-.-CW
        .dword      BYE     #End of input, no error!
ENDIF7:
        .dword      QERRUR
        .dword      ONEP
# debugging        DC      TDUP, LTYPE, CR 
        .dword      SETSRC
        .dword      INTER
        .dword      OK
        .dword      BRAN
        .dword      QUIT1-.-CW  #AGAIN
        .dword      SEMIS   #Unnecessary, but helpful for decompilation.
#

#  **********
#  *   OK   *
#  **********
#
        .align 3,0
N_OK:
        .dword      2
        .ascii      "OK"
        .align 3,0
OK:
        .dword    DOCOL
        .dword    OK+HEADSIZE
        .dword    0x0
        .dword    QUIT
        .dword    N_OK
        .dword    0

        .dword      STATE
        .dword      FETCH
        .dword      ZEQU
        .dword      ZBRAN
        .dword      OK2-.-CW #IF
        .dword      SKIP
         .dword      3
SB2: .ascii      " OK"
       .align 3,0
        .dword      LIT, SB2
        .dword      LIT, 3
        .dword      LTYPE
        .dword      CR
OK2:
        .dword      SEMIS
#


#  ***************
#  *   OPTIONS   *
#  ***************
#
        .align 3,0
N_OPTIONS:
        .dword      7
        .ascii      "OPTIONS"
        .align 3,0
OPTIONS:
        .dword    DOCOL
        .dword    OPTIONS+HEADSIZE
        .dword    0x0
        .dword    OK
        .dword    N_OPTIONS
        .dword    0

#       Execute option.
        .dword      ARGS, FETCH, CELLP, CELLP, FETCH, LDUP
        .dword      ZBRAN
        .dword      OPT4-.-CW # No options  




#Supposedly on the stack is now a pointer to the option.
        .dword      COUNT

        .dword      LIT, '-', UNEQ
        .dword      ZBRAN
        .dword      OPT3-.-CW
        .dword      LIT, 3, LDUP, LERROR
        .dword      XCODE, STOR, BYE
OPT3:
        .dword      CFET
        .dword      LIT, 0x1F, LAND
        .dword      LLOAD       #Supposedly stack-neutral.
        .dword      ZERO, SWAP # Sign on suppressed.
OPT4:
        .dword      DROP
        .dword      SEMIS   #Unnecessary, but helpful for decompilation.
#


#  *************
#  *   _INIT   *
#  *************
#
        .align 3,0
N_INIT:
        .dword      5
        .ascii      "_INIT"
        .align 3,0
INIT:
        .dword    DOCOL
        .dword    INIT+HEADSIZE
        .dword    0x0
        .dword    OPTIONS
        .dword    N_INIT
        .dword    0

        .dword      SZERO, FETCH, SPSTO
        .dword      ZERO, HANDLER, STOR
        .dword      DECA
        .dword      ONLY, FORTH, DEFIN

        .dword      LIT, 0, BLINI  #Default, open library file read-only!#

        .dword      SEMIS
#


#  ***********
#  *   S>D   *
#  ***********
#
        .align 3,0
N_STOD:
        .dword      3
        .ascii      "S>D"
        .align 3,0
STOD:
        .dword    STOD+HEADSIZE
        .dword    STOD+HEADSIZE
        .dword    0x0
        .dword    INIT
        .dword    N_STOD
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #S1
        SLT     a0, a1, zero
        NEG     a0, a0
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   ABS   *
#  ***********
#
        .align 3,0
N_LABS:
        .dword      3
        .ascii      "ABS"
        .align 3,0
LABS:
        .dword    DOCOL
        .dword    LABS+HEADSIZE
        .dword    0x0
        .dword    STOD
        .dword    N_LABS
        .dword    0

        .dword      LDUP
        .dword      ZLESS
        .dword      ZBRAN
        .dword      PM1-.-CW   #IF
        .dword      NEGATE   #THEN
PM1:
        .dword      SEMIS
#

#  ************
#  *   DABS   *
#  ************
#
        .align 3,0
N_DABS:
        .dword      4
        .ascii      "DABS"
        .align 3,0
DABS:
        .dword    DOCOL
        .dword    DABS+HEADSIZE
        .dword    0x0
        .dword    LABS
        .dword    N_DABS
        .dword    0

        .dword      LDUP
        .dword      ZLESS
        .dword      ZBRAN
        .dword      DPM1-.-CW  #IF
        .dword      DNEGA   #THEN
DPM1:
        .dword      SEMIS
#

#  ***********
#  *   MIN   *
#  ***********
#
        .align 3,0
N_MIN:
        .dword      3
        .ascii      "MIN"
        .align 3,0
MIN:
        .dword    DOCOL
        .dword    MIN+HEADSIZE
        .dword    0x0
        .dword    DABS
        .dword    N_MIN
        .dword    0

        .dword      TDUP
        .dword      GREAT
        .dword      ZBRAN
        .dword      MIN1-.-CW  #IF
        .dword      SWAP    #THEN
MIN1:   .dword      DROP
        .dword      SEMIS
#

#  ***********
#  *   MAX   *
#  ***********
#
        .align 3,0
N_MAX:
        .dword      3
        .ascii      "MAX"
        .align 3,0
MAX:
        .dword    DOCOL
        .dword    MAX+HEADSIZE
        .dword    0x0
        .dword    MIN
        .dword    N_MAX
        .dword    0

        .dword      TDUP
        .dword      LESS
        .dword      ZBRAN
        .dword      MAX1-.-CW  #IF
        .dword      SWAP    #THEN
MAX1:   .dword      DROP
        .dword      SEMIS
#

#  **************
#  *   LSHIFT   *
#  **************
#
        .align 3,0
N_LSHIFT:
        .dword      6
        .ascii      "LSHIFT"
        .align 3,0
LSHIFT:
        .dword    LSHIFT+HEADSIZE
        .dword    LSHIFT+HEADSIZE
        .dword    0x0
        .dword    MAX
        .dword    N_LSHIFT
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        SLL    a0, a0, a2
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **************
#  *   RSHIFT   *
#  **************
#
        .align 3,0
N_RSHIFT:
        .dword      6
        .ascii      "RSHIFT"
        .align 3,0
RSHIFT:
        .dword    RSHIFT+HEADSIZE
        .dword    RSHIFT+HEADSIZE
        .dword    0x0
        .dword    LSHIFT
        .dword    N_RSHIFT
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        SRL     a0, a0, a2
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***************
#  *   ARSHIFT   *
#  ***************
#
        .align 3,0
N_ARSHIFT:
        .dword      7
        .ascii      "ARSHIFT"
        .align 3,0
ARSHIFT:
        .dword    ARSHIFT+HEADSIZE
        .dword    ARSHIFT+HEADSIZE
        .dword    0x0
        .dword    RSHIFT
        .dword    N_ARSHIFT
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
        SRA     a0, a0, a2
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   M*   *
#  **********
#
        .align 3,0
N_MSTAR:
        .dword      2
        .ascii      "M*"
        .align 3,0
MSTAR:
        .dword    MSTAR+HEADSIZE
        .dword    MSTAR+HEADSIZE
        .dword    0x0
        .dword    ARSHIFT
        .dword    N_MSTAR
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        MUL      a1,a3,a2
        MULH     a0,a3,a2
        ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

# High level implementation from FIG-forth 
#  **************
#  *   SM/REM   *
#  **************
#
        .align 3,0
N_MSLAS:
        .dword      6
        .ascii      "SM/REM"
        .align 3,0
MSLAS:
        .dword    DOCOL
        .dword    MSLAS+HEADSIZE
        .dword    0x0
        .dword    MSTAR
        .dword    N_MSLAS
        .dword    0

        .dword      OVER, TOR, TOR, DABS  #   OVER >R >R DABS 
        .dword      RR, LABS, USLAS       #   R ABS UM/MOD 
        .dword      FROMR, RR, LXOR       #   R> R XOR 
        .dword      ZLESS                 #   <0 IF NEGATE THEN 
        .dword      ZBRAN
        .dword      MSLAS1-.-CW
        .dword      NEGATE
MSLAS1:
        .dword      SWAP, FROMR           #   SWAP R> 
        .dword      ZLESS                 #   <0 IF NEGATE THEN 
        .dword      ZBRAN
        .dword      MSLAS2-.-CW
        .dword      NEGATE
MSLAS2:
        .dword      SWAP                  #   SWAP ; 
        .dword      SEMIS


#  **********
#  *   2/   *
#  **********
#
        .align 3,0
N_TWOSL:
        .dword      2
        .ascii      "2/"
        .align 3,0
TWOSL:
        .dword    TWOSL+HEADSIZE
        .dword    TWOSL+HEADSIZE
        .dword    0x0
        .dword    MSLAS
        .dword    N_TWOSL
        .dword    0

        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))
	SRAI a0, a0, 1
	ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  **********
#  *   2*   *
#  **********
#
        .align 3,0
N_TWOST:
        .dword      2
        .ascii      "2*"
        .align 3,0
TWOST:
        .dword    DOCOL
        .dword    TWOST+HEADSIZE
        .dword    0x0
        .dword    TWOSL
        .dword    N_TWOST
        .dword    0

        .dword      TWO, STAR
        .dword      SEMIS
#

#  **********
#  *   1-   *
#  **********
#
        .align 3,0
N_ONEM:
        .dword      2
        .ascii      "1-"
        .align 3,0
ONEM:
        .dword    DOCOL
        .dword    ONEM+HEADSIZE
        .dword    0x0
        .dword    TWOST
        .dword    N_ONEM
        .dword    0

        .dword      ONE, LSUB
        .dword      SEMIS
#

#  **************
#  *   FM/MOD   *
#  **************
#
        .align 3,0
N_FMSLAS:
        .dword      6
        .ascii      "FM/MOD"
        .align 3,0
FMSLAS:
        .dword    DOCOL
        .dword    FMSLAS+HEADSIZE
        .dword    0x0
        .dword    ONEM
        .dword    N_FMSLAS
        .dword    0

        .dword      LDUP, TOR
        .dword      TDUP, LXOR, TOR
        .dword      MSLAS
        .dword      FROMR, ZLESS
        .dword      ZBRAN
        .dword      FMMOD1-.-CW
        .dword      OVER
        .dword      ZBRAN
        .dword      FMMOD1-.-CW
        .dword      ONE, LSUB
        .dword      SWAP, FROMR, PLUS, SWAP
        .dword      BRAN
        .dword      FMMOD2-.-CW
FMMOD1:
        .dword      RDROP
FMMOD2:
        .dword      SEMIS
#

#  *********
#  *   *   *
#  *********
#
        .align 3,0
N_STAR:
        .dword      1
        .ascii      "*"
        .align 3,0
STAR:
        .dword    STAR+HEADSIZE
        .dword    STAR+HEADSIZE
        .dword    0x0
        .dword    FMSLAS
        .dword    N_STAR
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        MUL      a0,a3,a2
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ************
#  *   /MOD   *
#  ************
#
        .align 3,0
N_SLMOD:
        .dword      4
        .ascii      "/MOD"
        .align 3,0
SLMOD:
        .dword    SLMOD+HEADSIZE
        .dword    SLMOD+HEADSIZE
        .dword    0x0
        .dword    STAR
        .dword    N_SLMOD
        .dword    0

        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))
        DIV a1, a3, a2
        REM a0, a3, a2
	ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
	ADDI    sp, sp, -(CW*(1))
        SD      a1, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#
#SUPPRESSED_


#  *********
#  *   /   *
#  *********
#
        .align 3,0
N_SLASH:
        .dword      1
        .ascii      "/"
        .align 3,0
SLASH:
        .dword    DOCOL
        .dword    SLASH+HEADSIZE
        .dword    0x0
        .dword    SLMOD
        .dword    N_SLASH
        .dword    0

        .dword      SLMOD
        .dword      NIP
        .dword      SEMIS
#

#  ***********
#  *   MOD   *
#  ***********
#
        .align 3,0
N_LMOD:
        .dword      3
        .ascii      "MOD"
        .align 3,0
LMOD:
        .dword    DOCOL
        .dword    LMOD+HEADSIZE
        .dword    0x0
        .dword    SLASH
        .dword    N_LMOD
        .dword    0

        .dword      SLMOD
        .dword      DROP
        .dword      SEMIS
#

#  *************
#  *   */MOD   *
#  *************
#
        .align 3,0
N_SSMOD:
        .dword      5
        .ascii      "*/MOD"
        .align 3,0
SSMOD:
        .dword    DOCOL
        .dword    SSMOD+HEADSIZE
        .dword    0x0
        .dword    LMOD
        .dword    N_SSMOD
        .dword    0

        .dword      TOR
        .dword      MSTAR
        .dword      FROMR
        .dword      MSLAS
        .dword      SEMIS
#

#  **********
#  *   */   *
#  **********
#
        .align 3,0
N_SSLA:
        .dword      2
        .ascii      "*/"
        .align 3,0
SSLA:
        .dword    DOCOL
        .dword    SSLA+HEADSIZE
        .dword    0x0
        .dword    SSMOD
        .dword    N_SSLA
        .dword    0

        .dword      SSMOD
        .dword      NIP
        .dword      SEMIS
#

#  ***************
#  *   UDM/MOD   *
#  ***************
#
        .align 3,0
N_MSMOD:
        .dword      7
        .ascii      "UDM/MOD"
        .align 3,0
MSMOD:
        .dword    DOCOL
        .dword    MSMOD+HEADSIZE
        .dword    0x0
        .dword    SSLA
        .dword    N_MSMOD
        .dword    0

#         .dword	DROP, DROP
#         .dword      LDUP, LIT, 15, LAND 
#         .dword      SWAP 
# 	.dword	LIT, 4, RSHIFT
#         .dword	LIT, 0
#         .dword	SEMIS
        .dword      TOR
        .dword      ZERO
        .dword      RR
        .dword      USLAS
        .dword      FROMR
        .dword      SWAP
        .dword      TOR
        .dword      USLAS
        .dword      FROMR
        .dword      SEMIS
#

#  **************
#  *   (LINE)   *
#  **************
#
        .align 3,0
N_PLINE:
        .dword      6
        .ascii      "(LINE)"
        .align 3,0
PLINE:
        .dword    DOCOL
        .dword    PLINE+HEADSIZE
        .dword    0x0
        .dword    MSMOD
        .dword    N_PLINE
        .dword    0

        .dword      TOR
        .dword      LIT,64
        .dword      MSTAR
        .dword      BBUF
        .dword      FMSLAS
        .dword      FROMR # This blocks, so is screens.
        .dword      PLUS
        .dword      BLOCK
        .dword      PLUS
        .dword      LIT,63
        .dword      SEMIS
#

#  **************
#  *   ERRSCR   *
#  **************
#
        .align 3,0
N_ERRSCR:
        .dword      6
        .ascii      "ERRSCR"
        .align 3,0
ERRSCR:
        .dword    DOVAR
        .dword    ERRSCR+HEADSIZE
        .dword    0x0
        .dword    PLINE
        .dword    N_ERRSCR
        .dword    0

        .dword ERRORSCREEN
#

#  ***************
#  *   MESSAGE   *
#  ***************
#
        .align 3,0
N_MESS:
        .dword      7
        .ascii      "MESSAGE"
        .align 3,0
MESS:
        .dword    DOCOL
        .dword    MESS+HEADSIZE
        .dword    0x0
        .dword    ERRSCR
        .dword    N_MESS
        .dword    0

        .dword      LWARN
        .dword      FETCH
        .dword      ZBRAN
        .dword      MESS1-.-CW #IF
        .dword      ERRSCR, FETCH
        .dword      PLINE, ONEP     # Also print the '\n' !
        .dword      ETYPE
        .dword      X
MESS1:                  #THEN
        .dword      DROP
        .dword      SEMIS
#
#
#

#
#

#  ***********
#  *   PC@   *
#  ***********
#
        .align 3,0
N_PCFET:
        .dword      3
        .ascii      "PC@"
        .align 3,0
PCFET:
        .dword    PCFET+HEADSIZE
        .dword    PCFET+HEADSIZE
        .dword    0x0
        .dword    MESS
        .dword    N_PCFET
        .dword    0

# FETCH CHARACTER (BYTE) FROM PORT
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      # PORT ADDR
        LB      a0,0(a1)
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   PC!   *
#  ***********
#
        .align 3,0
N_PCSTO:
        .dword      3
        .ascii      "PC!"
        .align 3,0
PCSTO:
        .dword    PCSTO+HEADSIZE
        .dword    PCSTO+HEADSIZE
        .dword    0x0
        .dword    PCFET
        .dword    N_PCSTO
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #PORT ADDR
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #DATA
        SB            a0, 0(a1)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   PW@   *
#  ***********
#
        .align 3,0
N_PFET:
        .dword      3
        .ascii      "PW@"
        .align 3,0
PFET:
        .dword    PFET+HEADSIZE
        .dword    PFET+HEADSIZE
        .dword    0x0
        .dword    PCSTO
        .dword    N_PFET
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #PORT ADDR
        LH            a0, 0(a1)
        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  ***********
#  *   PW!   *
#  ***********
#
        .align 3,0
N_PSTO:
        .dword      3
        .ascii      "PW!"
        .align 3,0
PSTO:
        .dword    PSTO+HEADSIZE
        .dword    PSTO+HEADSIZE
        .dword    0x0
        .dword    PFET
        .dword    N_PSTO
        .dword    0

        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))      #PORT ADDR
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))      #DATA
        SH            a0, 0(a1)
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0  
#

#  *************
#  *   _PREV   *
#  *************
#
        .align 3,0
N__PREV:
        .dword      5
        .ascii      "_PREV"
        .align 3,0
_PREV:
        .dword    DOVAR
        .dword    _PREV+HEADSIZE
        .dword    0x0
        .dword    PSTO
        .dword    N__PREV
        .dword    0

        .dword      BUF1

#
#

#  *************
#  *   #BUFF   *
#  *************
#
        .align 3,0
N_NOBUF:
        .dword      5
        .ascii      "#BUFF"
        .align 3,0
NOBUF:
        .dword    DOCON
        .dword    NBUF
        .dword    0x0
        .dword    _PREV
        .dword    N_NOBUF
        .dword    0

#

#  ************
#  *   +BUF   *
#  ************
#
        .align 3,0
N_PBUF:
        .dword      4
        .ascii      "+BUF"
        .align 3,0
PBUF:
        .dword    DOCOL
        .dword    PBUF+HEADSIZE
        .dword    0x0
        .dword    NOBUF
        .dword    N_PBUF
        .dword    0

        .dword      BBUF, CELLP, CELLP
        .dword      PLUS,LDUP
        .dword      _LIMIT,EQUAL
        .dword      ZBRAN
        .dword      PBUF1-.-CW
        .dword      DROP,_FIRST
PBUF1:  .dword      LDUP, _PREV, FETCH, LSUB
        .dword      SEMIS
#

#  **************
#  *   UPDATE   *
#  **************
#
        .align 3,0
N_UPDAT:
        .dword      6
        .ascii      "UPDATE"
        .align 3,0
UPDAT:
        .dword    DOCOL
        .dword    UPDAT+HEADSIZE
        .dword    0x0
        .dword    PBUF
        .dword    N_UPDAT
        .dword    0

        .dword      _PREV, FETCH
        .dword      LDUP, CELLP,CELLP
        .dword      SWAP, FETCH
        .dword      LOFFSET,  FETCH, PLUS
        .dword      BLWRITE
        .dword      SEMIS
#

#  *********************
#  *   EMPTY-BUFFERS   *
#  *********************
#
        .align 3,0
N_MTBUF:
        .dword      13
        .ascii      "EMPTY-BUFFERS"
        .align 3,0
MTBUF:
        .dword    DOCOL
        .dword    MTBUF+HEADSIZE
        .dword    0x0
        .dword    UPDAT
        .dword    N_MTBUF
        .dword    0

        .dword      _FIRST
        .dword      _LIMIT,OVER
        .dword      LSUB,LERASE
        .dword      _FIRST, _PREV, STOR
        .dword      SEMIS
        #
#

#  ****************
#  *   (BUFFER)   *
#  ****************
#
        .align 3,0
N_BUFFER:
        .dword      8
        .ascii      "(BUFFER)"
        .align 3,0
BUFFER:
        .dword    DOCOL
        .dword    BUFFER+HEADSIZE
        .dword    0x0
        .dword    MTBUF
        .dword    N_BUFFER
        .dword    0

# Find the buffer, if it is already here.
    .dword      _PREV, FETCH
BUFFER1:
    .dword          TOR, RR, FETCH, OVER, EQUAL
    .dword      ZBRAN
        .dword      BUFFER3-.-CW
    .dword        DROP, FROMR, EXIT
BUFFER3:
    .dword          FROMR
    .dword      PBUF, ZEQU
    .dword      ZBRAN
        .dword      BUFFER1-.-CW
# We are left with the previous buffer.
BUFFER2:
    .dword       PBUF, ZEQU, LIT, 48, QERR  # No free buffers available.
    .dword       LDUP, CELLP, FETCH
    .dword       LIT, -1, GREAT        # Unlocked buffer?
    .dword      ZBRAN
        .dword      BUFFER2-.-CW
    .dword       TOR
# Fill in the house keeping.
    .dword       RR, STOR
    .dword       ZERO, RR, CELLP, STOR

    .dword       RR, _PREV, STOR
    .dword       FROMR
    .dword  SEMIS
#


#  *************
#  *   BLOCK   *
#  *************
#
        .align 3,0
N_BLOCK:
        .dword      5
        .ascii      "BLOCK"
        .align 3,0
BLOCK:
        .dword    DOCOL
        .dword    BLOCK+HEADSIZE
        .dword    0x0
        .dword    BUFFER
        .dword    N_BLOCK
        .dword    0


        .dword      BUFFER
        .dword      LDUP, CELLP, FETCH, ONE, LAND, ZEQU
        .dword      ZBRAN
        .dword      BLOCK1-.-CW
        .dword      LDUP, CELLP, CELLP
        .dword      OVER, FETCH
        .dword      LOFFSET,  FETCH, PLUS
        .dword      BLREAD
        .dword      LDUP, CELLP, ONE, TOGGL
BLOCK1:
        .dword      LDUP, _PREV, STOR
        .dword      CELLP, CELLP
        .dword      SEMIS
#

#  *************
#  *   FLUSH   *
#  *************
#
        .align 3,0
N_FLUSH:
        .dword      5
        .ascii      "FLUSH"
        .align 3,0
FLUSH:
        .dword    DOCOL
        .dword    (MTBUF+HEADSIZE)
        .dword    0x0
        .dword    BLOCK
        .dword    N_FLUSH
        .dword    0


# Unlock all buffers
        .dword      _LIMIT
        .dword      _FIRST, CELLP
        .dword     XDO
        .dword      FLUS2-.-CW
FLUS1:  .dword      ZERO, IDO, STOR
        .dword      LIT,(KBBUF+2*CW)
        .dword      PLOOP
        .dword      (FLUS1-.)
FLUS2:
        .dword      SEMIS
#

#  ************
#  *   SAVE   *
#  ************
#
        .align 3,0
N_SAVE:
        .dword      4
        .ascii      "SAVE"
        .align 3,0
SAVE:
        .dword    DOCOL
        .dword    SAVE+HEADSIZE
        .dword    0x0
        .dword    FLUSH
        .dword    N_SAVE
        .dword    0

        .dword      FROMR
        .dword      SRC, TFET
        .dword      PP, FETCH
        .dword      TOR, TOR, TOR
        .dword      TOR
        .dword SEMIS
#

#  ***************
#  *   RESTORE   *
#  ***************
#
        .align 3,0
N_RESTO:
        .dword      7
        .ascii      "RESTORE"
        .align 3,0
RESTO:
        .dword    DOCOL
        .dword    RESTO+HEADSIZE
        .dword    0x0
        .dword    SAVE
        .dword    N_RESTO
        .dword    0

        .dword      FROMR
        .dword      FROMR, FROMR, FROMR
        .dword      PP, STOR
        .dword      SRC, TSTOR
        .dword      TOR
        .dword SEMIS
#

# 
#

# 
#

#  ************
#  *   LOCK   *
#  ************
#
        .align 3,0
N_LLOCK:
        .dword      4
        .ascii      "LOCK"
        .align 3,0
LLOCK:
        .dword    DOCOL
        .dword    LLOCK+HEADSIZE
        .dword    0x0
        .dword    RESTO
        .dword    N_LLOCK
        .dword    0

        .dword      BUFFER
        .dword      CELLP
        .dword      LIT, -2, SWAP, PSTOR
        .dword      SEMIS
#

#  **************
#  *   UNLOCK   *
#  **************
#
        .align 3,0
N_LUNLOCK:
        .dword      6
        .ascii      "UNLOCK"
        .align 3,0
LUNLOCK:
        .dword    DOCOL
        .dword    LUNLOCK+HEADSIZE
        .dword    0x0
        .dword    LLOCK
        .dword    N_LUNLOCK
        .dword    0

        .dword      BUFFER
        .dword      CELLP
        .dword      TWO, SWAP, PSTOR
        .dword      SEMIS
#

#  ************
#  *   LOAD   *
#  ************
#
        .align 3,0
N_LLOAD:
        .dword      4
        .ascii      "LOAD"
        .align 3,0
LLOAD:
        .dword    DOCOL
        .dword    LLOAD+HEADSIZE
        .dword    0x0
        .dword    LUNLOCK
        .dword    N_LLOAD
        .dword    0

        .dword      SAVE
        .dword      TOR
        .dword      RR, LIT, LLOCK, CATCH  # Be careful about always unlocking.
        .dword      LDUP, ZEQU
        .dword      ZBRAN
        .dword      LOAD5-.-CW
        .dword      DROP
        .dword      RR, BLOCK     # Can't fail anymore. 
        .dword      BBUF
        .dword      SETSRC
        .dword      LIT, INTER, CATCH
        .dword      RR, LUNLOCK
LOAD5:
        .dword      RDROP
        .dword      RESTO
        .dword      THROW
        .dword      SEMIS
#

#  ************
#  *   THRU   *
#  ************
#
        .align 3,0
N_THRU:
        .dword      4
        .ascii      "THRU"
        .align 3,0
THRU:
        .dword    DOCOL
        .dword    THRU+HEADSIZE
        .dword    0x0
        .dword    LLOAD
        .dword    N_THRU
        .dword    0

        .dword      ONEP, SWAP
        .dword     XDO
        .dword      THRU2-.-CW
THRU1:
        .dword      IDO, LLOAD
        .dword     ONE, XPLOOP, ZBRAN
        .dword      THRU1-.-CW
        .dword      UNLOOP
THRU2:
        .dword      SEMIS
#
#

#  ***********
#  *   BLK   *
#  ***********
#
        .align 3,0
N_BLK:
        .dword      3
        .ascii      "BLK"
        .align 3,0
BLK:
        .dword    DOCOL
        .dword    BLK+HEADSIZE
        .dword    0x0
        .dword    THRU
        .dword    N_BLK
        .dword    0

        .dword      PP, FETCH
        .dword      _FIRST, _LIMIT, WITHIN
        .dword      SRC, TFET, LSUB
        .dword      LIT, 1024, EQUAL, LAND
        .dword      ZBRAN
        .dword      BLK1-.-CW
        .dword      SRC, FETCH, TWO, LCELLS, LSUB, FETCH
        .dword      BRAN
        .dword      BLK2-.-CW
BLK1:
        .dword      ZERO
BLK2:
        .dword      PBLK, STOR
        .dword      PBLK
        .dword      SEMIS
#

#



#  ***********
#  *   XOS   *
#  ***********
#
        .align 3,0
N_XOS:
        .dword      3
        .ascii      "XOS"
        .align 3,0
XOS:
        .dword    XOS+HEADSIZE
        .dword    XOS+HEADSIZE
        .dword    0x0
        .dword    BLK
        .dword    N_XOS
        .dword    0

#


        ADDI    fp, fp, -(CW*(1))
        SD      a7, 0(fp) 

        LD      a7, 0(sp)
        ADDI    sp, sp, (CW*(1))       # Function number              
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))       # Third parameter, or dummy    
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))       # Second parameter, or dummy   
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))       # First parameter.             
        ECALL          #  Generic call on LINUX       
        LD      a7, 0(fp) 
        ADDI    fp, fp, (CW*(1))

        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0       # Positive means okay. Negative means -errno.
#
#
 # 




#


#  ************
#  *   XOS5   *
#  ************
#
        .align 3,0
N_XOS5:
        .dword      4
        .ascii      "XOS5"
        .align 3,0
XOS5:
        .dword    XOS5+HEADSIZE
        .dword    XOS5+HEADSIZE
        .dword    0x0
        .dword    XOS
        .dword    N_XOS5
        .dword    0

        ADDI    fp, fp, -(CW*(1))
        SD      a7, 0(fp) 

        LD      a7, 0(sp)
        ADDI    sp, sp, (CW*(1))        # Function number            
        LD      a4, 0(sp)
        ADDI    sp, sp, (CW*(1))        # 5th parameter, or dummy
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))        # 4th parameter, or dummy
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))        # Third parameter, or dummy
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))        # Second parameter, or dummy
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))        # First parameter.
        ECALL          # Generic call on LINUX      
        LD      a7, 0(fp) 
        ADDI    fp, fp, (CW*(1))

        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0            # Positive means okay. Negative means -errno.
#


#

#  ************
#  *   XOS7   *
#  ************
#
        .align 3,0
N_XOS7:
        .dword      4
        .ascii      "XOS7"
        .align 3,0
XOS7:
        .dword    XOS7+HEADSIZE
        .dword    XOS7+HEADSIZE
        .dword    0x0
        .dword    XOS5
        .dword    N_XOS7
        .dword    0

        ADDI    fp, fp, -(CW*(1))
        SD      a7, 0(fp) 

        LD      a7, 0(sp)
        ADDI    sp, sp, (CW*(1))        # Function number            
        LD      a6, 0(sp)
        ADDI    sp, sp, (CW*(1))        # 7th parameter, or dummy
        LD      a5, 0(sp)
        ADDI    sp, sp, (CW*(1))        # 6th parameter, or dummy
        LD      a4, 0(sp)
        ADDI    sp, sp, (CW*(1))        # 5th parameter, or dummy
        LD      a3, 0(sp)
        ADDI    sp, sp, (CW*(1))        # 4th parameter, or dummy
        LD      a2, 0(sp)
        ADDI    sp, sp, (CW*(1))        # Third parameter, or dummy
        LD      a1, 0(sp)
        ADDI    sp, sp, (CW*(1))        # Second parameter, or dummy
        LD      a0, 0(sp)
        ADDI    sp, sp, (CW*(1))        # First parameter.
        ECALL           # Generic call on LINUX      
        LD      a7, 0(fp) 
        ADDI    fp, fp, (CW*(1))

        ADDI    sp, sp, -(CW*(1))
        SD      a0, 0(sp) 
        LD      a0, 0(a7)
        ADDI    a7, a7, (CW*(1))
         LD      t0,0(a0)
         JALR    t0            # Positive means okay. Negative means -errno.

#

#
 # 


#  **********
#  *   MS   *
#  **********
#
        .align 3,0
N_MS:
        .dword      2
        .ascii      "MS"
        .align 3,0
MS:
        .dword    DOCOL
        .dword    MS+HEADSIZE
        .dword    0x0
        .dword    XOS7
        .dword    N_MS
        .dword    0

# Use select because nanosleep has a 20 MS resolution. 
# 

        .dword      LIT, 1000000, LIT, 1000000000, SSMOD # (ms -- ns s) 
        .dword      LIT, TV, TSTOR
        .dword      ZERO, ZERO
        .dword      ZERO, ZERO
        .dword      LIT, TV
        .dword      ZERO, ZERO
        .dword      LIT, select, XOS7
# 

        .dword      QERRUR
        .dword      SEMIS
TV:     .space    (CW*(2))    # A time_t and a long, both cells.
#
# 

        #
#------------------------------------
#       CONSOLE RELEATED I/O         
#------------------------------------







#  *****************
#  *   RW-BUFFER   *
#  *****************
#
        .align 3,0
N_RWBUF:
        .dword      9
        .ascii      "RW-BUFFER"
        .align 3,0
RWBUF:
        .dword    DOVAR
        .dword    RWBUF+HEADSIZE
        .dword    0x0
        .dword    MS
        .dword    N_RWBUF
        .dword    0

        .space    0x1000
# 

#


#  ***********
#  *   ZEN   *
#  ***********
#
        .align 3,0
N_ZEN:
        .dword      3
        .ascii      "ZEN"
        .align 3,0
ZEN:
        .dword    DOCOL
        .dword    ZEN+HEADSIZE
        .dword    0x0
        .dword    RWBUF
        .dword    N_ZEN
        .dword    0

        .dword      RWBUF, SSTOR
        .dword      ZERO, RWBUF, CHAPP
        .dword      RWBUF, CELLP

        .dword      SEMIS
# 



#


#  *****************
#  *   OPEN-FILE   *
#  *****************
#
        .align 3,0
N_OFILE:
        .dword      9
        .ascii      "OPEN-FILE"
        .align 3,0
OFILE:
        .dword    DOCOL
        .dword    OFILE+HEADSIZE
        .dword    0x0
        .dword    ZEN
        .dword    N_OFILE
        .dword    0


      .dword      TOR, ZEN
       .dword       LIT, AT_FDCWD, SWAP, FROMR
       .dword      LIT, openat, XOS
       .dword      LDUP, ZERO, MIN
       .dword      SEMIS 
#


#


#  ******************
#  *   CLOSE-FILE   *
#  ******************
#
        .align 3,0
N_CFILE:
        .dword      10
        .ascii      "CLOSE-FILE"
        .align 3,0
CFILE:
        .dword    DOCOL
        .dword    CFILE+HEADSIZE
        .dword    0x0
        .dword    OFILE
        .dword    N_CFILE
        .dword    0

        .dword      X, X
        .dword      LIT, close, XOS
        .dword      SEMIS
#


#



#  *******************
#  *   CREATE-FILE   *
#  *******************
#
        .align 3,0
N_CREATEF:
        .dword      11
        .ascii      "CREATE-FILE"
        .align 3,0
CREATEF:
        .dword    DOCOL
        .dword    CREATEF+HEADSIZE
        .dword    0x0
        .dword    CFILE
        .dword    N_CREATEF
        .dword    0


        .dword      TOR, TDUP, DFILE, DROP
      .dword      ZEN, LIT, AT_FDCWD, SWAP
        .dword     LIT, O_CREAT + O_TRUNC + O_WRONLY
        .dword      FROMR
        .dword      ZERO      # fifth wheel
        .dword      LIT, openat, XOS5
        .dword      LDUP, ZERO, MIN
        .dword      SEMIS 

#


#


#  *******************
#  *   DELETE-FILE   *
#  *******************
#
        .align 3,0
N_DFILE:
        .dword      11
        .ascii      "DELETE-FILE"
        .align 3,0
DFILE:
        .dword    DOCOL
        .dword    DFILE+HEADSIZE
        .dword    0x0
        .dword    CREATEF
        .dword    N_DFILE
        .dword    0


      .dword      ZEN
     .dword     LIT, AT_FDCWD, SWAP, ZERO
        .dword      LIT, unlinkat, XOS
        .dword      SEMIS 
#


#


#  *****************
#  *   READ-FILE   *
#  *****************
#
        .align 3,0
N_RFILE:
        .dword      9
        .ascii      "READ-FILE"
        .align 3,0
RFILE:
        .dword    DOCOL
        .dword    RFILE+HEADSIZE
        .dword    0x0
        .dword    DFILE
        .dword    N_RFILE
        .dword    0

        .dword      ROT, ROT
        .dword      LIT, read, XOS
        .dword      LDUP, ZERO, MAX, SWAP, ZERO, MIN
        .dword      SEMIS
# 
#

#

#  *********************
#  *   FILE-POSITION   *
#  *********************
#
        .align 3,0
N_FPFILE:
        .dword      13
        .ascii      "FILE-POSITION"
        .align 3,0
FPFILE:
        .dword    DOCOL
        .dword    FPFILE+HEADSIZE
        .dword    0x0
        .dword    RFILE
        .dword    N_FPFILE
        .dword    0

        .dword      ZERO
        .dword      LIT, SEEK_CUR
        .dword      LIT, lseek, XOS
        .dword      ZERO     #Double precision return result
        .dword      OVER, LDUP
        .dword      NEGATE, LIT, 256, ULESS, LAND # Heuristic 
        .dword      SEMIS
 # 

#


#  ***********************
#  *   REPOSITION-FILE   *
#  ***********************
#
        .align 3,0
N_PFILE:
        .dword      15
        .ascii      "REPOSITION-FILE"
        .align 3,0
PFILE:
        .dword    DOCOL
        .dword    PFILE+HEADSIZE
        .dword    0x0
        .dword    FPFILE
        .dword    N_PFILE
        .dword    0

        .dword      TOR, DROP, FROMR   #Convert to single precision.
        .dword      SWAP
        .dword      LIT, SEEK_SET
        .dword      LIT, lseek, XOS
        .dword      ZERO, MIN
        .dword      SEMIS
 # 



#


#  ******************
#  *   WRITE-FILE   *
#  ******************
#
        .align 3,0
N_WFILE:
        .dword      10
        .ascii      "WRITE-FILE"
        .align 3,0
WFILE:
        .dword    DOCOL
        .dword    WFILE+HEADSIZE
        .dword    0x0
        .dword    PFILE
        .dword    N_WFILE
        .dword    0

        .dword      ROT, ROT
        .dword      LIT, write, XOS
        .dword      ZERO, MIN
        .dword      SEMIS
# 



#


#  ****************
#  *   GET-FILE   *
#  ****************
#
        .align 3,0
N_GETFILE:
        .dword      8
        .ascii      "GET-FILE"
        .align 3,0
GETFILE:
        .dword    DOCOL
        .dword    GETFILE+HEADSIZE
        .dword    0x0
        .dword    WFILE
        .dword    N_GETFILE
        .dword    0

     .dword LALIGN
        .dword      TDUP, SCOMMA, DROP
        .dword      LIT, 'F'+(0x100*('i'+0x100*('L'+0x100*'e'))), COMMA #Magic number.
        .dword      ZERO, OFILE, THROW, TOR
        .dword      HERE, LDUP
        .dword      LEM, LIT, 6, SLASH, LDUP, ALLOT
        .dword      LIT, 1000, LSUB
        .dword      RR, RFILE, THROW
        .dword      FROMR, CFILE, THROW
        .dword      TDUP, PLUS, LDP, STOR        # No allocation if it fails.
        .dword      SEMIS
#

#  ****************
#  *   PUT-FILE   *
#  ****************
#
        .align 3,0
N_PUTFILE:
        .dword      8
        .ascii      "PUT-FILE"
        .align 3,0
PUTFILE:
        .dword    DOCOL
        .dword    PUTFILE+HEADSIZE
        .dword    0x0
        .dword    GETFILE
        .dword    N_PUTFILE
        .dword    0

        .dword      LIT, 0755
        .dword      CREATEF, THROW
        .dword      LDUP, TOR
        .dword      WFILE, THROW
        .dword      FROMR, CFILE, THROW
        .dword      SEMIS
#

#  ****************
#  *   INCLUDED   *
#  ****************
#
        .align 3,0
N_INCLUDED:
        .dword      8
        .ascii      "INCLUDED"
        .align 3,0
INCLUDED:
        .dword    DOCOL
        .dword    INCLUDED+HEADSIZE
        .dword    0x0
        .dword    PUTFILE
        .dword    N_INCLUDED
        .dword    0

        .dword      HERE, TOR
        .dword      LIT, GETFILE, CATCH
        .dword      LDUP
        .dword      ZBRAN
        .dword      INCLUD1-.-CW
        .dword      FROMR, LDP, STOR
        .dword      THROW
        .dword      BRAN
        .dword      INCLUD2-.-CW
INCLUD1:
        .dword      RDROP, DROP
INCLUD2:
        .dword      EVALUATE
        .dword      SEMIS
#

#  ***************
#  *   INCLUDE   *
#  ***************
#
        .align 3,0
N_LINCLUD:
        .dword      7
        .ascii      "INCLUDE"
        .align 3,0
LINCLUD:
        .dword    DOCOL
        .dword    LINCLUD+HEADSIZE
        .dword    0x0
        .dword    INCLUDED
        .dword    N_LINCLUD
        .dword    0

        .dword      LNAME, INCLUDED
        .dword      SEMIS
#

#  ******************
#  *   REFILL-TIB   *
#  ******************
#
        .align 3,0
N_REFTIB:
        .dword      10
        .ascii      "REFILL-TIB"
        .align 3,0
REFTIB:
        .dword    DOCOL
        .dword    REFTIB+HEADSIZE
        .dword    0x0
        .dword    LINCLUD
        .dword    N_REFTIB
        .dword    0

        .dword      REMAIND, FETCH, TOR
        .dword      TIB, FETCH, RR, PLUS
        .dword      LIT, TIBSIZE, RR, LSUB
        .dword      ZERO
        .dword      RFILE
        .dword      QERRUR
        .dword      LDUP, ZEQU, LIT, -EPIPE, LAND # Presumably end of pipe.
        .dword      QERRUR
        .dword      TIB, FETCH, SWAP, FROMR, PLUS
        .dword      REMAIND, TSTOR
        .dword      SEMIS
#
#

#

# All I/O goes through the modern device (Unix-standard)
# device I/O. Code fields are filled in during bootup. 
# Lower case labels starting with "c_.." are c-supplied facilities.

#  **************
#  *   ACCEPT   *
#  **************
#
        .align 3,0
N_ACCEP:
        .dword      6
        .ascii      "ACCEPT"
        .align 3,0
ACCEP:
        .dword    DOCOL
        .dword    ACCEP+HEADSIZE
        .dword    0x0
        .dword    REFTIB
        .dword    N_ACCEP
        .dword    0

        .dword      PACCEP
        .dword      TSWAP, ROT, MIN
        .dword      LDUP, TOR, LMOVE, FROMR
        .dword      SEMIS
#

#  ****************
#  *   (ACCEPT)   *
#  ****************
#
        .align 3,0
N_PACCEP:
        .dword      8
        .ascii      "(ACCEPT)"
        .align 3,0
PACCEP:
        .dword    DOCOL
        .dword    PACCEP+HEADSIZE
        .dword    0x0
        .dword    ACCEP
        .dword    N_PACCEP
        .dword    0

PACCEP2:
        .dword      REMAIND
        .dword      TFET
        .dword      LIT, ALF, SINDEX
        .dword      ZEQU
        .dword      ZBRAN
        .dword      PACCEP1-.-CW
        .dword      REMAIND, TFET
        .dword      TIB, FETCH
        .dword      SWAP, LMOVE
        .dword      TIB, FETCH
        .dword      REMAIND, CELLP, STOR
        .dword      REFTIB
        .dword      BRAN
        .dword      PACCEP2-.-CW
PACCEP1:
        .dword      REMAIND, TFET
        .dword      LIT, ALF, SSLAS
        .dword      TSWAP, REMAIND, TSTOR
        .dword      SEMIS


#  ***********
#  *   KEY   *
#  ***********
#
        .align 3,0
N_KEY:
        .dword      3
        .ascii      "KEY"
        .align 3,0
KEY:
        .dword    DOCOL
        .dword    KEY+HEADSIZE
        .dword    0x0
        .dword    PACCEP
        .dword    N_KEY
        .dword    0

        .dword      ONE, LIT, RAWIO, STTERM # Do not wait for CR
        .dword      ZERO, SPFET     # Expect single key on stack
        .dword      ZERO            # Standard input
        .dword      SWAP, ONE       # Buffer and length
        .dword      LIT, read, XOS
        .dword      LDUP, QERRUR
        .dword      ZEQU, LIT, -EPIPE, LAND, QERRUR # Presumably end of pipe.
        .dword      ONE, LIT, RAWIO, STTERM       # Toggle back
        .dword      SEMIS
#

#  ************
#  *   KEY?   *
#  ************
#
        .align 3,0
N_KEYQ:
        .dword      4
        .ascii      "KEY?"
        .align 3,0
KEYQ:
        .dword    DOCOL
        .dword    KEYQ+HEADSIZE
        .dword    0x0
        .dword    KEY
        .dword    N_KEYQ
        .dword    0

        .dword      ZERO, LIT, RAWIO, STTERM
        .dword      ZERO, ZERO     # Time out of zero. See MS.
        .dword      LIT, TV, TSTOR
        .dword      ONE # Waiting for standard in.
        .dword      LIT, RFDS, LSTOR
        .dword      ONE     # # of highest channel to watch.
        .dword      LIT, RFDS
        .dword      ZERO, ZERO  # Not output, not events.
        .dword      LIT, TV, ZERO, ZERO, LIT, select, XOS7
        
        .dword      LDUP,  QERRUR
        .dword      NEGATE        # 1 -> flag.
        .dword      ZERO, LIT, RAWIO, STTERM
        .dword      SEMIS

        .align 3,0
RFDS:   .space    (CW*(32))     #32 channels maximum.

#

#  ************
#  *   TYPE   *
#  ************
#
        .align 3,0
N_LTYPE:
        .dword      4
        .ascii      "TYPE"
        .align 3,0
LTYPE:
        .dword    DOCOL
        .dword    LTYPE+HEADSIZE
        .dword    0x0
        .dword    KEYQ
        .dword    N_LTYPE
        .dword    0

        .dword      LDUP, LOUT, PSTOR
        .dword      ONE, WFILE, THROW
        .dword      SEMIS
#

#  *************
#  *   ETYPE   *
#  *************
#
        .align 3,0
N_ETYPE:
        .dword      5
        .ascii      "ETYPE"
        .align 3,0
ETYPE:
        .dword    DOCOL
        .dword    ETYPE+HEADSIZE
        .dword    0x0
        .dword    LTYPE
        .dword    N_ETYPE
        .dword    0

        .dword      TWO, WFILE, THROW
        .dword      SEMIS
#

#  ************
#  *   EMIT   *
#  ************
#
        .align 3,0
N_EMIT:
        .dword      4
        .ascii      "EMIT"
        .align 3,0
EMIT:
        .dword    DOCOL
        .dword    EMIT+HEADSIZE
        .dword    0x0
        .dword    ETYPE
        .dword    N_EMIT
        .dword    0

        .dword      SPFET, ONE, LTYPE
        .dword      DROP
        .dword      SEMIS
#


#  **************
#  *   TERMIO   *
#  **************
#
        .align 3,0
N_TERMIO:
        .dword      6
        .ascii      "TERMIO"
        .align 3,0
TERMIO:
        .dword    DOVAR
        .dword    TERMIO+HEADSIZE
        .dword    0x0
        .dword    EMIT
        .dword    N_TERMIO
        .dword    0

        .space    SIZE_TERMIO     # Filled in during cold boot.
#

#  ****************
#  *   SET-TERM   *
#  ****************
#
        .align 3,0
N_STTERM:
        .dword      8
        .ascii      "SET-TERM"
        .align 3,0
STTERM:
        .dword    DOCOL
        .dword    STTERM+HEADSIZE
        .dword    0x0
        .dword    TERMIO
        .dword    N_STTERM
        .dword    0

SIZEL   = 4     # Also for 64 bits. 
        .dword      LIT, (TERMIO+HEADSIZE)+SIZEL*3
        .dword      SWAP, TOGGL
        .dword      LIT, (TERMIO+HEADSIZE)+SIZEL*4+7
        .dword      CSTOR
        .dword      ZERO, LIT, TCSETS, TERMIO
        .dword      LIT, ioctl, XOS, QERRUR
        .dword      SEMIS
#
# 






#

        #
#------------------------------------
#       SYSTEM DEPENDANT DISK I/O
#------------------------------------

#  ******************
#  *   DISK-ERROR   *
#  ******************
#
        .align 3,0
N_DERR:
        .dword      10
        .ascii      "DISK-ERROR"
        .align 3,0
DERR:
        .dword    DOVAR
        .dword    DERR+HEADSIZE
        .dword    0x0
        .dword    STTERM
        .dword    N_DERR
        .dword    0

        .dword      -1
#


#  ******************
#  *   BLOCK-FILE   *
#  ******************
#
        .align 3,0
N_BLFL:
        .dword      10
        .ascii      "BLOCK-FILE"
        .align 3,0
BLFL:
        .dword    DOVAR
        .dword    BLFL+HEADSIZE
        .dword    0x0
        .dword    DERR
        .dword    N_BLFL
        .dword    0

        .dword      9
        .ascii      "forth.lab"
        .space    252-9               # Allow for some path

#  ********************
#  *   BLOCK-HANDLE   *
#  ********************
#
        .align 3,0
N_BHAN:
        .dword      12
        .ascii      "BLOCK-HANDLE"
        .align 3,0
BHAN:
        .dword    DOVAR
        .dword    BHAN+HEADSIZE
        .dword    0x0
        .dword    BLFL
        .dword    N_BHAN
        .dword    0

        .dword      -1
#

#  *******************
#  *   ?DISK-ERROR   *
#  *******************
#
        .align 3,0
N_QDSKER:
        .dword      11
        .ascii      "?DISK-ERROR"
        .align 3,0
QDSKER:
        .dword    DOCOL
        .dword    QDSKER+HEADSIZE
        .dword    0x0
        .dword    BHAN
        .dword    N_QDSKER
        .dword    0

        .dword      LIT, 8, QERR
        .dword      SEMIS
#
#


#  ******************
#  *   BLOCK-INIT   *
#  ******************
#
        .align 3,0
N_BLINI:
        .dword      10
        .ascii      "BLOCK-INIT"
        .align 3,0
BLINI:
        .dword    DOCOL
        .dword    BLINI+HEADSIZE
        .dword    0x0
        .dword    QDSKER
        .dword    N_BLINI
        .dword    0

        .dword      BLFL, SFET, DTRAI
        .dword      ROT
        .dword      OFILE
        .dword      ZEQU, NEGATE  # 0 if disk problems, 1 if not.
        .dword      LWARN, FETCH, MIN # AND but WARNING is 0/1.
        .dword      LWARN, STOR
        .dword      BHAN, STOR
        .dword      MTBUF
        .dword      SEMIS
#

#  ******************
#  *   BLOCK-EXIT   *
#  ******************
#
        .align 3,0
N_BLEXI:
        .dword      10
        .ascii      "BLOCK-EXIT"
        .align 3,0
BLEXI:
        .dword    DOCOL
        .dword    BLEXI+HEADSIZE
        .dword    0x0
        .dword    BLINI
        .dword    N_BLEXI
        .dword    0

        .dword      FLUSH
        .dword      BHAN, FETCH
        .dword      CFILE
        .dword      ZERO, LWARN, STOR
        .dword      LIT, -1, BHAN, STOR    #Regardless of close errors.
        .dword      QDSKER
        .dword      SEMIS
#

#  **************
#  *   BLSEEK   *
#  **************
#
        .align 3,0
N_BLSEEK:
        .dword      6
        .ascii      "BLSEEK"
        .align 3,0
BLSEEK:
        .dword    DOCOL
        .dword    BLSEEK+HEADSIZE
        .dword    0x0
        .dword    BLEXI
        .dword    N_BLSEEK
        .dword    0

        .dword      BBUF
        .dword      USTAR
        .dword      BHAN, FETCH
        .dword      PFILE           # Disk position.
        .dword      QDSKER
        .dword      SEMIS
#

#  ******************
#  *   BLOCK-READ   *
#  ******************
#
        .align 3,0
N_BLREAD:
        .dword      10
        .ascii      "BLOCK-READ"
        .align 3,0
BLREAD:
        .dword    DOCOL
        .dword    BLREAD+HEADSIZE
        .dword    0x0
        .dword    BLSEEK
        .dword    N_BLREAD
        .dword    0

        .dword      BLSEEK # That's done
        .dword      BBUF
        .dword      BHAN, FETCH
        .dword      RFILE           # Disk read 
        .dword      SWAP, BBUF, UNEQ
        .dword      LOR, QDSKER
        .dword      SEMIS
#

#  *******************
#  *   BLOCK-WRITE   *
#  *******************
#
        .align 3,0
N_BLWRITE:
        .dword      11
        .ascii      "BLOCK-WRITE"
        .align 3,0
BLWRITE:
        .dword    DOCOL
        .dword    BLWRITE+HEADSIZE
        .dword    0x0
        .dword    BLREAD
        .dword    N_BLWRITE
        .dword    0

        .dword      BLSEEK # That's done
        .dword      BBUF
        .dword      BHAN, FETCH
        .dword      WFILE           # Disk write
        .dword      QDSKER
        .dword      SEMIS
#
#

#  *************
#  *   SHELL   *
#  *************
#
        .align 3,0
N_SHELL:
        .dword      5
        .ascii      "SHELL"
        .align 3,0
SHELL:
        .dword    DOVAR
        .dword    SHELL+HEADSIZE
        .dword    0x0
        .dword    BLWRITE
        .dword    N_SHELL
        .dword    0

                 .dword      7
        .ascii      "/bin/sh"

        .space    252-9               # Allow for some path

#


#  **************
#  *   SYSTEM   *
#  **************
#
        .align 3,0
N_SYSTEM:
        .dword      6
        .ascii      "SYSTEM"
        .align 3,0
SYSTEM:
        .dword    DOCOL
        .dword    SYSTEM+HEADSIZE
        .dword    0x0
        .dword    SHELL
        .dword    N_SYSTEM
        .dword    0

          .dword      LIT, COMBUF, SSTOR, ZERO, LIT, COMBUF, CHAPP
          .dword      ZERO, SHELL, CHAPP, LIT, -1, SHELL, PSTOR

         .dword      ZERO, ZERO, ZERO, ZERO, ZERO, LIT, clone, XOS5
          .dword      LDUP, QERRUR
          .dword      LDUP, ZEQU
          .dword      ZBRAN
        .dword      SYSTEM1-.-CW
        .dword      SHELL, CELLP, LIT, SYSARGS
#  Get environment vector
      .dword      ARGS, FETCH, SFET, ONEP, LCELLS, PLUS
        .dword      LIT, execve, XOS
        .dword      LIT, 49, LERROR, BYE   # Returned, this is always wrong.
SYSTEM1:
# int waitid(idtype_t idtype, id_t id, siginfo_t * infop , int options ) 

        .dword      LDUP   # pid of child
        .dword      LIT, P_PID, SWAP
        .dword      LIT, SIGINFOT
        .dword      LIT, WEXITED, ZERO
        .dword      LIT, waitid, XOS5
        .dword      LDUP
        .dword      LIT, EINTR, EQUAL
        .dword      BRAN
        .dword      SYSTEM2-.-CW
        .dword      DROP
        .dword      ZBRAN
        .dword      SYSTEM1-.-CW
SYSTEM2:
        .dword      QERRUR
        .dword      TDROP
        .dword      SEMIS
#
        .align 3,0
SIGINFOT: .space    CW*9
SYSARG2: .ascii     "-c"
         .byte      0
        .align 3,0
COMBUF:  .space    CW+256  #One cell for high level string manipulation.
        .align 3,0
SYSARGS:.dword   (SHELL+HEADSIZE)+CW, SYSARG2, COMBUF + CW, 0
#
 # 



#



#


#

#  *********
#  *   '   *
#  *********
#
        .align 3,0
N_ITICK:
        .dword      1
        .ascii      "'"
        .align 3,0
ITICK:
        .dword    DOCOL 
        .dword    ITICK+HEADSIZE
        .dword    0x0
        .dword    SYSTEM
        .dword    N_ITICK
        .dword    0

        .dword      LNAME, PRESENT
        .dword      LDUP, ZEQU
        .dword      LIT, 11, QERR
        .dword      SEMIS
#

#  ***********
#  *   [']   *
#  ***********
#
        .align 3,0
N_BTICK:
        .dword      3
        .ascii      "[']"
        .align 3,0
BTICK:
        .dword    DOCOL
        .dword    (TICK+HEADSIZE)
        .dword    B_IMMED
        .dword    ITICK
        .dword    N_BTICK
        .dword    0

#

#  ******************
#  *   FORGET-VOC   *
#  ******************
#
        .align 3,0
N_FORGV:
        .dword      10
        .ascii      "FORGET-VOC"
        .align 3,0
FORGV:
        .dword    DOCOL
        .dword    FORGV+HEADSIZE
        .dword    0x0
        .dword    BTICK
        .dword    N_FORGV
        .dword    0

        .dword      TDUP
        .dword      SWAP
        .dword      ULESS
        .dword      ZBRAN
        .dword      FORGV1-.-CW
#  Forget part of contents.
        .dword      SWAP
        .dword      TOR
        .dword      TWID
FORGV0:
        .dword      LDUP
        .dword      LDUP
FORGV3:
        .dword      TLFA,FETCH
        .dword      LDUP
        .dword      RR
        .dword      ULESS
        .dword      ZBRAN
        .dword      FORGV3-.-CW
        .dword      SWAP            #  Not to be forgotten entry found.
        .dword      TLFA
        .dword      STOR           #  Short other entries out.
        .dword      TLFA,FETCH
        .dword      LDUP
        .dword      ZEQU
        .dword      ZBRAN
        .dword      FORGV0-.-CW        #  Repeat until end of wordlist.
        .dword      DROP
        .dword      FROMR
        .dword      BRAN
        .dword      FORGV2-.-CW
FORGV1:
#       { Vocabulary itself is also forgotten.}
        .dword      TVFA
        .dword      FETCH     # Unlink by linking next namespace.
        .dword      VOCL
        .dword      STOR
        .dword      ONLY, FORTH
        .dword      DEFIN
FORGV2: .dword      SEMIS
#

#  **************
#  *   FORGET   *
#  **************
#
        .align 3,0
N_FORG:
        .dword      6
        .ascii      "FORGET"
        .align 3,0
FORG:
        .dword    DOCOL
        .dword    FORG+HEADSIZE
        .dword    0x0
        .dword    FORGV
        .dword    N_FORG
        .dword    0

        .dword      TICK
        .dword      LDUP
        .dword      FENCE
        .dword      FETCH
        .dword      LESS
        .dword      LIT, 21, QERR
        .dword      LIT,FORGV
        .dword      FORV
        .dword      TNFA, FETCH, LDP, STOR
        .dword      SEMIS
#

#  *************
#  *   (BACK   *
#  *************
#
        .align 3,0
N_PBACK:
        .dword      5
        .ascii      "(BACK"
        .align 3,0
PBACK:
        .dword    DOCOL
        .dword    PBACK+HEADSIZE
        .dword    0x0
        .dword    FORG
        .dword    N_PBACK
        .dword    0

        .dword      HERE
        .dword      SEMIS
#

#  *************
#  *   BACK)   *
#  *************
#
        .align 3,0
N_BACKP:
        .dword      5
        .ascii      "BACK)"
        .align 3,0
BACKP:
        .dword    DOCOL
        .dword    BACKP+HEADSIZE
        .dword    0x0
        .dword    PBACK
        .dword    N_BACKP
        .dword    0

        .dword      HERE
        .dword      CELLP
        .dword      LSUB
        .dword      COMMA
        .dword      SEMIS
#

#  ****************
#  *   (FORWARD   *
#  ****************
#
        .align 3,0
N_PFORWARD:
        .dword      8
        .ascii      "(FORWARD"
        .align 3,0
PFORWARD:
        .dword    DOCOL
        .dword    PFORWARD+HEADSIZE
        .dword    0x0
        .dword    BACKP
        .dword    N_PFORWARD
        .dword    0

        .dword      HERE
        .dword      X
        .dword      COMMA
        .dword      SEMIS
#

#  ****************
#  *   FORWARD)   *
#  ****************
#
        .align 3,0
N_FORWARDP:
        .dword      8
        .ascii      "FORWARD)"
        .align 3,0
FORWARDP:
        .dword    DOCOL
        .dword    FORWARDP+HEADSIZE
        .dword    0x0
        .dword    PFORWARD
        .dword    N_FORWARDP
        .dword    0

        .dword      HERE
        .dword      OVER
        .dword      CELLP
        .dword      LSUB
        .dword      SWAP
        .dword      STOR
        .dword      SEMIS
#

#  *************
#  *   BEGIN   *
#  *************
#
        .align 3,0
N_BEGIN:
        .dword      5
        .ascii      "BEGIN"
        .align 3,0
BEGIN:
        .dword    DOCOL
        .dword    BEGIN+HEADSIZE
        .dword    B_IMMED
        .dword    FORWARDP
        .dword    N_BEGIN
        .dword    0

        .dword      PBACK
        .dword      QCOMP, ONE
        .dword      SEMIS
#

#  ************
#  *   THEN   *
#  ************
#
        .align 3,0
N_THEN:
        .dword      4
        .ascii      "THEN"
        .align 3,0
THEN:
        .dword    DOCOL
        .dword    THEN+HEADSIZE
        .dword    B_IMMED
        .dword    BEGIN
        .dword    N_THEN
        .dword    0

        .dword      QCOMP, TWO, QPAIR
        .dword      FORWARDP
        .dword      SEMIS
#

#  **********
#  *   DO   *
#  **********
#
        .align 3,0
N_DO:
        .dword      2
        .ascii      "DO"
        .align 3,0
DO:
        .dword    DOCOL
        .dword    DO+HEADSIZE
        .dword    B_IMMED
        .dword    THEN
        .dword    N_DO
        .dword    0

         .dword      LIT, XDO, COMMA, PFORWARD, PBACK
        .dword      QCOMP, LIT,3    # Magic number
        .dword      SEMIS
#

#  ***********
#  *   ?DO   *
#  ***********
#
        .align 3,0
N_QDO:
        .dword      3
        .ascii      "?DO"
        .align 3,0
QDO:
        .dword    DOCOL
        .dword    QDO+HEADSIZE
        .dword    B_IMMED
        .dword    DO
        .dword    N_QDO
        .dword    0

         .dword      LIT, XQDO, COMMA, PFORWARD, PBACK
        .dword      QCOMP, LIT,3    # Magic number
        .dword      SEMIS
#

#  ************
#  *   LOOP   *
#  ************
#
        .align 3,0
N_LLOOP:
        .dword      4
        .ascii      "LOOP"
        .align 3,0
LLOOP:
        .dword    DOCOL
        .dword    LLOOP+HEADSIZE
        .dword    B_IMMED
        .dword    QDO
        .dword    N_LLOOP
        .dword    0

        .dword      LIT, ONE, COMMA # POSTPONE 1 
        .dword      PLOOP
        .dword      SEMIS
#

#  *************
#  *   +LOOP   *
#  *************
#
        .align 3,0
N_PLOOP:
        .dword      5
        .ascii      "+LOOP"
        .align 3,0
PLOOP:
        .dword    DOCOL
        .dword    PLOOP+HEADSIZE
        .dword    B_IMMED
        .dword    LLOOP
        .dword    N_PLOOP
        .dword    0

        .dword      QCOMP, LIT, 3, QPAIR
        .dword      LIT, XPLOOP, COMMA
        .dword      LIT, ZBRAN, COMMA, BACKP
        .dword      LIT, UNLOOP, COMMA
        .dword      FORWARDP # For DO to push the leave address.
        .dword      SEMIS
#

#  *************
#  *   UNTIL   *
#  *************
#
        .align 3,0
N_UNTIL:
        .dword      5
        .ascii      "UNTIL"
        .align 3,0
UNTIL:
        .dword    DOCOL
        .dword    UNTIL+HEADSIZE
        .dword    B_IMMED
        .dword    PLOOP
        .dword    N_UNTIL
        .dword    0

        .dword      QCOMP, ONE, QPAIR
        .dword      LIT, ZBRAN, COMMA, BACKP
        .dword      SEMIS
#

#  *************
#  *   AGAIN   *
#  *************
#
        .align 3,0
N_AGAIN:
        .dword      5
        .ascii      "AGAIN"
        .align 3,0
AGAIN:
        .dword    DOCOL
        .dword    AGAIN+HEADSIZE
        .dword    B_IMMED
        .dword    UNTIL
        .dword    N_AGAIN
        .dword    0

        .dword      QCOMP, ONE, QPAIR
        .dword      LIT, BRAN, COMMA, BACKP
        .dword      SEMIS
#

#  **************
#  *   REPEAT   *
#  **************
#
        .align 3,0
N_REPEA:
        .dword      6
        .ascii      "REPEAT"
        .align 3,0
REPEA:
        .dword    DOCOL
        .dword    REPEA+HEADSIZE
        .dword    B_IMMED
        .dword    AGAIN
        .dword    N_REPEA
        .dword    0

        .dword      QCOMP, ONE, QPAIR   # Matches BEGIN ?
        .dword      LIT, BRAN, COMMA, BACKP
        .dword      LIT, 4, QPAIR # Matches WHILE ?
        .dword      FORWARDP # WHILE target. 
        .dword      SEMIS
#

#  **********
#  *   IF   *
#  **********
#
        .align 3,0
N_LIF:
        .dword      2
        .ascii      "IF"
        .align 3,0
LIF:
        .dword    DOCOL
        .dword    LIF+HEADSIZE
        .dword    B_IMMED
        .dword    REPEA
        .dword    N_LIF
        .dword    0

        .dword      LIT, ZBRAN, COMMA, PFORWARD
        .dword      QCOMP, TWO     # Magic number
        .dword      SEMIS
#

#  ************
#  *   ELSE   *
#  ************
#
        .align 3,0
N_LELSE:
        .dword      4
        .ascii      "ELSE"
        .align 3,0
LELSE:
        .dword    DOCOL
        .dword    LELSE+HEADSIZE
        .dword    B_IMMED
        .dword    LIF
        .dword    N_LELSE
        .dword    0

        .dword      QCOMP, TWO, QPAIR
        .dword      LIT, BRAN, COMMA, PFORWARD
        .dword      SWAP
        .dword      FORWARDP
        .dword      TWO     # Magic number
        .dword      SEMIS
#

#  *************
#  *   WHILE   *
#  *************
#
        .align 3,0
N_LWHILE:
        .dword      5
        .ascii      "WHILE"
        .align 3,0
LWHILE:
        .dword    DOCOL
        .dword    LWHILE+HEADSIZE
        .dword    B_IMMED
        .dword    LELSE
        .dword    N_LWHILE
        .dword    0

        .dword      QCOMP, LDUP, ONE, QPAIR
        .dword      TOR    #  Save backward target. 
        .dword      TOR
        .dword      LIT, ZBRAN, COMMA, PFORWARD
        .dword      LIT, 4 # Magic number
        .dword      FROMR
        .dword      FROMR
        .dword      SEMIS
#

#  **************
#  *   SPACES   *
#  **************
#
        .align 3,0
N_SPACES:
        .dword      6
        .ascii      "SPACES"
        .align 3,0
SPACES:
        .dword    DOCOL
        .dword    SPACES+HEADSIZE
        .dword    0x0
        .dword    LWHILE
        .dword    N_SPACES
        .dword    0

        .dword      ZERO
        .dword      MAX
        .dword      ZERO
        .dword     XQDO
        .dword      SPAX1-.-CW
SPAX2:  .dword      SPACE
        .dword     ONE, XPLOOP, ZBRAN
        .dword      SPAX2-.-CW
        .dword      UNLOOP    #LOOP
SPAX1:
        .dword      SEMIS
#

#  **********
#  *   <#   *
#  **********
#
        .align 3,0
N_BDIGS:
        .dword      2
        .ascii      "<#"
        .align 3,0
BDIGS:
        .dword    DOCOL
        .dword    BDIGS+HEADSIZE
        .dword    0x0
        .dword    SPACES
        .dword    N_BDIGS
        .dword    0

        .dword      PAD
        .dword      HLD
        .dword      STOR
        .dword      SEMIS
#

#  **********
#  *   #>   *
#  **********
#
        .align 3,0
N_EDIGS:
        .dword      2
        .ascii      "#>"
        .align 3,0
EDIGS:
        .dword    DOCOL
        .dword    EDIGS+HEADSIZE
        .dword    0x0
        .dword    BDIGS
        .dword    N_EDIGS
        .dword    0

        .dword      DROP
        .dword      DROP
        .dword      HLD
        .dword      FETCH
        .dword      PAD
        .dword      OVER
        .dword      LSUB
        .dword      SEMIS
#

#  ************
#  *   SIGN   *
#  ************
#
        .align 3,0
N_SIGN:
        .dword      4
        .ascii      "SIGN"
        .align 3,0
SIGN:
        .dword    DOCOL
        .dword    SIGN+HEADSIZE
        .dword    0x0
        .dword    EDIGS
        .dword    N_SIGN
        .dword    0

        .dword      ZLESS
        .dword      ZBRAN
        .dword      SIGN1-.-CW #IF
        .dword      LIT, AMS
        .dword      HOLD    #THEN
SIGN1:  .dword      SEMIS
#

#  *********
#  *   #   *
#  *********
#
        .align 3,0
N_DIG:
        .dword      1
        .ascii      "#"
        .align 3,0
DIG:
        .dword    DOCOL
        .dword    DIG+HEADSIZE
        .dword    0x0
        .dword    SIGN
        .dword    N_DIG
        .dword    0



        .dword      BASE
        .dword      FETCH
        .dword      MSMOD
        .dword      ROT
        .dword      LIT,9
        .dword      OVER
        .dword      LESS
        .dword      ZBRAN
        .dword      DIG1-.-CW  #IF
        .dword      LIT,7
        .dword      PLUS    #THEN
DIG1:   .dword      LIT,0x30
        .dword      PLUS
        .dword      HOLD
        .dword      SEMIS
#

#  **********
#  *   #S   *
#  **********
#
        .align 3,0
N_DIGS:
        .dword      2
        .ascii      "#S"
        .align 3,0
DIGS:
        .dword    DOCOL
        .dword    DIGS+HEADSIZE
        .dword    0x0
        .dword    DIG
        .dword    N_DIGS
        .dword    0

DIGS1:  .dword      DIG     #BEGIN
        .dword      OVER
        .dword      OVER
        .dword      LOR
        .dword      ZEQU
        .dword      ZBRAN
        .dword      DIGS1-.-CW #UNTIL
        .dword      SEMIS
#
# 

#  *************
#  *   (D.R)   *
#  *************
#
        .align 3,0
N_PDDOTR:
        .dword      5
        .ascii      "(D.R)"
        .align 3,0
PDDOTR:
        .dword    DOCOL
        .dword    PDDOTR+HEADSIZE
        .dword    0x0
        .dword    DIGS
        .dword    N_PDDOTR
        .dword    0

        .dword      TOR
        .dword      SWAP
        .dword      OVER
        .dword      DABS
        .dword      BDIGS
        .dword      DIGS
        .dword      ROT
        .dword      SIGN
        .dword      EDIGS
        .dword      FROMR
        .dword      OVER
        .dword      LSUB, ZERO, MAX
        .dword      ZERO
        .dword     XQDO
        .dword      PDDOT1-.-CW
PDDOT2:  .dword      LBL, HOLD  #WARNING: HOLD outside of #>.
        .dword     ONE, XPLOOP, ZBRAN
        .dword      PDDOT2-.-CW
        .dword      UNLOOP
PDDOT1:
        .dword      EDIGS  #Drop string instead of number.
        .dword      SEMIS
#

#  ***********
#  *   D.R   *
#  ***********
#
        .align 3,0
N_DDOTR:
        .dword      3
        .ascii      "D.R"
        .align 3,0
DDOTR:
        .dword    DOCOL
        .dword    DDOTR+HEADSIZE
        .dword    0x0
        .dword    PDDOTR
        .dword    N_DDOTR
        .dword    0

        .dword      PDDOTR
        .dword      LTYPE
        .dword      SEMIS
#

#  **********
#  *   .R   *
#  **********
#
        .align 3,0
N_DOTR:
        .dword      2
        .ascii      ".R"
        .align 3,0
DOTR:
        .dword    DOCOL
        .dword    DOTR+HEADSIZE
        .dword    0x0
        .dword    DDOTR
        .dword    N_DOTR
        .dword    0

        .dword      TOR
        .dword      STOD
        .dword      FROMR
        .dword      DDOTR
        .dword      SEMIS
#

#  **********
#  *   D.   *
#  **********
#
        .align 3,0
N_DDOT:
        .dword      2
        .ascii      "D."
        .align 3,0
DDOT:
        .dword    DOCOL
        .dword    DDOT+HEADSIZE
        .dword    0x0
        .dword    DOTR
        .dword    N_DDOT
        .dword    0


        .dword      ZERO
        .dword      DDOTR
        .dword      SPACE
        .dword      SEMIS
#

#  *********
#  *   .   *
#  *********
#
        .align 3,0
N_DOT:
        .dword      1
        .ascii      "."
        .align 3,0
DOT:
        .dword    DOCOL
        .dword    DOT+HEADSIZE
        .dword    0x0
        .dword    DDOT
        .dword    N_DOT
        .dword    0


        .dword      STOD
        .dword      DDOT
        .dword      SEMIS
#

#  *********
#  *   ?   *
#  *********
#
        .align 3,0
N_QUES:
        .dword      1
        .ascii      "?"
        .align 3,0
QUES:
        .dword    DOCOL
        .dword    QUES+HEADSIZE
        .dword    0x0
        .dword    DOT
        .dword    N_QUES
        .dword    0

        .dword      FETCH
        .dword      DOT
        .dword      SEMIS
#

#  **********
#  *   U.   *
#  **********
#
        .align 3,0
N_UDOT:
        .dword      2
        .ascii      "U."
        .align 3,0
UDOT:
        .dword    DOCOL
        .dword    UDOT+HEADSIZE
        .dword    0x0
        .dword    QUES
        .dword    N_UDOT
        .dword    0


        .dword      ZERO
        .dword      DDOT
        .dword      SEMIS
#

#  *****************
#  *   FOR-WORDS   *
#  *****************
#
        .align 3,0
N_FORW:
        .dword      9
        .ascii      "FOR-WORDS"
        .align 3,0
FORW:
        .dword    DOCOL
        .dword    FORW+HEADSIZE
        .dword    0x0
        .dword    UDOT
        .dword    N_FORW
        .dword    0

        .dword      SWAP
        .dword      TOR
        .dword      TOR
FORW1:  .dword      FROMR
        .dword      RR
        .dword      OVER
        .dword      TLFA
        .dword      FETCH
        .dword      TOR
        .dword      EXEC
        .dword      RR
        .dword      ZEQU
        .dword      ZBRAN
        .dword      FORW1-.-CW
        .dword      RDROP
        .dword      RDROP
        .dword      SEMIS
#

#  ****************
#  *   FOR-VOCS   *
#  ****************
#
        .align 3,0
N_FORV:
        .dword      8
        .ascii      "FOR-VOCS"
        .align 3,0
FORV:
        .dword    DOCOL
        .dword    FORV+HEADSIZE
        .dword    0x0
        .dword    FORW
        .dword    N_FORV
        .dword    0

        .dword      TOR
        .dword      VOCL
        .dword      FETCH
        .dword      TOR
FORV1:  .dword      FROMR
        .dword      RR
        .dword      OVER
        .dword      TVFA
        .dword      FETCH
        .dword      TOR
        .dword      EXEC
        .dword      RR
        .dword      ZEQU
        .dword      ZBRAN
        .dword      FORV1-.-CW
        .dword      RDROP
        .dword      RDROP
        .dword      SEMIS
#

#  *************
#  *   WORDS   *
#  *************
#
        .align 3,0
N_WORDS:
        .dword      5
        .ascii      "WORDS"
        .align 3,0
WORDS:
        .dword    DOCOL
        .dword    WORDS+HEADSIZE
        .dword    0x0
        .dword    FORV
        .dword    N_WORDS
        .dword    0

        .dword      CSLL
        .dword      LOUT
        .dword      STOR
        .dword      ZERO            # End sentinel. 
        .dword      LIT, NOOP
        .dword      CONTEXT
        .dword      FETCH
        .dword      FORW
WORDS1:
        .dword      LDUP
        .dword      ZBRAN
        .dword      WORDS2-.-CW
        .dword      IDDOT
        .dword      BRAN
        .dword      WORDS1-.-CW
WORDS2:
        .dword DROP
        .dword      SEMIS
#



#  *****************
#  *   EXIT-CODE   *
#  *****************
#
        .align 3,0
N_XCODE:
        .dword      9
        .ascii      "EXIT-CODE"
        .align 3,0
XCODE:
        .dword    DOVAR
        .dword    XCODE+HEADSIZE
        .dword    0x0
        .dword    WORDS
        .dword    N_XCODE
        .dword    0

        .dword      0
#


#  ***********
#  *   BYE   *
#  ***********
#
        .align 3,0
N_BYE:
        .dword      3
        .ascii      "BYE"
        .align 3,0
BYE:
        .dword    DOCOL
        .dword    BYE+HEADSIZE
        .dword    0x0
        .dword    XCODE
        .dword    N_BYE
        .dword    0

# Exit to linux, with status as stored. 
        .dword      XCODE, FETCH, X, X, LIT, exit, XOS
        .dword      SEMIS   #Unnecessary, but helpful for decompilation.
# 
#

#  ************
#  *   LIST   *
#  ************
#
        .align 3,0
N_LLIST:
        .dword      4
        .ascii      "LIST"
        .align 3,0
LLIST:
        .dword    DOCOL
        .dword    LLIST+HEADSIZE
        .dword    0x0
        .dword    BYE
        .dword    N_LLIST
        .dword    0

        .dword      BASE, FETCH, TOR
        .dword      DECA
        .dword      SCR,STOR

        .dword      SKIP
         .dword      6
SB3: .ascii      "SCR # "
       .align 3,0
        .dword      LIT, SB3
        .dword      LIT, 6
        .dword      LTYPE
        .dword      SCR, FETCH, DOT
        .dword      SCR, FETCH, BLOCK, BBUF

        .dword      LIT, 16, ZERO
        .dword     XDO
        .dword      LLIST9-.-CW
LLIST1: .dword      LIT, ALF, SSLAS
        .dword      CR, IDO, LIT, 2, DOTR
        .dword      SPACE, LTYPE
        .dword     ONE, XPLOOP, ZBRAN
        .dword      LLIST1-.-CW
        .dword      UNLOOP
LLIST9:
        .dword      TDROP
        .dword      FROMR, BASE, STOR
        .dword      SEMIS
#

#  *************
#  *   INDEX   *
#  *************
#
        .align 3,0
N_INDEX:
        .dword      5
        .ascii      "INDEX"
        .align 3,0
INDEX:
        .dword    DOCOL
        .dword    INDEX+HEADSIZE
        .dword    0x0
        .dword    LLIST
        .dword    N_INDEX
        .dword    0

        .dword      LIT,AFF
        .dword      EMIT,CR
        .dword      ONEP,SWAP
        .dword     XDO
        .dword      INDE9-.-CW
INDE1:  .dword      CR,IDO
        .dword      LIT,3
        .dword      DOTR,SPACE
        .dword      ZERO,IDO
        .dword      PLINE, LTYPE
        .dword     ONE, XPLOOP, ZBRAN
        .dword      INDE1-.-CW
        .dword      UNLOOP
INDE9:
        .dword      SEMIS
#

#  **********
#  *   .S   *
#  **********
#
        .align 3,0
N_DOTS:
        .dword      2
        .ascii      ".S"
        .align 3,0
DOTS:
        .dword    DOCOL
        .dword    DOTS+HEADSIZE
        .dword    0x0
        .dword    INDEX
        .dword    N_DOTS
        .dword    0

        .dword      CR
        .dword      LIT, 'S', EMIT
        .dword      LIT, ASO, EMIT
        .dword      SPACE
        .dword      SPFET, SZERO, FETCH
DOC2:   .dword      OVER, OVER,  LSUB, ZLESS
        .dword      ZBRAN
        .dword      DOC1-.-CW
        .dword      ZERO, CELLP, LSUB, LDUP, FETCH
        .dword      DOT
        .dword      BRAN
        .dword      DOC2-.-CW
DOC1:    .dword DROP, DROP
        .dword      LIT, ASC, EMIT
        .dword SEMIS
#

#  ********************
#  *   ENVIRONMENT?   *
#  ********************
#
        .align 3,0
N_ENVQ:
        .dword      12
        .ascii      "ENVIRONMENT?"
        .align 3,0
ENVQ:
        .dword    DOCOL
        .dword    ENVQ+HEADSIZE
        .dword    0x0
        .dword    DOTS
        .dword    N_ENVQ
        .dword    0

        .dword      LIT, ENV, TWID, PFIND
        .dword      TOR, TDROP, FROMR
        .dword      LDUP
        .dword      ZBRAN
        .dword      ENVQ1-.-CW
        .dword      EXEC
        .dword      LIT, -1
ENVQ1:
        .dword      SEMIS
#

# 
#
# This word is not even fig!

#  ***************
#  *   .SIGNON   *
#  ***************
#
        .align 3,0
N_SIGNON:
        .dword      7
        .ascii      ".SIGNON"
        .align 3,0
SIGNON:
        .dword    DOCOL
        .dword    SIGNON+HEADSIZE
        .dword    0x0
        .dword    ENVQ
        .dword    N_SIGNON
        .dword    0

        .dword      CR
        .dword      BASE,FETCH
        .dword      LIT,36, BASE,STOR
        .dword      LCPU, DDOT
        .dword      BASE,STOR
#
        .dword      LENAME, LTYPE, SPACE
        .dword      LVERSION, LTYPE, SPACE
        .dword      CR
        .dword      SEMIS
#

#
#**** LAST DICTIONARY WORD OF FORTH ****

#  ************
#  *   TASK   *
#  ************
#
        .align 3,0
N_TASK:
        .dword      4
        .ascii      "TASK"
        .align 3,0
TASK:
        .dword    DOCOL
        .dword    TASK+HEADSIZE
        .dword    0x0
        .dword    SIGNON
        .dword    N_TASK
        .dword    0

        .dword      SEMIS
#
# End of ONLY namespace. Namespaces all end in a link to 0.

#  *********
#  *   '   *
#  *********
#
        .align 3,0
N_TICK:
        .dword      1
        .ascii      "'"
        .align 3,0
TICK:
        .dword    DOCOL
        .dword    TICK+HEADSIZE
        .dword    B_IMMED + B_DENOT
        .dword    0
        .dword    N_TICK
        .dword    0

        .dword      ITICK
        .dword      LITER
        .dword      SEMIS
#

#  *********
#  *   &   *
#  *********
#
        .align 3,0
N_DCHAR:
        .dword      1
        .ascii      "&"
        .align 3,0
DCHAR:
        .dword    DOCOL
        .dword    DCHAR+HEADSIZE
        .dword    B_IMMED + B_DENOT
        .dword    TICK
        .dword    N_DCHAR
        .dword    0

        .dword      PPFET
        .dword      NIP
        .dword      LDUP, QBL
        .dword      LIT, 10, QERR
        .dword      LITER
        .dword      QDELIM
        .dword      SEMIS

#

#  *********
#  *   ^   *
#  *********
#
        .align 3,0
N_DCTL:
        .dword      1
        .ascii      "^"
        .align 3,0
DCTL:
        .dword    DOCOL
        .dword    DCTL+HEADSIZE
        .dword    B_IMMED + B_DENOT
        .dword    DCHAR
        .dword    N_DCTL
        .dword    0

        .dword      PPFET
        .dword      NIP
        .dword      LDUP, QBL
        .dword      LIT, 10, QERR
        .dword      LIT, '@', LSUB
        .dword      LITER
        .dword      QDELIM
        .dword      SEMIS
#

#  *********
#  *   0   *
#  *********
#
        .align 3,0
N_DEN0:
        .dword      1
        .ascii      "0"
        .align 3,0
DEN0:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DCTL
        .dword    N_DEN0
        .dword    0

#  *********
#  *   1   *
#  *********
#
        .align 3,0
N_DEN1:
        .dword      1
        .ascii      "1"
        .align 3,0
DEN1:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN0
        .dword    N_DEN1
        .dword    0

#  *********
#  *   2   *
#  *********
#
        .align 3,0
N_DEN2:
        .dword      1
        .ascii      "2"
        .align 3,0
DEN2:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN1
        .dword    N_DEN2
        .dword    0

#  *********
#  *   3   *
#  *********
#
        .align 3,0
N_DEN3:
        .dword      1
        .ascii      "3"
        .align 3,0
DEN3:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN2
        .dword    N_DEN3
        .dword    0

#  *********
#  *   4   *
#  *********
#
        .align 3,0
N_DEN4:
        .dword      1
        .ascii      "4"
        .align 3,0
DEN4:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN3
        .dword    N_DEN4
        .dword    0

#  *********
#  *   5   *
#  *********
#
        .align 3,0
N_DEN5:
        .dword      1
        .ascii      "5"
        .align 3,0
DEN5:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN4
        .dword    N_DEN5
        .dword    0

#  *********
#  *   6   *
#  *********
#
        .align 3,0
N_DEN6:
        .dword      1
        .ascii      "6"
        .align 3,0
DEN6:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN5
        .dword    N_DEN6
        .dword    0


#  *********
#  *   7   *
#  *********
#
        .align 3,0
N_DEN7:
        .dword      1
        .ascii      "7"
        .align 3,0
DEN7:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN6
        .dword    N_DEN7
        .dword    0

#  *********
#  *   8   *
#  *********
#
        .align 3,0
N_DEN8:
        .dword      1
        .ascii      "8"
        .align 3,0
DEN8:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN7
        .dword    N_DEN8
        .dword    0

#  *********
#  *   9   *
#  *********
#
        .align 3,0
N_DEN9:
        .dword      1
        .ascii      "9"
        .align 3,0
DEN9:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN8
        .dword    N_DEN9
        .dword    0

#  *********
#  *   A   *
#  *********
#
        .align 3,0
N_DENA:
        .dword      1
        .ascii      "A"
        .align 3,0
DENA:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEN9
        .dword    N_DENA
        .dword    0

#  *********
#  *   B   *
#  *********
#
        .align 3,0
N_DENB:
        .dword      1
        .ascii      "B"
        .align 3,0
DENB:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DENA
        .dword    N_DENB
        .dword    0


#  *********
#  *   C   *
#  *********
#
        .align 3,0
N_DENC:
        .dword      1
        .ascii      "C"
        .align 3,0
DENC:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DENB
        .dword    N_DENC
        .dword    0

#  *********
#  *   D   *
#  *********
#
        .align 3,0
N_DEND:
        .dword      1
        .ascii      "D"
        .align 3,0
DEND:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DENC
        .dword    N_DEND
        .dword    0

#  *********
#  *   E   *
#  *********
#
        .align 3,0
N_DENE:
        .dword      1
        .ascii      "E"
        .align 3,0
DENE:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DEND
        .dword    N_DENE
        .dword    0

#  *********
#  *   F   *
#  *********
#
        .align 3,0
N_DENF:
        .dword      1
        .ascii      "F"
        .align 3,0
DENF:
        .dword    DOCOL
        .dword    LNUMB
        .dword    B_IMMED + B_DENOT
        .dword    DENE
        .dword    N_DENF
        .dword    0

#

#  *********
#  *   -   *
#  *********
#
        .align 3,0
N_DENM:
        .dword      1
        .ascii      "-"
        .align 3,0
DENM:
        .dword    DOCOL
        .dword    DENM+HEADSIZE
        .dword    B_IMMED + B_DENOT
        .dword    DENF
        .dword    N_DENM
        .dword    0

        .dword      PNUMB, DNEGA, SDLITE
        .dword      SEMIS
#

#  *********
#  *   +   *
#  *********
#
        .align 3,0
N_DENP:
        .dword      1
        .ascii      "+"
        .align 3,0
DENP:
        .dword    DOCOL
        .dword    DENP+HEADSIZE
        .dword    B_IMMED + B_DENOT
        .dword    DENM
        .dword    N_DENP
        .dword    0

        .dword      PNUMB, SDLITE
        .dword      SEMIS
#

#  *********
#  *   "   *
#  *********
#
        .align 3,0
N_DENQ:
        .dword      1
        .ascii      "\""
        .align 3,0
DENQ:
        .dword    DOCOL
        .dword    DENQ+HEADSIZE
        .dword    B_IMMED + B_DENOT
        .dword    DENP
        .dword    N_DENQ
        .dword    0

        .dword      LIT, SKIP, COMMA        #  'SKIP , HERE >R 0 ,
        .dword      HERE, TOR, ZERO, COMMA
DENQ1:
        .dword      LIT, '"', PARSE         #           BEGIN &" PARSE
        .dword      PPFET, LDUP, LIT, '"', EQUAL #           PP@@ DUP &" =
        .dword      ZBRAN
        .dword      DENQ2-.-CW                 #           WHILE
        .dword      TDROP, ONEP             #           2DROP 1+ R@ $+!
        .dword      LDUP, ALLOT, RR, SADD
        .dword      BRAN
        .dword      DENQ1-.-CW                  #           REPEAT
DENQ2:
        .dword      QBL, ZEQU
        .dword      LIT, 10, QERR           #           ?BLANK 0= 10 ?ERROR
        .dword      DROP                    #                DROP 
        .dword      LDUP, ALLOT, RR, SADD     #         DUP ALLOT R@ $+!
    .dword LALIGN
        .dword      FROMR, SFET, DLITE      #           R> $@ POSTPONE DLITERAL ;
        .dword      SEMIS
#

# The FORTH word is the only non-denotation in the ONLY namespace.
#  *************
#  *   FORTH   *
#  *************
#
        .align 3,0
N_FORTH:
        .dword      5
        .ascii      "FORTH"
        .align 3,0
FORTH:
        .dword    DODOE
        .dword    FORTH+HEADSIZE
        .dword    0x0
        .dword    DENQ
        .dword    N_FORTH
        .dword    0

        .dword      DOVOC
        .dword      0       # END OF NAMESPACE LIST

        .align 3,0

        .dword    0x0
        .dword    0
        .dword    B_DUMMY
        .dword    TASK
        .dword    (ZERO+(CW*(D_HOFFSET)))
        .dword    0

#
STARTOFONLY    = FORTH
#
#

#  ************
#  *   CORE   *
#  ************
#
        .align 3,0
N_CORE:
        .dword      4
        .ascii      "CORE"
        .align 3,0
CORE:
        .dword    DOCOL
        .dword    CORE+HEADSIZE
        .dword    0x0
        .dword    0
        .dword    N_CORE
        .dword    0

        .dword      ZERO    # Not (fully) present.
        .dword      SEMIS
#

#  ***********
#  *   CPU   *
#  ***********
#
        .align 3,0
N_LCPU:
        .dword      3
        .ascii      "CPU"
        .align 3,0
LCPU:
        .dword    DOCOL
        .dword    LCPU+HEADSIZE
        .dword    0x0
        .dword    CORE
        .dword    N_LCPU
        .dword    0

       .dword      LIT, 0x02C15B0F, ZERO      # 'RISCV'
        .dword      SEMIS
#

#  ***************
#  *   VERSION   *
#  ***************
#
        .align 3,0
N_LVERSION:
        .dword      7
        .ascii      "VERSION"
        .align 3,0
LVERSION:
        .dword    DOCOL
        .dword    LVERSION+HEADSIZE
        .dword    0x0
        .dword    LCPU
        .dword    N_LVERSION
        .dword    0

 /* 
#       If this is there it is an official release
        .dword      SKIP
        .dword      22
IBMPC:  .ascii     "IBM-PC ciforth"
        .byte      FIGREL+0x40,ADOT,FIGREV+0x30,ADOT,USRVER+0x30
        .dword      LIT, IBMPC, LIT, 22
 */  #       If M4_VERSION exists and contains a . it is an official release
        .dword      SKIP
         .dword      14
SB4: .ascii      "beta 2023Mar30"
       .align 3,0
        .dword      LIT, SB4
        .dword      LIT, 14
        .dword      SEMIS
#

#  ************
#  *   NAME   *
#  ************
#
        .align 3,0
N_LENAME:
        .dword      4
        .ascii      "NAME"
        .align 3,0
LENAME:
        .dword    DOCOL
        .dword    LENAME+HEADSIZE
        .dword    0x0
        .dword    LVERSION
        .dword    N_LENAME
        .dword    0

        .dword      SKIP
         .dword      7
SB5: .ascii      "ciforth"
       .align 3,0
        .dword      LIT, SB5
        .dword      LIT, 7
        .dword      SEMIS
#

#  ****************
#  *   SUPPLIER   *
#  ****************
#
        .align 3,0
N_SUPPLIER:
        .dword      8
        .ascii      "SUPPLIER"
        .align 3,0
SUPPLIER:
        .dword    DOCOL
        .dword    SUPPLIER+HEADSIZE
        .dword    0x0
        .dword    LENAME
        .dword    N_SUPPLIER
        .dword    0

        .dword      SKIP
         .dword      20
SB6: .ascii      "Albert van der Horst"
       .align 3,0
        .dword      LIT, SB6
        .dword      LIT, 20
        .dword      SEMIS
#
STARTOFENVIRONMENT    =   SUPPLIER
#
#
 #  

 /* 

The remaining memory ( up to 'EM' ) is
used for:

        1. EXTENSION DICTIONARY
        2. PARAMETER STACK
        3. TERMINAL INPUT BUFFER
        4. RETURN STACK
        5. USER VARIABLE AREA
        6. DISK BUFFERS (UNLESS REQUIRED <1 MBYTE)


 */  

#       This is the proper way to do it.
#       No memory addresses should be arrived at through equates.
#       However now we must teach the linker to keep the
#       two sections together.

FORTHSIZE       =     . - _main
# We must use bss, because of the c-library.
# The dictionary is not a normal bss, because it is executable.
# The linker -N option takes care of that. 
       .bss 

INITDP:                 #  It may be that it is not consecutive with TASK.
                        #  Because of bad linking, resulting in a unusable Forth.
        BUFFERSIZE      =  (KBBUF+2*CW)*NBUF

# For more task the TAS area is duplicated; the remainder is the same.
        .space    EM - FORTHSIZE - TAS - BUFFERSIZE 
# INITR1:                         Return stack for next task
#          .space    US              #  For next task, if present. 
# The following framework may be repeated for more tasks.
# It is floating, i.e. it may be shifted up and down, in order
# to have INITR0 at a STACKSIZE boundary. 
        .space    STACKSIZE - US    # Data stack, only limited if more tasks present.
INITS0:                         # Grows down.
STRTIB: .space    TIBSIZE     # Terminal input buffer.
        .space    STACKSIZE       # Return stack, grows down.
INITR0:                         # Grows down
STRUSA: .space    US              # User area
BUF1:   .space    BUFFERSIZE      # FIRST DISK BUFFER
ACTUAL_EM:  # Differs from EM if code is relocated.

        
# 
#


        #
 /* 

  MISC. NOTES AND SCATTERED THOUGHTS

- Remember that all the FORTH words in this version are
  upper case letters.  Use <CAPS LOCK> when in FORTH.



- Subscribe to FORTH Dimensions.  It is a valuable source
  of system and application ideas.  Talking with fellow
  FORTH programmers is sure to stir up some exciting ideas.
  Consider joining a FIG chapter.  See the back of FORTH
  Dimensions for more info.

 */  
# Define the entry point, not valid for auto booting.
        
        # _ORIG


























#





